/*
   This file is part of the RELXILL model code.

   RELXILL is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   RELXILL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.
   For a copy of the GNU General Public License see
   <http://www.gnu.org/licenses/>.

    Copyright 2016 Thomas Dauser, Remeis Observatory & ECAP

    Modified by Michal Szanecki (2018):                                                                                removed parts related with relativistic transfer of radiation 
*/
#ifndef COMMON_H_
#define COMMON_H_

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <float.h>

/*********** DEFINE STATEMENTS *********/

/** define primary spectrum Type **/
#define PRIM_SPEC_ECUT 1
#define PRIM_SPEC_NTHCOMP 2
/***************************************/

/****** TYPE DEFINITIONS ******/

typedef struct{
	double gam;
	double afe;
	double lxi;
	double ect;
	double incl;
	double z;
	double refl_frac;
	double dens;
	int fixReflFrac;
	int model_type;
	int prim_type;
} xillParam;

/** the XILLVER table structure */
typedef struct{

	float* elo;
	float* ehi;
	int n_ener;

	int num_param;

	float* gam;
	float* afe;
	float* lxi;
	float* ect;
	float* incl;

	int n_gam;
	int n_afe;
	int n_lxi;
	int n_ect;
	int n_incl;

	int n_ca_gam;
	int n_ca_afe;
	int n_ca_lxi;
	int n_ca_ect;
	int n_ca_incl;

	float****** dat;

}xillTable;

typedef struct{
	double* ener;
	double* incl;
	double** flu;
	int n_ener;
	int n_incl;
}xill_spec;


/******************************/
/* define the c_donthcomp function here */
void c_donthcomp(double *ear, int ne, double* param, double *photar);

#endif /* COMMON_H_ */
