/*
   This file is part of the RELXILL model code.

   RELXILL is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   RELXILL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.
   For a copy of the GNU General Public License see
   <http://www.gnu.org/licenses/>.

    Copyright 2017 Thomas Dauser, Remeis Observatory & ECAP

    removed parts related with relativistic transfer of radiation 
    Modified by Michal Szanecki (2018) 
*/
#ifndef MODELS_H_
#define MODELS_H_

#include "xilbase.h"
#include "xilutility.h"
#include "xilltable.h"

/**** DEFINES **/

#define MOD_TYPE_XILLVER 0
#define NUM_PARAM_XILLVER 7

#define MOD_TYPE_XILLVERDENS -100
#define NUM_PARAM_XILLVERDENS 7

/**** FUNCTION DEFINITIONS ****/
xillParam* init_par_xillver(const double* inp_par, const int n_parameter, int* status);

/** basic xillver model function **/
void xillver_base(const double* ener0, const int n_ener0, double* photar, xillParam* param_struct, int* status);

/** internal MODEL FUNCTIONS **/
void tdxillver(const double* ener0, const int n_ener0, double* photar, const double* parameter, const int n_parameter, int* status);
void tdxillverdens(const double* ener0, const int n_ener0, double* photar, const double* parameter, const int n_parameter, int* status);
void tdxillver_nthcomp(const double* ener0, const int n_ener0, double* photar, const double* parameter, const int n_parameter, int* status);

/* get the version number text on the screen (if not already printed before */
void print_version_number(int* status);

xillParam* new_xillParam(int model_type, int prim_type, int* status);
void free_xillParam(xillParam*);

/* xspec local model wrapper functions **/
void lmodxillver(const double* ener0, const int n_ener0, const double* parameter, int ifl, double* photar, double* photer, const char* init);
void lmodxillverdens(const double* ener0, const int n_ener0, const double* parameter, int ifl, double* photar, double* photer, const char* init);
void lmodxillvernthcomp(const double* ener0, const int n_ener0, const double* parameter, int ifl, double* photar, double* photer, const char* init);

#endif /* MODELS_H_ */
