! This file is part of the REFLKERR model.                                         
! You can redistribute it.       
! Please refer to Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942  

! This file (xreflecth.f90) contains basic model functions and variables and
! local frame model routines

MODULE par_global
  IMPLICIT NONE
  SAVE
  double precision, parameter :: pi=3.141592654
  character (len=240) :: &
    reflkerr_tables = '/work/aaz_team/XSPEC_MODELS/reflkerr_lpbb/dev/4'
  integer :: mtp = -1 
  double precision, parameter :: inenl=0.01!7.17702e-2/1e-3
  double precision, parameter :: inenh=2500.0 !5000.0
  double precision, parameter :: inenl_iref=3. !3.
  double precision, parameter :: inenh_iref=2500. !2500.
  INTEGER, PARAMETER :: inbn=2970 !5000/2500/2970
  double precision :: inen(0:inbn)
  REAL :: ginen(0:inbn)
  REAL :: inphotar_p(inbn),inphotar_r(inbn)

  logical :: cached_inen = .false.
  logical :: cached_primary = .false.

  REAL :: gshft=-1.0
  logical :: is_lamppost = .false.
  LOGICAL :: rchatty=.FALSE.
  LOGICAL :: rnobb=.FALSE.
  logical :: is_lamppost_bb = .false.

END MODULE par_global

MODULE par_xilcps
  USE par_global
  IMPLICIT NONE
  SAVE

  character(LEN=120):: fGkTtau,fGkTmutau,firrslb
  logical :: cached_kTt_G_tbp = .false.
  logical :: cached_firrslb_tbp = .false.

  REAL :: pRedshft,pEcut,gEcut,puse_Ec,pTauGam,pGamma,pLgXi,pXi,pAFe,pIncl,pLgN,pRel_refl
  REAL :: pkTe,pKTbb, pTauy, pGeom, pCosi
  REAL :: pEind=2.0
  REAL :: pGmin =-1.0
  REAL :: pGmax =-1.0
  REAL :: pHR =1.0
  REAL :: pCovf =1.0
  REAL :: pAMe =1.0
  REAL :: pTdisk =1.0e6
  REAL :: pBet10 =-10.0
  REAL :: pRin = 10.0
  REAL :: pRout = 1000.0
  REAL :: flux2,flux2_30

  REAL, PARAMETER :: g1_kT_l=10.0,g1_kT_h=500.0
  REAL, PARAMETER :: g1_kTbb_l=1.0e-3,g1_kTbb_h=1.0
  REAL, PARAMETER :: g1_tau_l=0.05,g1_tau_h=4.0
  REAL, PARAMETER :: g1_incl_l=9.0,g1_incl_h=89.3
  REAL :: g1_y_l,g1_y_h
  
  REAL :: pLineE

  REAL :: RMtype

  logical :: use_iref=.TRUE.
  logical :: use_xil=.TRUE.
  logical :: cached_refl_1 = .false.
  logical :: cached_refl_2ir = .false.
  logical :: cached_refl_2ps = .false.
  logical :: cached_refl_fl1 = .false.

END MODULE par_xilcps


SUBROUTINE get_primary()

  USE par_global, only: mtp,inbn,inen,inphotar_p
  USE par_global, only: cached_primary,rnobb,rchatty
  USE par_xilcps, only: pGeom
  IMPLICIT NONE

  REAL, ALLOCATABLE :: param(:)
  REAL :: photer(inbn)
  INTEGER :: lmtp(10)
  INTEGER :: npar(10)
  LOGICAL,SAVE::firstcall=.TRUE.
  REAL :: tau
  REAL,SAVE :: cached_tau=0.0
  INTEGER :: i
  REAL::photar_bb(inbn),tmp

  npar(1:10)=(/19,19,9,1,19,19,9,19,19,19/)
  lmtp(1:10)=(/1,1,2,5,1,1,2,1,1,1/)

  if (allocated(param)) then                               
     deallocate (param)                                    
  endif

  allocate(param(npar(mtp)))
  
  call init_inen()

  call INIT_PAR_PRIMARY(param,npar(mtp))

  call check_cache_primary()

  IF(.not.cached_primary) THEN

     if(lmtp(mtp).eq.1)then
        if(param(6).ge.0.0)then
           tau=param(6)
        else
           tau=127.75*abs(param(6))/param(1)
        endif
        !        if(firstcall.or.((cached_tau.le.3.4).and.(tau.gt.3.4)))then
        if(firstcall.or.(tau.gt.3.4))then
           tmp=param(6)
           param(6)=sign(1.1*abs(param(6)),param(6))
           call get_xspec_model(lmtp(mtp), real(inen), inbn, param, npar(mtp),inphotar_p, photer)
           param(6)=tmp
           firstcall=.FALSE.
           if(rchatty)write(*,*)"Primary preinitiated"
        endif
        cached_tau=tau
     endif

     call get_xspec_model(lmtp(mtp), real(inen), inbn, param, npar(mtp),inphotar_p, photer)

     IF((lmtp(mtp).eq.1).and.rnobb.and.(pGeom.ne.1))THEN
        param(9)=-param(9)
        call get_xspec_model(lmtp(mtp), real(inen), inbn, param, npar(mtp),photar_bb, photer)
        param(9)=-param(9)
        do i=1,inbn
           inphotar_p(i)=inphotar_p(i)-photar_bb(i)
        enddo
        if(rchatty)write(*,*)"BB removed"
     ENDIF
  ENDIF

  deallocate (param)

  return
END SUBROUTINE get_primary


SUBROUTINE get_reflection()
  USE par_global, only: mtp,inbn,inen,inphotar_p,inphotar_r,rchatty,pi,inenl_iref,inenh_iref
  USE par_xilcps, only: cached_refl_1, cached_refl_2ir, cached_refl_2ps, cached_refl_fl1, RMtype, pGeom
  USE par_xilcps, only: flux2,flux2_30,pXi,use_iref,use_xil,pGamma,pIncl,pAFe
  
  IMPLICIT NONE

  INTEGER :: ifl, i, ilg, ilAfe

  REAL, ALLOCATABLE :: param1(:),param2(:)
  REAL :: photar1(inbn),photar2(inbn)
  DOUBLE PRECISION :: photar2_dp(inbn)
  DOUBLE PRECISION,ALLOCATABLE,SAVE :: photar2_iref(:)
  REAL,ALLOCATABLE,SAVE :: photar2_iref_real(:)
  double precision :: photar(inbn)
  REAL :: param_iref(7)
  INTEGER :: npar(2,10)
  INTEGER :: lmtp(2,10)
  REAL :: photer(inbn)
  REAL :: flux1,flux30,f_prim

  INTEGER,PARAMETER::nbgl=13
  INTEGER,PARAMETER::nbAfe=19

  REAL,PARAMETER :: eming=0.9
  REAL,PARAMETER :: eflx=1.0
  REAL,PARAMETER :: emaxg=1.1

  REAL,PARAMETER :: eminxi=89.0
  REAL,PARAMETER :: emaxxi=91.0
  REAL,PARAMETER :: eflxxi=90.0

  REAL,PARAMETER :: gl_l=1.6
  REAL,PARAMETER :: gl_h=2.9
  REAL,PARAMETER :: Afel_l=0.5
  REAL,PARAMETER :: Afel_h=10.0

  
  INTEGER,PARAMETER :: fln=200
  REAL :: de,efl(0:fln),photfl(inbn),photfler(inbn)
  REAL :: enflx(0:fln)
  LOGICAL :: cross=.TRUE.
  INTEGER :: xi_iter
  REAL :: xi_tmp

  REAL :: parg1(2),parg2(2),phg1(inbn),phg2(inbn),scale_lka

  REAL :: dgl,gamma,g_l,g_h,glfac,lgnormi
  REAL :: dAfe,Afe,Afe_l,Afe_h,Afefac,Afenormi
  REAL :: lgnorm(nbgl+1),AFenorm(nbAfe+1)

  character FGMSTR*20, FGMSTR_val*20

  INTEGER,SAVE :: ienl_iref,ienh_iref,inbn_iref
  REAL,ALLOCATABLE,SAVE :: inen_iref(:)
  

  SAVE photar1, photar2, flux1, f_prim
 
  FGMSTR_val= trim(FGMSTR("RUSEIREF"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) i
     if(i.eq.1)then
        use_iref=.TRUE.
     else
        use_iref=.FALSE.
     endif
  else
     use_iref=.TRUE.
  endif

  FGMSTR_val= trim(FGMSTR("RUSEXIL"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) i
     if(i.eq.1)then
        use_xil=.TRUE.
     else
        use_xil=.FALSE.
     endif
  else
     use_xil=.TRUE.
  endif

  ifl=0

  npar(1,1:10)=(/7,7,7,0,7,7,7,7,7,7/)
  npar(2,1:10)=(/19,19,9,0,19,19,9,19,19,19/)
  lmtp(1,1:10)=(/3,3,4,0,6,6,6,3,4,4/)
  lmtp(2,1:10)=(/1,1,2,0,1,1,2,1,1,1/)

  lgnorm(1:nbgl+1)=(/2.5184097,1.9967512,1.5779703,1.2472153,1.0,0.79115099, &
       0.62886757,0.50293936,0.40052599,0.31868812,0.25340347,0.20436262,0.16336634,0.13041460/)
  AFenorm(1:nbAfe+1)=(/0.60279990,1.0,1.1953046,1.3906092,1.5859138,1.7812185, &
       1.9765231,2.1718276,2.3671323,2.5624370,2.6413670,2.7202970,2.7992267, &
       2.8781567,2.9570867,3.0360167,3.1149467,3.1938766,3.2728064,3.3517364/)

  if (allocated(param1)) then  
     deallocate (param1)                                    
  endif

  if (allocated(param2)) then  
     deallocate (param2)                                    
  endif

  allocate(param1(npar(1,mtp)))
  allocate(param2(npar(2,mtp)))

  call init_inen()

  de=Log10(emaxg/eming)/fln
  do i=0,fln
     efl(i)=eming*10.0**(i*de)
  enddo

  de=Log10(emaxxi/eminxi)/fln
  do i=0,fln
     enflx(i)=eminxi*10.0**(i*de)
  enddo

  call INIT_PAR_REFLECTION(param1,npar(1,mtp),param2,npar(2,mtp),param_iref)

  call check_cache_reflection()

  xi_iter=0
  xi_tmp=pXi

1 CONTINUE

  IF(.not.use_iref)GOTO 20!no ireflect

  if(.not.allocated(inen_iref))then
     call f_binary_search_float(inenl_iref,ienl_iref,inen,inbn+1)
     call f_binary_search_float(inenh_iref,ienh_iref,inen,inbn+1)
     inbn_iref=int((ienh_iref-ienl_iref+1)/5.5)     
!     write(*,*)inbn_iref
     allocate(inen_iref(0:inbn_iref))

     de=real(dlog10(inenh_iref/inenl_iref)/inbn_iref)
     do i=0,inbn_iref
        inen_iref(i) = real(inenl_iref*10.0**(i*de))
     enddo
     allocate(photar2_iref(inbn_iref))
     allocate(photar2_iref_real(inbn_iref))
  endif

  
  IF(RMtype.eq.0.0) THEN
     !empty block
  ELSE IF(RMtype.eq.1.0) THEN
     if(((mtp.eq.1).or.(mtp.eq.2).or.(mtp.eq.5).or. &
          (mtp.eq.6).or.(mtp.eq.8).or.(mtp.eq.9).or.(mtp.eq.10)).and. &
          (pGeom.eq.1.0))then
        IF((.not.cached_refl_2ps).or.(.not.cross))THEN
           if(.not.cross)param2(14)=xi_tmp
           call get_xspec_model(lmtp(2,mtp), real(inen), inbn, param2, npar(2,mtp),photar2,photer)
           !write(*,*)"calling compps"
        ENDIF
     else
        IF((.not.cached_refl_2ir).or.(.not.cross))THEN
           call f_rebin_spectrum(1.d0*inen_iref,inbn_iref,1.d0*inen,inbn,photar2_iref,1.d0*inphotar_p)
           photar2_iref_real=real(photar2_iref)
           
           if(.not.cross)param_iref(7)=xi_tmp
           call ireflct(inen_iref,inbn_iref,param_iref,ifl,photar2_iref_real,photer)
           photar2_dp=0.0
           photar2=0.0
           call f_rebin_spectrum(1.d0*inen,inbn,1.d0*inen_iref,inbn_iref,photar2_dp,1.d0*photar2_iref_real)
           photar2=real(photar2_dp)
        ENDIF
     endif
  ELSE IF(RMtype.eq.2.0)then !not used now
     IF((.not.cached_refl_2ps))THEN
        call get_xspec_model(lmtp(2,mtp), real(inen), inbn, param2, npar(2,mtp),photar2,photer)
     ENDIF
  ENDIF

20 CONTINUE !no ireflect

  if(use_xil)then

     IF(.not.cached_refl_fl1)THEN
        param1(7)=0.0
        call get_xspec_model(lmtp(1,mtp), real(inen), inbn, param1, npar(1,mtp),photfl,photfler)
        param1(7)=-1.0
        call f_binary_search_float(1.d0,i,inen,inbn+1)
        flux1=real(photfl(i)/(inen(i)-inen(i-1)))
        f_prim=1.0
        if(((mtp.eq.1).or.(mtp.eq.2).or.(mtp.eq.8)).and.(pGeom.ne.1))then
           call f_binary_search_float(30.0d0,i,inen,inbn+1)
           flux30=real(photfl(i)/(inen(i)-inen(i-1)))
           f_prim=(flux2_30/flux2)/(flux30/flux1)
           if(f_prim.lt.0.7)f_prim=1.0
        endif
        !write(*,*)"calling xillver (flux)"
     ENDIF

     IF(rchatty)write(*,*)"f_prim=",f_prim

     IF(.not.cached_refl_1)THEN
        !write(*,*)"calling xillver refl"
        call get_xspec_model(lmtp(1,mtp), real(inen), inbn, param1, npar(1,mtp),photar1,photer)
     ENDIF

     IF((.not.cached_refl_1).or.(.not.cached_refl_2ps) &
          .or.(.not.cached_refl_2ir).or.(.not.cached_refl_fl1) &
          .or.(.not.cross))THEN
        call merge(photar1, photar2, real(inen), inbn, flux1, flux2*f_prim, photar, cross)
        if((.not.cross).and.(xi_iter.lt.3))then
           xi_iter=xi_iter+1
           xi_tmp=3.3*xi_tmp
           if(xi_tmp.gt.1.0e5)then
              xi_tmp=1.0e5
              xi_iter=3
           endif
           GOTO 1
        endif

        do i=1,inbn
           inphotar_r(i)=real(photar(i))
        enddo

     ENDIF

  else

     !gaussian Kalpha & Kbeta lines                                                                                    
     parg1(1)=6.405
     parg1(2)=1.0e-6
     parg2(1)=7.059
     parg2(2)=1.0e-6

     call xsgaul(real(inen),inbn,parg1,ifl,phg1,photer)
     call xsgaul(real(inen),inbn,parg2,ifl,phg2,photer)

     dgl=(gl_h-gl_l)/nbgl
     dAfe=(Afel_h-Afel_l)/nbAfe

     gamma=pGamma
     Afe=pAfe
     if(gamma.lt.gl_l)gamma=gl_l
     if(gamma.gt.gl_h)gamma=gl_h
     if(Afe.lt.Afel_l)Afe=Afel_l
     if(Afe.gt.Afel_h)Afe=Afel_h

     ilg=INT(1+(gamma-gl_l)/dgl)
     ilAfe=INT(1+(Afe-Afel_l)/dAfe)
     if(ilg.lt.1)ilg=1
     if(ilg.gt.nbgl)ilg=ilg-1
     if(ilAfe.lt.1)ilAfe=1
     if(ilAfe.gt.nbAfe)ilAfe=ilAfe-1

     g_l=gl_l+(ilg-1)*dgl
     g_h=gl_l+ilg*dgl
     Afe_l=Afel_l+(ilAfe-1)*dAfe
     Afe_h=Afel_l+ilAfe*dAfe

     glfac=(gamma-g_l)/(g_h-g_l)
     Afefac=(Afe-Afe_l)/(Afe_h-Afe_l)

     lgnormi=(1.0-glfac)*lgnorm(ilg)+glfac*lgnorm(ilg+1)
     Afenormi=(1.0-Afefac)*Afenorm(ilAfe)+Afefac*Afenorm(ilAfe+1)
     call f_binary_search_float(6.405*1.0d0,i,inen,inbn+1)

     scale_lka=REAL(COS(pIncl*pi/180.0)/phg1(i)*lgnormi*Afenormi*0.007)
     do i=1,inbn
        inphotar_r(i)=photar2(i)+scale_lka*flux2*(phg1(i)+0.113333333*phg2(i))
     enddo

  endif

  deallocate (param1)
  deallocate (param2)

return
END SUBROUTINE get_reflection


SUBROUTINE get_xspec_model(lmtp, ear, ne, param, npar, photar, photer)
  IMPLICIT NONE
  
  INTEGER :: lmtp, ne, npar, ifl
  REAL :: ear(0:ne), param(npar), photar(ne),photer(ne)
  
  ifl=0

  IF(lmtp.eq.1) call xscompps(ear, ne, param, ifl, photar, photer)
  IF(lmtp.eq.2) call xspexriv(ear, ne, param, ifl, photar, photer)
  IF(lmtp.eq.3) call f_rxillvernthcomp(ear, ne, param, ifl, photar,photer)
  IF(lmtp.eq.4) call f_rxillver(ear, ne, param, ifl, photar,photer)
  IF(lmtp.eq.5) call xsline(ear, ne, param, ifl, photar, photer)
  IF(lmtp.eq.6) call f_rxillverD(ear, ne, param, ifl, photar, photer)

  return
END SUBROUTINE get_xspec_model


SUBROUTINE f_hrefl(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_xilcps, only: RMtype,pGeom,puse_Ec
  use par_xilcps, only: pTauGam,pEcut,pIncl,pKTbb
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_kTbb_l,g1_kTbb_h,g1_tau_l,g1_tau_h, &
       g1_incl_l,g1_incl_h,g1_y_l,g1_y_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 10
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.

  call init_par_hrefl(param,num_param)

  IF(abs(puse_Ec).gt.0.0)THEN
     mtp=9
  ELSE
     mtp=1
  ENDIF

  RMtype=1.0

  if(pGeom.eq.-1.0) pGeom=1.0
  if((abs(pGeom).gt.1.0).and.(abs(pGeom).lt.4.0)) pGeom=0.0
  
  IF(pGeom.eq.1.0)THEN
     if(pTauGam.lt.0.0)then
        g1_y_l=-g1_tau_h*(0.0078277886*pEcut)
        g1_y_h=-g1_tau_l*(0.0078277886*pEcut)
        IF(pTauGam.lt.g1_y_l)pTauGam=g1_y_l
        IF(pTauGam.gt.g1_y_h)pTauGam=g1_y_h
     else
        IF(pTauGam.lt.g1_tau_l)pTauGam=g1_tau_l
        IF(pTauGam.gt.g1_tau_h)pTauGam=g1_tau_h
     endif
     
     if(pEcut.lt.g1_kT_l)pEcut=g1_kT_l
     if(pEcut.gt.g1_kT_h)pEcut=g1_kT_h
     IF(pkTbb.gt.0.)THEN
        if(pkTbb.lt.g1_kTbb_l)pkTbb=g1_kTbb_l
        if(pkTbb.gt.g1_kTbb_h)pkTbb=g1_kTbb_h
     ELSE
        if((-pkTbb).lt.g1_kTbb_l)pkTbb=-g1_kTbb_l
        if((-pkTbb).gt.g1_kTbb_h)pkTbb=-g1_kTbb_h
     ENDIF
     if(pIncl.lt.g1_incl_l)pIncl=g1_incl_l
     if(pIncl.gt.g1_incl_h)pIncl=g1_incl_h
  ENDIF

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hrefl


SUBROUTINE f_hreflG(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_xilcps, only: pGeom,puse_Ec,pkTbb,RMtype
  use par_xilcps, only: pIncl,pEcut
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_incl_l,g1_incl_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 9
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.

  call init_par_hreflG(param,num_param)

  IF(abs(puse_Ec).gt.0.0)THEN
     mtp=10
  ELSE
     mtp=2
  ENDIF

  RMtype=1.0
  pkTbb=1.0e-3

  if(pGeom.eq.-1.0) pGeom=1.0
  if(abs(pGeom).ne.1.0)pGeom=0.0

  IF(pGeom.eq.1.0)THEN
     if(pEcut.lt.g1_kT_l)pEcut=g1_kT_l
     if(pEcut.gt.g1_kT_h)pEcut=g1_kT_h
     if(pIncl.lt.g1_incl_l)pIncl=g1_incl_l
     if(pIncl.gt.g1_incl_h)pIncl=g1_incl_h
  ENDIF

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hreflG


SUBROUTINE f_hreflD(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_xilcps, only: pGeom,RMtype
  use par_xilcps, only: pTauGam,pEcut,pIncl,pKTbb
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_kTbb_l,g1_kTbb_h,g1_tau_l,g1_tau_h, &
       g1_incl_l,g1_incl_h,g1_y_l,g1_y_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 10
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.

  call init_par_hreflD(param,num_param)

  mtp=5
  RMtype=1.0

  if(pGeom.eq.-1.0) pGeom=1.0
  if((abs(pGeom).gt.1.0).and.(abs(pGeom).lt.4.0)) pGeom=0.0  

  IF(pGeom.eq.1.0)THEN
     if(pTauGam.lt.0.0)then
        g1_y_l=-g1_tau_h*(0.0078277886*pEcut)
        g1_y_h=-g1_tau_l*(0.0078277886*pEcut)
        IF(pTauGam.lt.g1_y_l)pTauGam=g1_y_l
        IF(pTauGam.gt.g1_y_h)pTauGam=g1_y_h
     else
        IF(pTauGam.lt.g1_tau_l)pTauGam=g1_tau_l
        IF(pTauGam.gt.g1_tau_h)pTauGam=g1_tau_h
     endif
     
     if(pEcut.lt.g1_kT_l)pEcut=g1_kT_l
     if(pEcut.gt.g1_kT_h)pEcut=g1_kT_h
     IF(pkTbb.gt.0.)THEN
        if(pkTbb.lt.g1_kTbb_l)pkTbb=g1_kTbb_l
        if(pkTbb.gt.g1_kTbb_h)pkTbb=g1_kTbb_h
     ELSE
        if((-pkTbb).lt.g1_kTbb_l)pkTbb=-g1_kTbb_l
        if((-pkTbb).gt.g1_kTbb_h)pkTbb=-g1_kTbb_h
     ENDIF
     if(pIncl.lt.g1_incl_l)pIncl=g1_incl_l
     if(pIncl.gt.g1_incl_h)pIncl=g1_incl_h
  ENDIF

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hreflD


SUBROUTINE f_hreflDG(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_xilcps, only: pGeom,RMtype,pkTbb
  use par_xilcps, only: pEcut,pIncl
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_incl_l,g1_incl_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 9
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.  

  call init_par_hreflDG(param,num_param)

  mtp=6
  RMtype=1.0
  pkTbb=1.0e-3

  if(pGeom.eq.-1.0) pGeom=1.0
  if(abs(pGeom).ne.1.0) pGeom=0.0

  IF(pGeom.eq.1.0)THEN
     if(pEcut.lt.g1_kT_l)pEcut=g1_kT_l
     if(pEcut.gt.g1_kT_h)pEcut=g1_kT_h
     if(pIncl.lt.g1_incl_l)pIncl=g1_incl_l
     if(pIncl.gt.g1_incl_h)pIncl=g1_incl_h
  ENDIF

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hreflDG


SUBROUTINE f_hreflexp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  USE par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 7
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.
  
  mtp = 3
  RMtype = 1.0

  call init_par_hreflexp(param,num_param)

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hreflexp


SUBROUTINE f_hreflexpD(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  USE par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 8
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)

  ifl=0
  photer(1)=0.0

  is_lamppost = .false.
  is_lamppost_bb = .false.
  
  mtp = 7
  RMtype=1.0

  call init_par_hreflexpD(param,num_param)

  call main_local_model(ear, ne, photar)

  return
END SUBROUTINE f_hreflexpD


SUBROUTINE f_xsline(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,inen,ginen,inbn,inphotar_p,is_lamppost,is_lamppost_bb
  USE par_xilcps, only: pRedshft
  IMPLICIT NONE

  integer,parameter :: num_param = 2
  INTEGER :: ne, ifl, i
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)

  REAL:: redshft_scl
  double precision :: dp_photar(ne)
  REAL::g_inphotar(inbn)

  ifl=0
  photer(1)=0.0

  mtp=4
  is_lamppost = .false.
  is_lamppost_bb = .false.  

  call INIT_PAR_XLINE(param, num_param)

  call get_primary()

  IF(pRedshft.gt.0.0)THEN
     redshft_scl = 1.0/ (1 + pRedshft)
     call shift_photar(real(inen),real(inphotar_p),inbn,ginen,g_inphotar,redshft_scl)
     call f_rebin_spectrum(1.d0*ear,ne,1.d0*ginen,inbn,dp_photar,1.0d0*g_inphotar)
  ELSE
     call f_rebin_spectrum(1.d0*ear,ne,inen,inbn,dp_photar,1.0d0*inphotar_p)
  ENDIF

  do i=1,ne
     photar(i)=real(dp_photar(i))
  enddo

  return
END SUBROUTINE f_xsline


SUBROUTINE main_local_model(ear, ne, photar)
  USE par_global, only: inbn,inen,ginen,inphotar_p,inphotar_r,rchatty
  use par_xilcps, only: pRel_refl,pRedshft
  IMPLICIT NONE

  INTEGER :: ne, i, k
  REAL :: ear(0:ne), photar(ne)

  double precision, allocatable :: dp_infinal(:),dp_photar(:)
  REAL(kind=8), ALLOCATABLE :: g_inphotar(:)
  REAL, ALLOCATABLE :: g_inphotar_tmp(:)
  REAL :: redshft_scl,norm_1

  character FGMSTR*20, FGMSTR_val*20

  LOGICAL,SAVE :: message_diplayed=.FALSE.

  IF(.not.message_diplayed)THEN
     write(*,"(A73)")"*************************************************************************"
     write(*,"(A73)")"* Hybrid model (xillver + ireflect) for reflection from ionized medium. *"
     write(*,"(A73)")"*           If you use results of this model, please refer to           *"
     write(*,"(A73)")"*        Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942        *"
     write(*,"(A73)")"*************************************************************************"
     message_diplayed=.TRUE.
  ENDIF

  FGMSTR_val= trim(FGMSTR("RCHATTY"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) i
     if(i.eq.1)then
        rchatty=.TRUE.
     else
        rchatty=.FALSE.
     endif
  else
     rchatty=.FALSE.
  endif

  IF(allocated(dp_infinal))deallocate(dp_infinal)
  IF(allocated(dp_photar)) deallocate(dp_photar)
  IF(allocated(g_inphotar))deallocate(g_inphotar)
  IF(allocated(g_inphotar_tmp))deallocate(g_inphotar_tmp)

  allocate(dp_infinal(inbn))
  allocate(dp_photar(ne))
  allocate(g_inphotar(inbn))
  allocate(g_inphotar_tmp(inbn))

  call get_primary()

  call f_binary_search_float(1.d0,k,inen,inbn+1)

  IF(pRedshft.ne.0.0)THEN
     redshft_scl = abs(1.0/ (1.0 + pRedshft))
     call shift_photar(real(inen),inphotar_p,inbn,ginen,g_inphotar_tmp,redshft_scl)
     call f_rebin_spectrum(inen,inbn,1.d0*ginen,inbn,g_inphotar,1.0d0*g_inphotar_tmp)
     norm_1=real(g_inphotar(k)/(inen(k)-inen(k-1)))
  ELSE
     norm_1=real(inphotar_p(k)/(inen(k)-inen(k-1)))
  ENDIF

  IF(pRel_refl.ne.0.0) call get_reflection()

  do i=1,inbn
     dp_infinal(i)=1.d0*inphotar_r(i)
  enddo

  call get_finalsp(inbn, dp_infinal, 1.0d0*inphotar_p,pRel_refl)

  IF(pRedshft.ne.0.0)THEN
     call shift_photar(real(inen),real(dp_infinal),inbn,ginen,g_inphotar_tmp,redshft_scl)
     call f_rebin_spectrum(1.d0*ear,ne,1.d0*ginen,inbn,dp_photar,1.0d0*g_inphotar_tmp)
  ELSE
     call f_rebin_spectrum(1.d0*ear,ne,inen,inbn,dp_photar,dp_infinal)
  ENDIF

  do i=1,ne
     photar(i)=real(dp_photar(i))/norm_1
  enddo
  IF(rchatty)write(*,*)"Final norm=",1.0/norm_1

  deallocate(dp_infinal)
  deallocate(dp_photar) 
  deallocate(g_inphotar)  
  deallocate(g_inphotar_tmp)  

  return
END SUBROUTINE main_local_model


SUBROUTINE xsline(ear, ne, param, ifl, photar, photer)                 
  IMPLICIT NONE      

  INTEGER :: ne, ifl, i
  integer, parameter :: npm= 1                                                                                             
  REAL :: ear(0:ne), param(npm), photar(ne), photer(ne)

  ifl=0
  photer(1)=0.0

  do i=1,ne                                                                                                                
     if((ear(i-1).lt.param(1)).and.(ear(i).gt.param(1))) then                                                              
        photar(i) = 1.                                                                                                     
     else                                                                                                                  
        photar(i) = 0.                                                                                                     
     endif                                                                                                                 
  enddo                                                                                                                    

  return                                                                                                                   
END SUBROUTINE xsline 


SUBROUTINE init_inen()
  use par_global, only: inen,inenl,inenh,inbn,cached_inen
  DOUBLE PRECISION :: de

  IF(.not.cached_inen)THEN
     de=dlog10(inenh/inenl)/inbn
     do i=0,inbn
        inen(i) = inenl*10.0**(i*de)
     enddo
     cached_inen=.TRUE.
  ENDIF

  return
END SUBROUTINE init_inen


SUBROUTINE INIT_PAR_PRIMARY(param,npar)
  use par_global, only: mtp,pi,rchatty,rnobb
  use par_xilcps, only: pEcut,pIncl,pTauGam,pLgXi,pAFe,pGamma
  use par_xilcps, only: pkTe, pKTbb, pTauy, pGeom, pCosi
  use par_xilcps, only: pEind,pGmin,pGmax,pHR,pCovf,pAMe,pTdisk,pBet10,pRin,pRout
  use par_xilcps, only: pLineE

  IMPLICIT NONE
  INTEGER :: npar,i
  REAL :: param(npar)
  LOGICAL, SAVE :: firstcall= .TRUE.
  LOGICAL, SAVE :: TGT_loaded= .FALSE.
  LOGICAL, SAVE :: TGT_mu_loaded= .FALSE.

  INTEGER, PARAMETER :: GBn=220
  INTEGER, PARAMETER :: kTBn=990
  REAL, SAVE :: kTtGlim(0:kTBn,2),GTau(0:GBn,0:kTBn)

  INTEGER, PARAMETER :: cosiBn=10
  INTEGER, PARAMETER :: muGBn=220
  INTEGER, PARAMETER :: mukTBn=198
  REAL, SAVE :: mukTtGlim(0:cosiBn,0:mukTBn,2),muGTau(0:cosiBn,0:muGBn,0:mukTBn)

  REAL :: get_tau, get_tau_mu
  INTEGER :: in_mtp

  character(LEN=20)::FGMSTR,FGMSTR_val

  in_mtp=mtp
  if(mtp.eq.5)in_mtp=1
  if(mtp.eq.6)in_mtp=2
  if(mtp.eq.7)in_mtp=3
  if(mtp.eq.8)in_mtp=1
  if(mtp.eq.9)in_mtp=1
  if(mtp.eq.10)in_mtp=2

  IF(firstcall) THEN
     call check_table_paths()
     call FPMSTR("IREFLECT_PRECISION", "1.2")
!     call FPMSTR("IREFLECT_MAX_E", "5000.0")     
     call FPMSTR("COMPPS_PRECISION", "1.2")
     call FPMSTR("PEXRIV_PRECISION", "1.2")
     firstcall=.FALSE.
  ENDIF

  FGMSTR_val= trim(FGMSTR("RNOBB"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) i
     if(i.eq.1)then
        rnobb=.TRUE.
     else
        rnobb=.FALSE.
     endif
  else
     rnobb=.FALSE.
  endif

  IF(in_mtp.eq.1) THEN

     pTauy=pTauGam
     pkTe=pEcut
     pCosi=cos(pIncl*real(Pi)/180.0)
     param(1)=pkTe
     param(2)=pEind
     param(3)=pGmin
     param(4)=pGmax
     param(5)=pkTbb
     param(6)=pTauy
     param(7)=pGeom
     param(8)=pHR
     param(9)=pCosi
     param(10)=pCovf
     param(11)=0.0
     param(12)=pAFe
     param(13)=pAMe
     param(14)=10.0**pLgXi
     param(15)=pTdisk
     param(16)=pBet10
     param(17)=pRin    
     param(18)=pRout
     param(19)=0.0

  ELSE IF(in_mtp.eq.2) THEN

     pkTe=pEcut
     pCosi=cos(pIncl*real(Pi)/180.0)

     if(pGeom.eq.0.0) then
        IF(.not.TGT_loaded) THEN
           call INIT_TABLE_kTtG_PATHS()
           call INIT_kTtG_TABLES(kTtGlim,GTau,GBn,kTBn)
           write(*,*) "Gamma parametrization enabled for spherical geometry"
           TGT_loaded=.TRUE.
        ENDIF
        pTauy=get_tau(pTauGam,pkTe,kTtGlim,GTau,GBn,kTBn)
     else
        IF(.not.TGT_mu_loaded) THEN
           call INIT_TABLE_kTtG_PATHS()
           call INIT_mukTtG_TABLES(mukTtGlim,muGTau,cosiBn,muGBn,mukTBn)
           write(*,*) "Gamma parametrization enabled for slab geometry"
           TGT_mu_loaded=.TRUE.
        ENDIF
        pTauy=get_tau_mu(pTauGam,pCosi,pkTe,mukTtGlim,muGTau,cosiBn,muGBn,mukTBn)
     endif

     param(1)=pkTe
     param(2)=pEind
     param(3)=pGmin
     param(4)=pGmax
     param(5)=pkTbb
     param(6)=pTauy
     param(7)=pGeom
     param(8)=pHR
     param(9)=pCosi
     param(10)=pCovf
     param(11)=0.0
     param(12)=pAFe
     param(13)=pAMe
     param(14)=10.0**pLgXi
     param(15)=pTdisk
     param(16)=pBet10
     param(17)=pRin    
     param(18)=pRout
     param(19)=0.0

  ELSE IF(in_mtp.eq.3) THEN

     pCosi=cos(pIncl*real(Pi)/180.0)
     pGamma=pTauGam
     param(1)=pGamma
     param(2)=pEcut
     param(3)=0.0
     param(4)=0.0
     param(5)=pAMe
     param(6)=pAFe
     param(7)=pCosi
     param(8)=pTdisk
     param(9)=10.0**pLgXi

  ELSE IF(in_mtp.eq.4) THEN

     param(1)=pLineE

  ENDIF
    IF(rchatty)write(*,*) "Primary tau/y=",pTauy

  return
END SUBROUTINE INIT_PAR_PRIMARY


SUBROUTINE INIT_PAR_REFLECTION(param1,npar1,param2,npar2,param_iref)
  use par_global, only: mtp,pi,inphotar_p,inen,inbn,is_lamppost,gshft,rchatty
  use par_xilcps, only: pEcut,gEcut,pIncl,pTauGam,pLgXi,pXi,pAFe,pGamma,pLgN,flux2,flux2_30
  use par_xilcps, only: pkTe, pKTbb, pTauy, pGeom, pCosi
  use par_xilcps, only: pEind,pGmin,pGmax,pHR,pCovf,pAMe,pTdisk,pBet10,pRin,pRout
  use par_xilcps, only: cached_firrslb_tbp

  IMPLICIT NONE
  INTEGER :: npar1, npar2,i
  REAL :: param1(npar1),param2(npar2)
  REAL :: param_iref(7)
!  REAL, PARAMETER :: eming=1.0
!  REAL, PARAMETER :: emaxg=20.0
  REAL::eming,emaxg
  REAL :: dum,flscl,Gamma_o,get_xEcut,get_mkTe
  INTEGER :: in_mtp

  LOGICAL, SAVE :: irrslb_tab_loaded= .FALSE.
  INTEGER, PARAMETER :: kTBn = 40                                                        
  INTEGER, PARAMETER :: kTbbBn = 10
  INTEGER, PARAMETER :: TBn = 80
  INTEGER, PARAMETER :: cosiBn = 6

  REAL, PARAMETER :: escale_irr=20.0

  REAL, SAVE :: slope(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn),fluxscl(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn)

  in_mtp=mtp
  if(mtp.eq.5)in_mtp=1
  if(mtp.eq.6)in_mtp=2
  if(mtp.eq.7)in_mtp=3
  if(mtp.eq.8)in_mtp=1
  if(mtp.eq.9)in_mtp=1
  if(mtp.eq.10)in_mtp=2

  IF(in_mtp.eq.1) THEN

     eming=1.0
     emaxg=20.0
     eming=eming*(5.8997*abs(pkTbb)+0.9941) !shifting due to low e. cutoff and GR rshft
     if(eming.gt.10.0)eming=10.0
     eming=abs(gshft)*eming
     emaxg=abs(gshft)*emaxg

     gEcut=pEcut
     pTauy=pTauGam
     pkTe=pEcut

     if(pGeom.eq.1)then
        IF(.not.cached_firrslb_tbp)THEN
           call INIT_TABLE_IRRSLB_PATH()
        ENDIF
        IF(.not.irrslb_tab_loaded) THEN
           call INIT_IRRSLB_TABLES(slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn)
           irrslb_tab_loaded=.TRUE.
        ENDIF
        call get_slope(eming,emaxg,inphotar_p,real(inen),inbn,Gamma_o,flux2)
        IF(Gamma_o.lt.1.2)THEN
           Gamma_o=1.2
           flux2=1.0
        ENDIF
        IF(Gamma_o.gt.3.4)THEN
           Gamma_o=3.4
           flux2=1.0
        ENDIF
        
        call get_slope_irrslb(pGamma,flscl,slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn,pkTe,abs(pkTbb),pTauGam,pIncl)
        IF(pGamma.lt.1.2)THEN
           pGamma=1.2
           fluxscl=1.0
        ENDIF
        IF(pGamma.gt.3.4)THEN
           pGamma=3.4
           fluxscl=1.0
        ENDIF
        
        flscl=flscl*escale_irr**(pGamma-Gamma_o) !factor calculated on 20 keV!
        if(flscl.le.0.0)flscl=1.0
        
        IF(rchatty)write(*,*)"*************Diagnose info*************"
        IF(rchatty)write(*,*)"Norm ps (from pl fit)=",1.0/flux2,"Norm ps rescaled=",1.0/(flux2*flscl),"scl_irr=",flscl,&
             "gam_irr=",pGamma,"gam_o=",Gamma_o
        flux2=flscl*flux2
     else
        call get_slope(eming,emaxg,inphotar_p,real(inen),inbn,pGamma,flux2)
        IF(pGamma.lt.1.2)THEN
           pGamma=1.2
           flux2=1.0
        ENDIF
        IF(pGamma.gt.3.4)THEN
           pGamma=3.4
           flux2=1.0
        ENDIF

        IF(rchatty)write(*,*)"*************Diagnose info*************"
        IF(rchatty)write(*,*)"Norm ps (from pl fit)=",1.0/flux2,"gam_o=",pGamma
     endif

     IF(mtp.eq.9) THEN
        gEcut=get_xEcut(pGamma,pkTe)
        if(is_lamppost)then 
           gEcut=gshft*gEcut
           IF(rchatty)write(*,*)"Use Ec, xEc=",gEcut,"g=",gshft
        else
           IF(rchatty)write(*,*)"Use Ec, xEc=",gEcut
        endif
     ELSE IF(mtp.ne.5) THEN
        if(is_lamppost)then
           dum=0.435*gshft+0.565 !dum-need additional factor
           gEcut=get_mkTe(pGamma,dum*gshft*pkTe)
           IF(rchatty)write(*,*)"Use Cp,mkT=",get_mkTe(pGamma,pkTe),"gmkT=",gEcut,"g=",gshft,"g_dum=",dum
        else
           gEcut=get_mkTe(pGamma,pkTe)
           IF(rchatty)write(*,*)"Use Cp, mkT=",gEcut
        endif
        flux2_30=1.0
        if(pGeom.ne.1)then
           call f_binary_search_float(30.0d0,i,inen,inbn+1)
           flux2_30=real(inphotar_p(i)/(inen(i)-inen(i-1)))
        endif
     ELSE
        IF(rchatty)write(*,*)"Use D"
     ENDIF

     param1(1)=pGamma
     IF(param1(1).lt.1.2)param1(1)=1.2
     IF(param1(1).gt.3.4)param1(1)=3.4

     param1(2)=  pAFe
     param1(3)=  gEcut
     IF(mtp.eq.5)param1(3)=pLgN
     param1(4)=  pLgXi
     param1(5)=  0.0
     param1(6)=  pIncl  
     param1(7)=  -1.0

     pCosi=cos(pIncl*real(Pi)/180.0)
     param2(1)=pkTe
     param2(2)=pEind
     param2(3)=pGmin
     param2(4)=pGmax
     param2(5)=pkTbb
     param2(6)=pTauy
     param2(7)=pGeom
     param2(8)=pHR
     param2(9)=pCosi
     param2(10)=pCovf
     param2(11)=-1.0
     param2(12)=pAFe
     param2(13)=pAMe
     IF(mtp.ne.8)pXi=10.0**pLgXi
     param2(14)=pXi
     param2(15)=pTdisk
     param2(16)=pBet10
     param2(17)=pRin    
     param2(18)=pRout
     param2(19)=0.0

     param_iref(1)=   -1.0
     param_iref(2)=   0.0
     param_iref(3)=   pAMe
     param_iref(4)=   pAFe
     param_iref(5)=   pCosi
     param_iref(6)=   pTdisk
     param_iref(7)=   pXi

  ELSE IF(in_mtp.eq.2) THEN

     eming=1.0
     emaxg=20.0
     eming=eming*(5.8997*abs(pkTbb)+0.9941) !shifting due to low e. cutoff and GR rshft
     if(eming.gt.10.0)eming=10.0
     eming=abs(gshft)*eming
     emaxg=abs(gshft)*emaxg

     pkTe=pEcut
     pCosi=cos(pIncl*real(Pi)/180.0)

     if(pGeom.eq.1)then
        IF(.not.cached_firrslb_tbp)THEN
           call INIT_TABLE_IRRSLB_PATH()
        ENDIF
        IF(.not.irrslb_tab_loaded) THEN
           call INIT_IRRSLB_TABLES(slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn)
           irrslb_tab_loaded=.TRUE.
        ENDIF
        call get_slope(eming,emaxg,inphotar_p,real(inen),inbn,Gamma_o,flux2)
        IF(Gamma_o.lt.1.2)THEN
           Gamma_o=1.2
           flux2=1.0
        ENDIF
        IF(Gamma_o.gt.3.4)THEN
           Gamma_o=3.4
           flux2=1.0
        ENDIF

        call get_slope_irrslb(pGamma,flscl,slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn,pkTe,abs(pkTbb),pTauy,pIncl)
        IF(pGamma.lt.1.2)THEN
           pGamma=1.2
           fluxscl=1.0
        ENDIF
        IF(pGamma.gt.3.4)THEN
           pGamma=3.4
           fluxscl=1.0
        ENDIF

        flscl=flscl*escale_irr**(pGamma-Gamma_o) !factor calculated on 20 keV!

        if(flscl.le.0.0)flscl=1.0
        
        IF(rchatty)write(*,*)"*************Diagnose info*************"
        IF(rchatty) write(*,*)"Norm ps (from pl fit)=",1.0/flux2,"Norm ps rescaled=",1.0/(flux2*flscl),"scl_irr=",flscl,&
             "gam_irr=",pGamma,"gam_o=",Gamma_o
        flux2=flscl*flux2
     else
        call get_slope(eming,emaxg,inphotar_p,real(inen),inbn,pGamma,flux2)
        IF(pGamma.lt.1.2)THEN
           pGamma=1.2
           flux2=1.0
        ENDIF
        IF(pGamma.gt.3.4)THEN
           pGamma=3.4
           flux2=1.0
        ENDIF

        IF(rchatty)write(*,*)"*************Diagnose info*************"
        IF(rchatty) write(*,*)"Norm ps (from pl fit)=",1.0/flux2,"gam_o=",pGamma
     endif

     gEcut=pEcut

     IF(mtp.eq.10) THEN
        gEcut=get_xEcut(pGamma,pkTe)
        if(is_lamppost)then 
           gEcut=gshft*gEcut
           IF(rchatty)write(*,*)"Use Ec, xEc=",gEcut,"g=",gshft
        else
           IF(rchatty)write(*,*)"Use Ec, xEc=",gEcut
        endif
     ELSE IF(mtp.ne.6) THEN
        if(is_lamppost)then
           dum=0.435*gshft+0.565 !dum-need additional factor
           gEcut=get_mkTe(pGamma,dum*gshft*pkTe)
           IF(rchatty)write(*,*)"Use Cp, mkT=",gEcut,"g=",gshft,"g_dum=",dum
        else
           gEcut=get_mkTe(pGamma,pkTe)
           IF(rchatty)write(*,*)"Use Cp, mkT=",gEcut
        endif
        flux2_30=1.0
        if(pGeom.ne.1)then
           call f_binary_search_float(30.0d0,i,inen,inbn+1)
           flux2_30=real(inphotar_p(i)/(inen(i)-inen(i-1)))
        endif
     ELSE
        IF(rchatty)write(*,*)"Use D"
     ENDIF

     pXi=10.0**pLgXi

     param1(1)=pGamma
     IF(param1(1).lt.1.2)param1(1)=1.2
     IF(param1(1).gt.3.4)param1(1)=3.4

     param1(2)= pAFe
     param1(3)= gEcut
     IF(mtp.eq.6)param1(3)=pLgN
     param1(4)=  pLgXi
     param1(5)=  0.0
     param1(6)=  pIncl  
     param1(7)=  -1.0

     param2(1)=pkTe
     param2(2)=pEind
     param2(3)=pGmin
     param2(4)=pGmax
     param2(5)=pkTbb
     param2(6)=pTauy
     param2(7)=pGeom
     param2(8)=pHR
     param2(9)=pCosi
     param2(10)=pCovf
     param2(11)=-1.0
     param2(12)=pAFe
     param2(13)=pAMe
     param2(14)=pXi
     param2(15)=pTdisk
     param2(16)=pBet10
     param2(17)=pRin    
     param2(18)=pRout
     param2(19)=0.0

     param_iref(1)=   -1.0
     param_iref(2)=   0.0
     param_iref(3)=   pAMe
     param_iref(4)=   pAFe
     param_iref(5)=   pCosi
     param_iref(6)=   pTdisk
     param_iref(7)=   pXi

  ELSE IF(in_mtp.eq.3) THEN

     pXi=10.0**pLgXi

     call f_binary_search_float(1.d0,i,inen,inbn+1)
     flux2=real(inphotar_p(i)/(inen(i)-inen(i-1)))

     gEcut=pEcut
     IF(is_lamppost) gEcut=gshft*pEcut
     
     pGamma=pTauGam
     param1(1)=pGamma
     IF(param1(1).lt.1.2)param1(1)=1.2
     IF(param1(1).gt.3.4)param1(1)=3.4

     param1(2)=  pAFe
     param1(3)=  gEcut
     IF(mtp.eq.7)param1(3)=pLgN
     param1(4)=  pLgXi
     param1(5)=  0.0
     param1(6)=  pIncl  
     param1(7)=  -1.0

     pCosi=cos(pIncl*real(Pi)/180.0)
     param2(1)=pGamma
     param2(2)=pEcut
     param2(3)=-1.0
     param2(4)=0.0
     param2(5)=pAMe
     param2(6)=pAFe
     param2(7)=pCosi
     param2(8)=pTdisk
     param2(9)=pXi

     param_iref(1)=   -1.0
     param_iref(2)=   0.0
     param_iref(3)=   pAMe
     param_iref(4)=   pAFe
     param_iref(5)=   pCosi
     param_iref(6)=   pTdisk
     param_iref(7)=   pXi

  ENDIF

  IF(rchatty) write(*,*) "Primary gamma=",pGamma

  return
END SUBROUTINE INIT_PAR_REFLECTION


SUBROUTINE INIT_PAR_HREFL(param, num_param)

  use par_xilcps, only: pRedshft,pEcut,pIncl,pTauGam,pRel_refl,pLgXi,pAFe,pKTbb,pGeom,puse_Ec

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =     param(3)
  pLgXi =     param(4)
  pRedshft =  param(5)
  pIncl =     param(6)
  pKTbb =     param(7)
  pGeom =     param(8)
  puse_Ec =   param(9)
  pRel_refl = param(10)

  RETURN
END SUBROUTINE INIT_PAR_HREFL


SUBROUTINE INIT_PAR_HREFLG(param, num_param)

  use par_xilcps, only: pRedshft,pEcut,pIncl,pTauGam,pRel_refl,pLgXi,pAFe,pGeom,puse_Ec

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =     param(3)
  pLgXi =     param(4)
  pRedshft =  param(5)
  pIncl =     param(6)
  pGeom =     param(7)
  puse_Ec =   param(8)
  pRel_refl = param(9)

  RETURN
END SUBROUTINE INIT_PAR_HREFLG


SUBROUTINE INIT_PAR_HREFLD(param, num_param)

  use par_xilcps, only: pRedshft,pLgN,pIncl,pTauGam,pEcut,pRel_refl,pLgXi,pAFe,pKTbb,pGeom

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =       param(3)
  pLgXi =       param(4)
  pRedshft =      param(5)
  pIncl =     param(6)
  pLgN =     param(7)
  pKTbb =     param(8)
  pGeom =     param(9)
  pRel_refl = param(10)

  RETURN
END SUBROUTINE INIT_PAR_HREFLD


SUBROUTINE INIT_PAR_HREFLDG(param, num_param)

  use par_xilcps, only: pRedshft,pLgN,pIncl,pTauGam,pEcut,pRel_refl,pLgXi,pAFe,pGeom

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =       param(3)
  pLgXi =       param(4)
  pRedshft =      param(5)
  pIncl =     param(6)
  pLgN =     param(7)
  pGeom =     param(8)
  pRel_refl = param(9)

  RETURN
END SUBROUTINE INIT_PAR_HREFLDG


SUBROUTINE INIT_PAR_HREFLEXP(param, num_param)

  use par_xilcps, only: pRedshft,pEcut,pIncl,pTauGam,pRel_refl,pLgXi,pAFe

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =       param(3)
  pLgXi =      param(4)
  pRedshft =      param(5)
  pIncl =     param(6)
  pRel_refl = param(7)

END SUBROUTINE INIT_PAR_HREFLEXP


SUBROUTINE INIT_PAR_HREFLEXPD(param, num_param)

  use par_xilcps, only: pRedshft,pLgN,pEcut,pIncl,pTauGam,pRel_refl,pLgXi,pAFe

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pTauGam =   param(1)
  pAFe =      param(2)
  pEcut =       param(3)
  pLgXi =    param(4)
  pRedshft =      param(5)
  pIncl =     param(6)
  pLgN=      param(7)
  pRel_refl = param(8)

  RETURN
END SUBROUTINE INIT_PAR_HREFLEXPD


SUBROUTINE INIT_PAR_XLINE(param, num_param)

  use par_xilcps, only: pLineE,pRedshft

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  pLineE =   param(1)
  pRedshft =   param(2)
  
  RETURN
END SUBROUTINE INIT_PAR_XLINE

SUBROUTINE merge(photar1, photar2, ear, ne, flux1, flux2, photar, cross)
  use par_global, only: pi,rchatty
  use par_xilcps, only: pIncl,use_iref
  IMPLICIT NONE

  INTEGER :: ne
  REAL :: ear(0:ne), photar1(ne), photar2(ne), flux1, flux2
  double precision :: photar(ne)
  LOGICAL :: cross

  REAL, PARAMETER :: macc=5.0e-3
  REAL :: dum, scale, scalef, emin, emax, ec
  REAL,ALLOCATABLE ::slp_mrg(:)
  REAL :: slp_tmp

  INTEGER :: i, bmin, bmax,bmin_hlf,bmax_hlf,ic,ic2,p12i
  double precision :: p1w, p2w
  REAL :: dfl,dfh,slp_mrg_l, slp_mrg_r, phot_l, phot_r
  REAL, ALLOCATABLE :: diff_phot(:), photar1_in(:)
  INTEGER, ALLOCATABLE :: iarr(:)
  INTEGER :: sgn_l,sgn_r

  REAL, PARAMETER :: slp_mrg_scale=2.5

  character(len=20) :: FGSOLR
  character(len=10) :: FGSOLR_val

  FGSOLR_val = trim(FGSOLR())

  IF(FGSOLR_val.eq.'angr')THEN
     scale=real(flux2/flux1*(1.0187+4.66693*sin(pi/180.0*pIncl*0.228909)**2.48271+ &
          5.93573*sin(pi/180.0*pIncl*0.727118)**23.6696))
  ELSE IF(FGSOLR_val.eq.'wilm')THEN
     scale=real(flux2/flux1*(1.20486+3.84012*sin(pi/180.0*pIncl*0.206153)**2.19502+ &
          20.3397*sin(pi/180.0*pIncl*0.702119)**26.8753))
  ELSE
     scale=real(flux2/flux1) !to turn-off angular scaling
  ENDIF
  
!  scale=real(flux2/flux1)!TEMPORARY OFF ANGULAR SCALING
  
  IF(.not.use_iref)THEN
     do i=1,ne
        photar(i)=photar1(i)*scale !no merging                                                               
     enddo
     cross=.TRUE. !no merging                                                                                
     GOTO 10 !no merging                                                                                     
  ELSE
     IF(allocated(photar1_in))deallocate(photar1_in)
     allocate(photar1_in(ne))
     do i=1,ne
        photar1_in(i)=photar1(i)*scale
     enddo
  ENDIF

  emin=10.0
  emax=100.0

  call f_binary_search_float(1.d0*emin,bmin,1.d0*ear,ne+1)
  call f_binary_search_float(1.d0*emax,bmax,1.d0*ear,ne+1)

  ic=-1
  p12i=int(sign(1.0,photar1_in(bmin)-photar2(bmin)))
  do i=bmin+1,bmax
     if( p12i.ne.sign(1.0,photar1_in(i)-photar2(i)) ) then
        ic=i
        exit
     endif
  enddo

   IF (ic.gt.0) THEN
     ec=sqrt(ear(ic)*ear(ic-1)) 

     emin=MAX(emin,0.5*ec)
     emax=2.0*ec

     IF(rchatty) write(*,*)"Cross: e_l,e_cross,e_h:", emin,ec,emax
     IF(rchatty) write(*,*)"Prim_norm_x=",1.0/flux1,"Scl_fac_ang=",scale*flux1/flux2
     cross=.TRUE.
  ELSE
     if(allocated(diff_phot))deallocate(diff_phot)
     if(allocated(iarr))deallocate(iarr)
     allocate(diff_phot(bmax-bmin))
     allocate(iarr(bmax-bmin))

     if(photar2(bmin+10).gt.photar1_in(bmin+10))then
        do i=1,bmax-bmin
           diff_phot(i)=photar2(bmin+i)/photar1_in(bmin+i)
        enddo
     else
        do i=1,bmax-bmin
           diff_phot(i)=photar1_in(bmin+i)/photar2(bmin+i)
        enddo
     endif

     call piksrt(bmax-bmin,diff_phot,iarr)

     ic=iarr(1)+bmin

     ec=sqrt(ear(ic)*ear(ic-1)) 

     scalef=photar2(ic)/photar1_in(ic)

     do i=1,ne
        photar1_in(i)=photar1_in(i)*scalef
     enddo

     emin=MAX(emin,0.5*ec)
     emax=2.0*ec

     deallocate(diff_phot)
     deallocate(iarr)

     IF(rchatty) write(*,*)"NO CROSS: e_l,e_nearest,e_h:", emin,ec,emax
     IF(rchatty) write(*,*)"Prim_norm_x=",1.0/flux1,"Scal_fac_nocross=",scalef,&
          "Scal_fac_angle=",scale*flux1/flux2
     cross=.FALSE.
  ENDIF

     call f_binary_search_float(1.d0*emin,bmin,1.d0*ear,ne+1)
     call f_binary_search_float(1.d0*emax,bmax,1.d0*ear,ne+1)

     bmin_hlf=(ic+bmin)/2
     bmax_hlf=(ic+bmax)/2

     ic2=-1;
     p12i=int(sign(1.0,photar1_in(bmax_hlf)-photar2(bmax_hlf)))
     IF(p12i.gt.0)THEN
        dum=photar1_in(bmax_hlf)/photar2(bmax_hlf)
        DO i=bmax_hlf+1,bmax
           if(photar1_in(i)/photar2(i).le.(3.0*dum+1.0)/4.0)then
              ic2=i
              exit
           endif
        ENDDO
     ELSE
        dum=photar2(bmax_hlf)/photar1_in(bmax_hlf)
        DO i=bmax_hlf+1,bmax
           if(photar2(i)/photar1_in(i).le.(3.0*dum+1.0)/4.0)then
              ic2=i
              exit
           endif
        ENDDO
     ENDIF

     IF(ic2.gt.0)THEN
        bmax=ic2;
        emax=sqrt(ear(bmax)*ear(bmax-1)) 
        IF(rchatty) write(*,*) "RIGHT FADE FOUND AT:",emax
     ENDIF

     do i=1,bmin-1
        photar(i)=REAL(photar1_in(i),8)
     enddo
     do i=bmax+1,ne
        photar(i)=REAL(photar2(i),8)
     enddo

     dum=photar1_in(bmin)

     dfl=(photar1_in(bmin_hlf)/photar2(bmin_hlf))
     dfh=(photar2(bmax_hlf)/photar1_in(bmax_hlf))

     sgn_l=1
     sgn_r=1

     IF((photar1_in(bmin_hlf).lt.photar2(bmin_hlf))) sgn_l=-1
     IF((photar1_in(bmax_hlf).gt.photar2(bmax_hlf))) sgn_r=-1

     if(dfl.lt.1.0)then
        dfl=1.0/dfl
     endif
     if(dfh.lt.1.0)then 
        dfh=1.0/dfh
     endif

     slp_mrg_l=slp_mrg_scale*log(2.0)/log(dfl)
     slp_mrg_r=slp_mrg_scale*log(2.0)/log(dfh)
     if(slp_mrg_l.gt.42.0)slp_mrg_l=42.0
     if(slp_mrg_l.lt.3.8)slp_mrg_l=3.8
     if(slp_mrg_r.gt.42.0)slp_mrg_r=42.0
     if(slp_mrg_r.lt.3.8)slp_mrg_r=3.8

     if(allocated(slp_mrg))deallocate(slp_mrg)
     allocate(slp_mrg(bmax-bmin+1))

     do i=bmin,bmax
        slp_mrg(i-bmin+1)=(ear(bmax)-ear(i))/(ear(bmax)-ear(bmin))*slp_mrg_l+&
             (ear(i)-ear(bmin))/(ear(bmax)-ear(bmin))*slp_mrg_r
     enddo
     
     IF((sgn_l*sgn_r).lt.0.0)THEN
        do i=bmin,bmin_hlf-1
           slp_tmp=sgn_l*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           photar(i)=((p1w+p2w)**(1.0/slp_tmp))*dum
        enddo

        do i=bmin_hlf,bmax_hlf
           slp_tmp=sgn_l*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           phot_l=REAL(((p1w+p2w)**(1.0/slp_tmp))*dum)

           slp_tmp=sgn_r*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           phot_r=REAL(((p1w+p2w)**(1.0/slp_tmp))*dum)

           p1w=(ear(bmax_hlf)-ear(i))/(ear(bmax_hlf)-ear(bmin_hlf))*phot_l
           p2w=(ear(i)-ear(bmin_hlf))/(ear(bmax_hlf)-ear(bmin_hlf))*phot_r
           photar(i)=p1w+p2w
        enddo

        do i=bmax_hlf+1,bmax
           slp_tmp=sgn_r*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           photar(i)=((p1w+p2w)**(1.0/slp_tmp))*dum
        enddo

     ELSE

        do i=bmin,ic
           slp_tmp=sgn_l*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           photar(i)=((p1w+p2w)**(1.0/slp_tmp))*dum
        enddo
        do i=ic+1,bmax
           slp_tmp=sgn_r*slp_mrg(i-bmin+1)
           p1w=REAL(photar1_in(i)/dum,8)**slp_tmp
           p2w=REAL(photar2(i)/dum,8)**slp_tmp
           photar(i)=((p1w+p2w)**(1.0/slp_tmp))*dum
        enddo

     ENDIF

     if(abs((photar1_in(bmin)-REAL(photar(bmin)))/photar1_in(bmin)).gt.macc)then
        dum=1.2*emin
        i=bmin
        do while(ear(i).lt.dum)
           p1w=(dum-ear(i))/(dum-emin)*photar1_in(i)
           p2w=(ear(i)-emin)/(dum-emin)*photar(i)
           photar(i)=p1w+p2w
           i=i+1
        enddo
     endif

     if(abs((photar2(bmax)-REAL(photar(bmax)))/photar2(bmax)).gt.macc)then
        dum=1.0/1.2*emax
        i=bmax
        do while(ear(i).gt.dum)
           p1w=(emax-ear(i))/(emax-dum)*photar(i)
           p2w=(ear(i)-dum)/(emax-dum)*photar2(i)
           photar(i)=p1w+p2w
           i=i-1
        enddo
     endif

     deallocate(slp_mrg)
     deallocate(photar1_in)

10 CONTINUE !no merging
     
  return
END SUBROUTINE merge


subroutine shift_photar(en_in,photar_in,ne,en_out,photar_out,redshft_scl)

  implicit none
  integer :: i, ne
  REAL :: en_in(0:ne),photar_in(ne)
  REAL :: en_out(0:ne),photar_out(ne)
  REAL :: redshft_scl

  do i=0,ne
     en_out(i) = en_in(i)*redshft_scl
  enddo

  do i=1,ne
     photar_out(i) = photar_in(i)
  enddo

  return
end subroutine shift_photar


SUBROUTINE get_finalsp(ne, photar, photarp,refl_frac)

  IMPLICIT NONE

  INTEGER :: ne, i
  REAL :: refl_frac
  double precision :: photar(ne), photarp(ne)

  if(abs(refl_frac).gt.1.0d-22) then
     do i=1, ne
        photar(i)=abs(refl_frac)*photar(i)
     enddo
     if(refl_frac.gt.0.0) then
        do i=1,ne
           photar(i)=photar(i)+photarp(i)
        enddo
     endif
  else
     do i=1,ne
        photar(i)=photarp(i)
     enddo
  endif

  return
END SUBROUTINE get_finalsp


SUBROUTINE get_slope(emin, emax, photar,ear, ne, gamma, flux)
  IMPLICIT NONE

  REAL :: gamma, flux
  REAL :: emin, emax
  INTEGER :: ne, mwt, fne
  REAL :: ear(0:ne), photar(ne)
  REAL :: sig(ne), SIGA, SIGB, CHI2, Q
  INTEGER :: bmin, bmax

  integer :: i

  real, allocatable :: fen(:)                                                                            
  real, allocatable :: fphotar(:)


  mwt=0 !no significances                                                                                                                                    
!  mwt=1                                                                                                                                                     
  gamma=0.0
  flux=0.0

  call f_binary_search_float(1.d0*emin,bmin,1.d0*ear,ne+1)
  call f_binary_search_float(1.d0*emax,bmax,1.d0*ear,ne+1)

  fne=bmax-bmin+1

  if (allocated(fen)) then
     deallocate (fen)
  endif

  if (allocated(fphotar)) then
     deallocate (fphotar)
  endif
  allocate (fen(fne))
  allocate (fphotar(fne))

  do i=1,fne
     fen(i) = 0.5*log10(ear(bmin-2+i)*ear(bmin-1+i))
     fphotar(i) = log10(photar(bmin-1+i)/(ear(bmin-1+i)-ear(bmin-2+i)))
     sig(i)=1.0
!     sig(i)=sqrt(photar(bmin-1+i)/(ear(bmin-1+i)-ear(bmin-2+i))) !significances??
  enddo

  call FIT_LIN(fen, fphotar, fne, sig, mwt, flux, gamma, SIGA, SIGB, CHI2, Q)
  gamma=-gamma
  flux=10.0**flux

  deallocate (fen)
  deallocate (fphotar)

  return
END SUBROUTINE get_slope

!------------Function get_phot_flux temporaily not used------------
!   REAL function get_phot_flux(photar_in,en_in,ne,emin,emax)
!  implicit none
!  INTEGER :: ne, j
!  double precision :: photar_in(ne),en_in(0:ne)
!  double precision :: emin,emax
!
!  INTEGER :: bmin,bmax
!  double precision :: dmin,dmax
!
!  get_phot_flux=0.0
!
!  IF((emin.lt.en_in(0)).or.(emax.gt.en_in(ne)))THEN
!     write(*,*)"Flux calc impossible - wrong boundaries. Exiting..."
!     return
!  ELSE
!     call f_binary_search_float(emin,bmin,en_in,ne+1)
!     call f_binary_search_float(emax,bmax,en_in,ne+1)
!     bmin = bmin-1
!     bmax = bmax-1
!
!     if (bmin.eq.bmax) then
!        get_phot_flux = REAL((emax-emin)/(en_in(bmin+1)-en_in(bmin))*photar_in(bmin+1))
!     else
!        dmin=(en_in(bmin+1)-emin)/(en_in(bmin+1)-en_in(bmin))
!        dmax=(emax-en_in(bmax))/(en_in(bmax+1)-en_in(bmax))
!
!        get_phot_flux=REAL(get_phot_flux+photar_in(bmin+1)*dmin)
!        do j=bmin+2,bmax
!           get_phot_flux=REAL(get_phot_flux+photar_in(j))
!        enddo
!        get_phot_flux=REAL(get_phot_flux+photar_in(bmax+1)*dmax)
!     endif
!  ENDIF
!
!  return
!
!end function get_phot_flux
!------------Function get_phot_flux temporaily not used------------

REAL function get_energy_flux(photar_in,en_in,ne,emin,emax)

  implicit none
  INTEGER :: ne, j
  double precision :: photar_in(ne),en_in(0:ne)
  double precision :: emin,emax

  INTEGER :: bmin,bmax
  double precision :: dmin,dmax

  get_energy_flux=0.0

  IF((emin.lt.en_in(0)).or.(emax.gt.en_in(ne)))THEN
     write(*,*)"Flux calc impossible - wrong boundaries. Exiting..."
     return
  ELSE
     call f_binary_search_float(emin,bmin,en_in,ne+1)
     call f_binary_search_float(emax,bmax,en_in,ne+1)
     bmin = bmin-1
     bmax = bmax-1

     if (bmin.eq.bmax) then
        get_energy_flux = REAL((emax-emin)/(en_in(bmin+1)-en_in(bmin))*photar_in(bmin+1) &
             *sqrt(en_in(bmin+1)*en_in(bmin)))
     else
        dmin=(en_in(bmin+1)-emin)/(en_in(bmin+1)-en_in(bmin))
        dmax=(emax-en_in(bmax))/(en_in(bmax+1)-en_in(bmax))

        get_energy_flux=REAL(get_energy_flux+photar_in(bmin+1)*dmin*sqrt(en_in(bmin+1)*en_in(bmin)))
        do j=bmin+2,bmax
           get_energy_flux=REAL(get_energy_flux+photar_in(j)*sqrt(en_in(j+1)*en_in(j)))
        enddo
        get_energy_flux=REAL(get_energy_flux+photar_in(bmax+1)*dmax*sqrt(en_in(bmax+1)*en_in(bmax)))
     endif
  ENDIF

  return
end function get_energy_flux


SUBROUTINE FIT_LIN(X,Y,NDATA,SIG,MWT,A,B,SIGA,SIGB,CHI2,Q)
  DIMENSION X(NDATA),Y(NDATA),SIG(NDATA)
  
  SX  = 0.
  SY  = 0.
  ST2 = 0.

  IF (MWT .NE. 0) THEN
     SS = 0.
     DO 11, I=1,NDATA 
        WT = 1./(SIG(I)**2)
        SS = SS + WT
        SX = SX + X(I)*WT
        SY = SY + Y(I)*WT
11      CONTINUE
     ELSE             
        DO 12, I=1,NDATA
           SX = SX + X(I)
           SY = SY + Y(I)
12         CONTINUE
           SS = FLOAT(NDATA)
        ENDIF
        !                                                                                                                 
        SXOSS = SX/SS
        !                                                                                                                 
        IF (MWT .NE. 0) THEN
           DO 13,I=1,NDATA
              T   = (X(I)-SXOSS) / SIG(I)
              ST2 = ST2 + T*T
              B   = B + T*Y(I)/SIG(I)
13            CONTINUE
           ELSE
              DO 14,I=1,NDATA
                 T   = X(I) - SXOSS
                 ST2 = ST2 + T*T
                 B   = B + T*Y(I)
14               CONTINUE
              ENDIF

              B = B/ST2
              A = (SY - SX*B) / SS
              SIGA = SQRT((1. + SX*SY / (SS*ST2)) / SS)
              SIGB = SQRT(1. / ST2)
              CHI2 = 0.

              IF (MWT .EQ. 0) THEN
                 DO 15, I=1,NDATA
                    CHI2 = CHI2 + (Y(I) - A - B*X(I))**2
15                  CONTINUE

                    Q = 1.
                    SIGDAT = SQRT(CHI2 / (NDATA-2))
                    SIGA   = SIGA*SIGDAT           
                    SIGB   = SIGB*SIGDAT           
                 ELSE
                    DO 16, I=1,NDATA
                       CHI2 = CHI2 + ((Y(I) - A - B*X(I)) / SIG(I))**2
16                     CONTINUE
                       Q = GAMMQ(0.5 * (NDATA-2), 0.5*CHI2)
   ENDIF

 RETURN
END SUBROUTINE FIT_LIN
                  

FUNCTION GAMMQ(A,X)

  !  Returns the incomplete gamma-function P(a,x) -> 1-P(a,x).           !                                                   
  !  Press, W.H., et al., 1989. Numerical Recipes: p162.                 !                                                   


  IF (X .LT. 0. .OR. A .LE. 0.) PAUSE
  IF (X .LT.         A+1.) THEN
     CALL GSER(GAMSER,A,X,GLN)
     GAMMQ = 1. - GAMSER       
  ELSE                         
     CALL GCF(GAMMCF,A,X,GLN)
     GAMMQ = GAMMCF
  ENDIF
  !                                                                                                                          
  RETURN
END FUNCTION GAMMQ

SUBROUTINE GSER(GAMSER,A,X,GLN)

  !  Returns the incomplete gamma function P(a,x) evaluated by its se-   !                                                   
  !  ries representation as GAMSER. Also returns ln(..a) as GLN.         !                                                   
  !  Press, W.H., et al., 1989. Numerical Recipes: p162.                 !                                                   

  PARAMETER (ITMAX=100, EPS=3.E-7)
  !                                                                                                                          
  GLN = GAMMLN(A)
  IF (X .LE. 0.) THEN
     IF (X .LT. 0.) PAUSE
     GAMSER = 0.
     !                                                                                                                          
     RETURN
  ENDIF
  AP  = A
  SUM = 1./A
  DEL = SUM
  DO 11, N=1,ITMAX
     AP  = AP + 1.
     DEL = DEL*X/AP
     SUM = SUM+DEL
     IF (ABS(DEL) .LT. ABS(SUM)*EPS) GO TO 1
11   CONTINUE
     !                                                                                                                          
     PAUSE 'A too large, ITMAX too small'
     !                                                                                                                          
1    GAMSER = SUM*EXP(-X + A*LOG(X) - GLN)
     !                                                                                                                          
     RETURN
END SUBROUTINE GSER


SUBROUTINE GCF(GAMMCF,A,X,GLN)
!  Returns the incomplete gamma function Q(a,x) evaluated by its con-  !                                                   
!  tinued fraction representation as GAMMCF. Also returns ln(..a)      !                                                   
!  as GLN.  Press, W.H., et al., 1989. Numerical Recipes: p162.        !                                                   

      PARAMETER (ITMAX=100, EPS=3.E-7)

      GLN  = GAMMLN(A)
      GOLD = 0.     
      A0 = 1.       
      A1 = X        
      B0 = 0.       
      B1 = 1.
      FAC = 1.      
      DO 11, N=1,ITMAX
         AN  = FLOAT(N)
         ANA = AN-A
         A0  = (A1 + A0*ANA)*FAC 
         B0  = (B1 + B0*ANA)*FAC
         ANF = AN*FAC
         A1  = X*A0 + ANF*A1     
         B1  = X*B0 + ANF*B1
!                                                                                                                          
         IF (A1 .NE. 0.) THEN    
            FAC = 1./A1          
            G   = B1*FAC          
            IF (ABS((G-GOLD)/G) .LT. EPS) GO TO 1
            GOLD = G                             
         ENDIF
         !                                                                                                                          
11       CONTINUE
         !                                                                                                                          
         PAUSE 'A too large, ITMAX too small'
1        GAMMCF = EXP(-X + A*ALOG(X) - GLN)*G    
         !                                                                                                                          
         RETURN
       END SUBROUTINE GCF

FUNCTION GAMMLN(XX)

!  See Press, W.H., et al., 1989. Numerical Recipes: p157. for full    !                                                   
!  explanation...                                                                                                          


      REAL(kind=8) COF(6),STP,HALF,ONE,FPF,X,TMP,SER
!                                                                                                                          
!  Internal arithmetic could be done in double precision, a nicety that                                                    
!  has been omitted here; five-figure accuracy is considered as good enough                                                
!                                                                                                                          

      DATA COF,STP / 76.18009173,-86.50532033,24.01409822, -1.231739516,.120858003E-2,-.536382E-5,2.50662827465 /
      DATA HALF,ONE,FPF / 0.5,1.0,5.5 /
!                                                                                                                          
      X   = XX-ONE
      TMP = X+FPF
      TMP = (X+HALF) * LOG(TMP) - TMP
      SER = ONE
!                                                                                                                          
      DO 11, J=1,6
         X   = X+ONE
         SER = SER + COF(J)/X
   11 CONTINUE
!                                                                                                                          
      GAMMLN = TMP+LOG(STP*SER)
      !                                                                                                                    
      RETURN
   END FUNCTION GAMMLN


SUBROUTINE check_cache_primary()
  USE par_global, only: cached_primary,mtp,rnobb
  USE par_xilcps, only: pkTe,pKTbb,pTauy,pGeom,pCosi
  USE par_xilcps, only: pEcut, pGamma
  IMPLICIT NONE

  LOGICAL :: cached_mtp
  REAL, allocatable, SAVE :: cached_param(:)
  INTEGER, SAVE :: cached_mtype=-1
  LOGICAL, SAVE :: cached_rnobb
  INTEGER :: npar(3),in_mtp
  
  npar(1:3)=(/5,5,2/)

  in_mtp=mtp
  if(mtp.eq.5)in_mtp=1
  if(mtp.eq.6)in_mtp=2
  if(mtp.eq.7)in_mtp=3
  if(mtp.eq.8)in_mtp=1
  if(mtp.eq.9)in_mtp=1
  if(mtp.eq.10)in_mtp=2

  IF(cached_mtype.ne.in_mtp)THEN
     cached_mtp=.FALSE.
     cached_mtype=in_mtp
     cached_primary=.FALSE.
  ELSE
     cached_mtp=.TRUE.
  ENDIF


  IF (.not.cached_mtp)THEN
     if (allocated(cached_param)) then                                                                                  
        deallocate (cached_param)                                                                                       
     endif
     allocate(cached_param(npar(in_mtp)))

     IF(in_mtp.eq.1) THEN
        cached_param(1)=pGeom
        cached_param(2)=pkTbb
        cached_param(3)=pTauy
        cached_param(4)=pkTe
        cached_param(5)=pCosi
        cached_rnobb=rnobb
     ELSE IF(in_mtp.eq.2) THEN
        cached_param(1)=pGeom
        cached_param(2)=pkTbb
        cached_param(3)=pTauy
        cached_param(4)=pkTe
        cached_param(5)=pCosi
        cached_rnobb=rnobb
     ELSE IF(in_mtp.eq.3) THEN
        cached_param(1)=pGamma
        cached_param(2)=pEcut
     ENDIF
  ELSE
     cached_primary=.TRUE.
     IF(in_mtp.eq.1) THEN
        if(cached_param(1).ne.pGeom) cached_primary=.FALSE.
        if(cached_param(2).ne.pkTbb) cached_primary=.FALSE.
        if(cached_param(3).ne.pTauy) cached_primary=.FALSE.
        if(cached_param(4).ne.pkTe) cached_primary=.FALSE.
        if((pGeom.eq.1.0).and.(cached_param(5).ne.pCosi))cached_primary=.FALSE.
        if(cached_rnobb.neqv.rnobb)cached_primary=.FALSE.
        cached_param(1)=pGeom
        cached_param(2)=pkTbb
        cached_param(3)=pTauy
        cached_param(4)=pkTe
        cached_param(5)=pCosi
        cached_rnobb=rnobb
     ELSE IF(in_mtp.eq.2) THEN
        if(cached_param(1).ne.pGeom) cached_primary=.FALSE.
        if(cached_param(2).ne.pkTbb) cached_primary=.FALSE.
        if(cached_param(3).ne.pTauy) cached_primary=.FALSE.
        if(cached_param(4).ne.pkTe) cached_primary=.FALSE.
        if((pGeom.eq.1.0).and.(cached_param(5).ne.pCosi))cached_primary=.FALSE.
        if(cached_rnobb.neqv.rnobb)cached_primary=.FALSE.
        cached_param(1)=pGeom
        cached_param(2)=pkTbb
        cached_param(3)=pTauy
        cached_param(4)=pkTe
        cached_param(5)=pCosi
        cached_rnobb=rnobb
     ELSE IF(in_mtp.eq.3) THEN
        if(cached_param(1).ne.pGamma)cached_primary=.FALSE.
        if(cached_param(2).ne.pEcut)cached_primary=.FALSE.
        cached_param(1)=pGamma
        cached_param(2)=pEcut
     ELSE IF(in_mtp.eq.4) THEN
        cached_primary=.FALSE.
     ENDIF

  ENDIF

  return
END SUBROUTINE check_cache_primary


SUBROUTINE check_cache_reflection()
  USE par_global, only: mtp,is_lamppost,gshft,rnobb
  USE par_xilcps, only: pkTe,pKTbb, pTauy, pGeom, pCosi
  USE par_xilcps, only: pAFe, pEcut, gEcut, pLgXi, pXi, pIncl, pGamma, pLgN
  USE par_xilcps, only: cached_refl_1,cached_refl_2ir,cached_refl_2ps,cached_refl_fl1
  USE par_xilcps, only: RMtype,use_iref,use_xil
  IMPLICIT NONE

  LOGICAL :: cached_mtp
  REAL, allocatable, SAVE :: cached_param1(:)
  REAL, allocatable, SAVE :: cached_param2(:)
  LOGICAL, SAVE :: cached_rel_type
  REAL, SAVE :: cached_gshft
  INTEGER, SAVE :: cached_mtype=-1
  INTEGER :: npar(2,3),in_mtp
  LOGICAL, SAVE :: cached_rnobb,cached_use_iref,cached_use_xil

  character(len=20) :: FGSOLR
  character(len=10) :: FGSOLR_val
  character(len=10), SAVE :: cached_FGSOLR_val=''

  FGSOLR_val = trim(FGSOLR())

  npar(1,1:3)=(/5,5,5/)
  npar(2,1:3)=(/8,8,6/)

  in_mtp=mtp
  if(mtp.eq.5)in_mtp=1
  if(mtp.eq.6)in_mtp=2
  if(mtp.eq.7)in_mtp=3
  if(mtp.eq.8)in_mtp=1
  if(mtp.eq.9)in_mtp=1
  if(mtp.eq.10)in_mtp=2

  IF(cached_mtype.ne.mtp)THEN
     cached_mtp=.FALSE.
     cached_mtype=mtp

     cached_refl_1=.FALSE.
     cached_refl_2ir=.FALSE.
     cached_refl_2ps=.FALSE.
     cached_refl_fl1=.FALSE.
  ELSE
     cached_mtp=.TRUE.
  ENDIF

  IF (.not.cached_mtp)THEN
     if (allocated(cached_param1)) then                                                                                  
        deallocate (cached_param1)                                                                                       
     endif
     if (allocated(cached_param2)) then                                                                                  
        deallocate (cached_param2)                                                                                       
     endif

     allocate(cached_param1(npar(1,in_mtp)))
     allocate(cached_param2(npar(2,in_mtp)))

     IF(in_mtp.eq.1) THEN
        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(3)=  gEcut
        if(mtp.eq.5)cached_param1(3)=pLgN
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pkTe
        cached_param2(2)=pkTbb
        cached_param2(3)=pTauy
        cached_param2(4)=pGeom
        cached_param2(5)=pCosi
        cached_param2(6)=pAFe
        cached_param2(7)=pXi
        cached_param2(8)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_rnobb=rnobb
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ELSE IF(in_mtp.eq.2) THEN
        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(3)= gEcut
        if(mtp.eq.6)cached_param1(3)=pLgN
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pkTe
        cached_param2(2)=pkTbb
        cached_param2(3)=pTauy
        cached_param2(4)=pGeom
        cached_param2(5)=pCosi
        cached_param2(6)=pAFe
        cached_param2(7)=10.0**pLgXi
        cached_param2(8)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_rnobb=rnobb
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ELSE IF(in_mtp.eq.3) THEN
        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(3)=  gEcut
        if(mtp.eq.7)cached_param1(3)=pLgN
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pGamma
        cached_param2(2)=pEcut
        cached_param2(3)=pAFe
        cached_param2(4)=pCosi
        cached_param2(5)=10.0**pLgXi
        cached_param2(6)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ENDIF

  ELSE

     cached_refl_1=.TRUE.
     cached_refl_2ir=.TRUE.
     cached_refl_2ps=.TRUE.
     cached_refl_fl1=.TRUE.

     IF(in_mtp.eq.1) THEN

        if(cached_param1(1).ne.pGamma)cached_refl_1=.FALSE.
        if(cached_param1(2).ne.pAFe)cached_refl_1=.FALSE.
        IF(mtp.eq.5)THEN
           if(cached_param1(3).ne.pLgN)cached_refl_1=.FALSE.
           cached_param1(3)=  pLgN
        ELSE
           if(cached_param1(3).ne.gEcut)cached_refl_1=.FALSE.
           if(cached_param1(3).ne.gEcut)cached_refl_fl1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_fl1=.FALSE.
           cached_param1(3)=  gEcut
        ENDIF

        if(cached_param1(4).ne.pLgXi)cached_refl_1=.FALSE.
        if(cached_param1(5).ne.pIncl)cached_refl_1=.FALSE.
        if(cached_param1(1).ne.pGamma)cached_refl_fl1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_fl1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_fl1=.FALSE.

        if(cached_param2(1).ne.pkTe)cached_refl_2ps=.FALSE.
        if(cached_param2(2).ne.pkTbb)cached_refl_2ps=.FALSE.
        if(cached_param2(3).ne.pTauy)cached_refl_2ps=.FALSE.
        if(cached_param2(4).ne.pGeom)cached_refl_2ps=.FALSE.
        if(cached_param2(5).ne.pCosi)cached_refl_2ps=.FALSE.
        if(cached_param2(6).ne.pAFe)cached_refl_2ps=.FALSE.
        if(cached_param2(7).ne.pXi)cached_refl_2ps=.FALSE.
        if(cached_param2(8).ne.RMtype)cached_refl_2ps=.FALSE.
        if(cached_FGSOLR_val.ne.FGSOLR_val)cached_refl_2ps=.FALSE.
        if(cached_rnobb.neqv.rnobb)cached_refl_2ps=.FALSE.
        if(cached_use_iref.neqv.use_iref)cached_refl_2ps=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_2ps=.FALSE.
        
        if(.not.cached_refl_2ps)cached_refl_2ir=.FALSE.
        if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_2ir=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_2ir=.FALSE.

        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pkTe
        cached_param2(2)=pkTbb
        cached_param2(3)=pTauy
        cached_param2(4)=pGeom
        cached_param2(5)=pCosi
        cached_param2(6)=pAFe
        cached_param2(7)=pXi
        cached_param2(8)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_rnobb=rnobb
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ELSE IF(in_mtp.eq.2) THEN

        if(cached_param1(1).ne.pGamma)cached_refl_1=.FALSE.
        if(cached_param1(2).ne.pAFe)cached_refl_1=.FALSE.
        IF(mtp.eq.2)THEN
           if(cached_param1(3).ne.gEcut)cached_refl_1=.FALSE.
           if(cached_param1(3).ne.gEcut)cached_refl_fl1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_fl1=.FALSE.
           cached_param1(3)=  gEcut
        ELSE
           if(cached_param1(3).ne.pLgN)cached_refl_1=.FALSE.
           cached_param1(3)=  pLgN
        ENDIF
        if(cached_param1(4).ne.pLgXi)cached_refl_1=.FALSE.
        if(cached_param1(5).ne.pIncl)cached_refl_1=.FALSE.
        if(cached_param1(1).ne.pGamma)cached_refl_fl1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_fl1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_fl1=.FALSE.


        if(cached_param2(1).ne.pkTe)cached_refl_2ps=.FALSE.
        if(cached_param2(2).ne.pkTbb)cached_refl_2ps=.FALSE.
        if(cached_param2(3).ne.pTauy)cached_refl_2ps=.FALSE.
        if(cached_param2(4).ne.pGeom)cached_refl_2ps=.FALSE.
        if(cached_param2(5).ne.pCosi)cached_refl_2ps=.FALSE.
        if(cached_param2(6).ne.pAFe)cached_refl_2ps=.FALSE.
        if(cached_param2(7).ne.(10.0**pLgXi))cached_refl_2ps=.FALSE.
        if(cached_param2(8).ne.RMtype)cached_refl_2ps=.FALSE.
        if(cached_FGSOLR_val.ne.FGSOLR_val)cached_refl_2ps=.FALSE.
        if(cached_rnobb.neqv.rnobb)cached_refl_2ps=.FALSE.
        if(cached_use_iref.neqv.use_iref)cached_refl_2ps=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_2ps=.FALSE.
        
        if(.not.cached_refl_2ps)cached_refl_2ir=.FALSE.
        if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_2ir=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_2ir=.FALSE.

        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pkTe
        cached_param2(2)=pkTbb
        cached_param2(3)=pTauy
        cached_param2(4)=pGeom
        cached_param2(5)=pCosi
        cached_param2(6)=pAFe
        cached_param2(7)=10.0**pLgXi
        cached_param2(8)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_rnobb=rnobb
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ELSE IF(in_mtp.eq.3) THEN

        if(cached_param1(1).ne.pGamma)cached_refl_1=.FALSE.
        if(cached_param1(2).ne.pAFe)cached_refl_1=.FALSE.
        IF(mtp.eq.3)THEN
           if(cached_param1(3).ne.gEcut)cached_refl_1=.FALSE.
           if(cached_param1(3).ne.gEcut)cached_refl_fl1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_1=.FALSE.
           if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_fl1=.FALSE.
           cached_param1(3)=  gEcut
        ELSE
           if(cached_param1(3).ne.pLgN)cached_refl_1=.FALSE.
           cached_param1(3)=  pLgN
        ENDIF
        
        if(cached_param1(4).ne.pLgXi)cached_refl_1=.FALSE.
        if(cached_param1(5).ne.pIncl)cached_refl_1=.FALSE.
        if(cached_param1(1).ne.pGamma)cached_refl_fl1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_1=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_fl1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_1=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_fl1=.FALSE.

        if(cached_param2(1).ne.pGamma)cached_refl_2ps=.FALSE.
        if(cached_param2(2).ne.pEcut)cached_refl_2ps=.FALSE.
        if(cached_param2(3).ne.pAFe)cached_refl_2ps=.FALSE.
        if(cached_param2(4).ne.pCosi)cached_refl_2ps=.FALSE.
        if(cached_param2(5).ne.(10.0**pLgXi))cached_refl_2ps=.FALSE.
        if(cached_param2(6).ne.RMtype)cached_refl_2ps=.FALSE.
        if(cached_use_iref.neqv.use_iref)cached_refl_2ps=.FALSE.
        if(cached_use_xil.neqv.use_xil)cached_refl_2ps=.FALSE.

        if(.not.cached_refl_2ps)cached_refl_2ir=.FALSE.
        if(is_lamppost.and.(cached_gshft.ne.gshft))cached_refl_2ir=.FALSE.
        if(cached_rel_type.neqv.is_lamppost)cached_refl_2ir=.FALSE.
        if(cached_FGSOLR_val.ne.FGSOLR_val)cached_refl_2ir=.FALSE.


        cached_param1(1)=  pGamma
        cached_param1(2)=  pAFe
        cached_param1(4)=  pLgXi
        cached_param1(5)=  pIncl  

        cached_param2(1)=pGamma
        cached_param2(2)=pEcut
        cached_param2(3)=pAFe
        cached_param2(4)=pCosi
        cached_param2(5)=10.0**pLgXi
        cached_param2(6)=RMtype
        cached_rel_type=is_lamppost
        cached_gshft=gshft

        cached_FGSOLR_val=FGSOLR_val
        cached_use_iref=use_iref
        cached_use_xil=use_xil

     ENDIF
  ENDIF

  return
END SUBROUTINE check_cache_reflection


SUBROUTINE INIT_TABLE_kTtG_PATHS()
  use par_global, only: reflkerr_tables
  use par_xilcps, only: fGkTtau,fGkTmutau
  use par_xilcps, only: cached_kTt_G_tbp
  IMPLICIT NONE
  
  IF(.not.cached_kTt_G_tbp) THEN
     fGkTtau=trim(reflkerr_tables)//'/GkTtau.fits'
     fGkTmutau=trim(reflkerr_tables)//'/GkTmutau.fits'

     cached_kTt_G_tbp = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE INIT_TABLE_kTtG_PATHS


SUBROUTINE INIT_TABLE_IRRSLB_PATH()
  use par_global, only: reflkerr_tables
  use par_xilcps, only: firrslb
  use par_xilcps, only: cached_firrslb_tbp
  IMPLICIT NONE

  IF(.not.cached_firrslb_tbp) THEN
     firrslb=trim(reflkerr_tables)//'/irrslb.fits'
     cached_firrslb_tbp = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE INIT_TABLE_IRRSLB_PATH


SUBROUTINE INIT_kTtG_TABLES(kTtGlim,GTau,GBn,kTBn)
  use par_xilcps, only: fGkTtau
  IMPLICIT NONE

  INTEGER :: GBn, kTBn
  REAL :: kTtGlim(0:kTBn,2),GTau(0:GBn,0:kTBn)
  logical, save :: FIRSTCALL = .TRUE.
  INTEGER :: ikT,iG

  INTEGER :: unit,rw=0,blk_size=1,stat=0
  INTEGER :: hdu_nr,hdu_curr,hdu_t
  INTEGER :: c_nr,first_el=1
  INTEGER :: ANULL,DNULL=0                                                                                              

  IF(FIRSTCALL) then

     call ftgiou(unit,stat)

     call ftopen(unit,fGkTtau,rw,blk_size,stat)
     call printfitserror(stat)
     if (stat.ne.0) stop

     call FTTHDU(unit, hdu_nr, stat)
     if (hdu_nr.ne.(3)) then
        print *,"Wrong number of extensions => check:",trim(fGkTtau)," table!"
        stop
     endif

     hdu_curr = 2
     call FTMAHD(unit,hdu_curr,hdu_t,stat)

     DO ikT=1,kTBn+1
        c_nr=2
        call ftgcve(unit,c_nr,ikT,first_el,1,DNULL,kTtGlim(ikT-1,1), &
             ANULL,stat)
        c_nr=3
        call ftgcve(unit,c_nr,ikT,first_el,1,DNULL,kTtGlim(ikT-1,2), &
             ANULL,stat)
     ENDDO

     hdu_curr = hdu_curr+1
     call FTMAHD(unit,hdu_curr,hdu_t,stat)

     DO iG=1,GBn+1
        c_nr=2
        call ftgcve(unit,c_nr,iG,first_el,kTBn+1,DNULL,GTau(iG-1,:), &
             ANULL,stat)
     ENDDO

     call ftclos(unit, stat)
     call ftfiou(unit, stat)

     FIRSTCALL=.FALSE.

  ENDIF

  RETURN
END SUBROUTINE INIT_kTtG_TABLES


SUBROUTINE INIT_mukTtG_TABLES(mukTtGlim,muGTau,cosiBn,muGBn,mukTBn)
  use par_xilcps, only: fGkTmutau
  IMPLICIT NONE

  INTEGER :: cosiBn,muGBn,mukTBn
  REAL :: mukTtGlim(0:cosiBn,0:mukTBn,2),muGTau(0:cosiBn,0:muGBn,0:mukTBn)
  logical, save :: FIRSTCALL = .TRUE.
  INTEGER :: imu,iG
  INTEGER :: unit,status

  integer :: readwrite,blocksize                                                                                           
  integer :: hdunum,chdu                                                                                                   
  integer :: hdutype,colnum                                                                                                
  integer :: felem=1, nulld=0                                                                                              
  logical :: anynull                                                                                                       

  status=0                                                                                                                 
  readwrite=0                                                                                                              
  blocksize=1

  IF(FIRSTCALL) then

     call ftgiou(unit,status)

     call ftopen(unit,fGkTmutau,readwrite,blocksize,status)
     call printfitserror(status)
     if (status.ne.0) stop

     call FTTHDU(unit, hdunum, status)
     if (hdunum.ne.3) then
        print *,'Number of Extensions is not correct - check table!'
        stop
     endif

     chdu = 2
     call FTMAHD(unit,chdu,hdutype,status)

     DO imu=1,cosiBn+1                                                                                                  
        colnum=3                                                                                                        
        call ftgcve(unit,colnum,imu,felem,mukTBn+1,nulld,mukTtGlim(imu-1,:,1), &
             anynull,status)
        colnum=4                                                                                                        
        call ftgcve(unit,colnum,imu,felem,mukTBn+1,nulld,mukTtGlim(imu-1,:,2), &
             anynull,status)
     ENDDO

     chdu = chdu+1
     call FTMAHD(unit,chdu,hdutype,status)

     DO iG=1,muGBn+1
        do imu=0,cosiBn
           colnum=imu+2
           call ftgcve(unit,colnum,iG,felem,mukTBn+1,nulld,muGTau(imu,iG-1,:), &
                anynull,status)
        enddo
     ENDDO

     call ftclos(unit, status)
     call ftfiou(unit, status)

     FIRSTCALL=.FALSE.

  ENDIF

  RETURN
END SUBROUTINE INIT_mukTtG_TABLES


SUBROUTINE INIT_IRRSLB_TABLES(slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn)
  use par_xilcps, only: firrslb
  IMPLICIT NONE

  INTEGER :: kTBn,kTbbBn,TBn,cosiBn
  REAL :: slope(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn),fluxscl(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn)
  logical, save :: FIRSTCALL = .TRUE.
  INTEGER :: imu,ikT,ikTbb
  INTEGER :: unit,status

  integer :: readwrite,blocksize                                                                                           
  integer :: hdunum,hdutype
  integer :: felem=1, nulld=0                                                                                              
  logical :: anynull                                                                                                       

  status=0                                                                                                                 
  readwrite=0                                                                                                              
  blocksize=1

  IF(FIRSTCALL) then

     call ftgiou(unit,status)

     call ftopen(unit,firrslb,readwrite,blocksize,status)
     call printfitserror(status)
     if (status.ne.0) stop

     call FTTHDU(unit, hdunum, status)
     if (hdunum.ne.(2+(cosiBn+1))) then
        print *,'Number of Extensions is not correct - check table!'
        stop
     endif

     do imu=3,3+cosiBn
        call FTMAHD(unit,imu,hdutype,status)

        DO ikTbb=1,kTbbBn+1
           do ikT=1,kTBn+1
              call ftgcve(unit,ikTbb,ikT,felem,TBn+1,nulld,slope(imu-3,ikTbb-1,ikT-1,:), &
                   anynull,status)
           enddo
        ENDDO
        DO ikTbb=kTbbBn+2,2*(kTbbBn+1)
           do ikT=1,kTBn+1
              call ftgcve(unit,ikTbb,ikT,felem,TBn+1,nulld,fluxscl(imu-3,ikTbb-kTbbBn-2,ikT-1,:), &
                   anynull,status)
           enddo
        ENDDO
     enddo

     call ftclos(unit, status)
     call ftfiou(unit, status)

     FIRSTCALL=.FALSE.
  ENDIF

  RETURN
END SUBROUTINE INIT_IRRSLB_TABLES


REAL function get_tau(gamma,kTe,kTtGlim,GTau,GBn,kTBn)
  IMPLICIT NONE

  INTEGER :: GBn,kTBn
  REAL :: gamma,kTe,kTtGlim(0:kTBn,2),GTau(0:GBn,0:kTBn)

  REAL, PARAMETER::Gl=1.2
  REAL, PARAMETER::Gh=3.4
  REAL, PARAMETER::kTl=10.0
  REAL, PARAMETER::kTh=1000.0

  INTEGER :: iG,ikT
  REAL :: G_l,G_h,factor_g
  REAL :: kT_l,kT_h,factor_t
  REAL :: G_min,G_max

  get_tau=-1.0

  if(gamma.ge.(Gh-0.0051))gamma=Gh-0.0051

  iG = NINT((gamma-Gl)*GBn/(Gh-Gl))
  G_l=Gl+iG*((Gh-Gl)/GBn)
  G_h=Gl+(iG+1)*((Gh-Gl)/GBn)
  factor_g=(gamma-G_l)/(G_h-G_l)

10 CONTINUE
  ikT = NINT((kTe-kTl)*kTBn/(kTh-kTl))

  G_min=CEILING(kTtGlim(ikT,1)*100.0)/100.0
  G_max=FLOOR(kTtGlim(ikT+1,2)*100.0)/100.0

  IF((gamma.gt.G_min).and.(gamma.lt.G_max)) THEN

     kT_l=kTl+ikT*((kTh-kTl)/kTBn)
     kT_h=kTl+(ikT+1)*((kTh-kTl)/kTBn)
     factor_t=(kTe-kT_l)/(kT_h-kT_l)

     get_tau=(1.0-factor_g)*(1.0-factor_t)*GTau(iG,ikT) &
          +(1.0-factor_g)*factor_t*GTau(iG,ikT+1) &
          +factor_g*(1.0-factor_t)*GTau(iG+1,ikT) &
          +factor_g*factor_t*GTau(iG+1,ikT+1)

  ELSE

     IF(gamma.lt.G_min)THEN
        write(*,*) "Warning! Slope=", gamma,  "is below the lower limit for kTe=",kTe
        DO WHILE(gamma.lt.G_min)
           ikT=ikT+1
           G_min=CEILING(kTtGlim(ikT,1)*100.0)/100.0
        ENDDO
        kTe=kTl+ikT*(kTh-kTl)/kTBn
        write(*,*) "Setting kTe=",kTe
        GOTO 10
     ELSE IF(gamma.eq.G_min)THEN
        kT_l=kTl+ikT*((kTh-kTl)/kTBn)
        kT_h=kTl+(ikT+1)*((kTh-kTl)/kTBn)
        factor_t=(kTe-kT_l)/(kT_h-kT_l)
        get_tau=(1.0-factor_t)*GTau(iG,ikT)+factor_t*GTau(iG,ikT+1)
     ELSE IF(gamma.gt.G_max)THEN
        write(*,*) "Warning! Slope=", gamma,  "is above the upper limit for kTe=",kTe
        DO WHILE(gamma.gt.G_max)
           ikT=ikT-1
           G_max=FLOOR(kTtGlim(ikT+1,2)*100.0)/100.0
        ENDDO
        kTe=kTl+ikT*(kTh-kTl)/kTBn
        write(*,*) "Setting kTe=",kTe
        GOTO 10
     ELSE IF(gamma.eq.G_max)THEN
        kT_l=kTl+ikT*((kTh-kTl)/kTBn)
        kT_h=kTl+(ikT+1)*((kTh-kTl)/kTBn)
        factor_t=(kTe-kT_l)/(kT_h-kT_l)
        get_tau=(1.0-factor_t)*GTau(iG,ikT)+factor_t*GTau(iG,ikT+1)
     ENDIF

  ENDIF

  RETURN
end function get_tau


REAL function get_tau_mu(gamma,cosi,kTe,mukTtGlim,muGTau,cosiBn,muGBn,mukTBn)
  use par_global, only: Pi
  IMPLICIT NONE

  INTEGER :: cosiBn,muGBn,mukTBn
  REAL :: gamma,cosi,kTe,mukTtGlim(0:cosiBn,0:mukTBn,2),muGTau(0:cosiBn,0:muGBn,0:mukTBn)

  REAL, PARAMETER::Gl=1.2
  REAL, PARAMETER::Gh=3.4
  REAL, PARAMETER::cosil=0.012217
  REAL, PARAMETER::cosih=0.987688
  REAL, PARAMETER::kTl=10.0
  REAL, PARAMETER::kTh=1000.0

  INTEGER :: iG,imu,ikT
  REAL :: G_l,G_h,c_l,c_h,kT_l,kT_h,factor_g,factor_c,factor_t
  REAL :: G_min, G_max

  get_tau_mu=-1.0
  if(gamma.ge.(Gh-0.0051))gamma=Gh-0.0051
  iG = INT((gamma-Gl)*muGBn/(Gh-Gl))
  imu = INT((cosi-cosil)*cosiBn/(cosih-cosil))

  G_l=Gl+iG*((Gh-Gl)/muGBn)
  G_h=Gl+(iG+1)*((Gh-Gl)/muGBn)
  c_l=cosil+imu*((cosih-cosil)/cosiBn)
  c_h=cosil+(imu+1)*((cosih-cosil)/cosiBn)

  factor_g=(gamma-G_l)/(G_h-Gl)
  factor_c=(cosi-c_l)/(c_h-c_l)

10 CONTINUE

  ikT = INT((kTe-kTl)*mukTBn/(kTh-kTl))

  G_min=CEILING(mukTtGlim(imu,ikT,1)*100.0)/100.0
  G_max=FLOOR(mukTtGlim(imu,ikT+1,2)*100.0)/100.0
 
  IF((gamma.gt.G_min).and.(gamma.lt.G_max)) THEN
     kT_l=kTl+ikT*((kTh-kTl)/mukTBn)
     kT_h=kTl+(ikT+1)*((kTh-kTl)/mukTBn)
     factor_t=(kTe-kT_l)/(kT_h-kT_l)

     get_tau_mu=(1.0-factor_c)*(1.0-factor_g)*(1.0-factor_t)*muGTau(imu,iG,ikT) &
          +factor_c*(1.0-factor_g)*(1.0-factor_t)*muGTau(imu+1,iG,ikT) &
          +(1.0-factor_c)*factor_g*(1.0-factor_t)*muGTau(imu,iG+1,ikT) &
          +(1.0-factor_c)*(1.0-factor_g)*factor_t*muGTau(imu,iG,ikT+1) &
          +factor_c*factor_g*(1.0-factor_t)*muGTau(imu+1,iG+1,ikT) &
          +factor_c*(1.0-factor_g)*factor_t*muGTau(imu+1,iG,ikT+1) &
          +(1.0-factor_c)*factor_g*factor_t*muGTau(imu,iG+1,ikT+1) &
          +factor_c*factor_g*factor_t*muGTau(imu+1,iG+1,ikT+1)
  ELSE

     IF(gamma.lt.G_min)THEN
        write(*,*) "Warning! Slope=", gamma,  "is below the lower limit for kTe=",kTe, &
             "and Incl=",real(ACOS(cosi)*180.0/Pi)
        DO WHILE(gamma.lt.G_min)
           ikT=ikT+1
           G_min=CEILING(mukTtGlim(imu,ikT,1)*100.0)/100.0
        ENDDO
        kTe=kTl+ikT*(kTh-kTl)/mukTBn
        write(*,*) "Setting kTe=",kTe
        GOTO 10
     ELSE IF(gamma.eq.G_min)THEN
        kT_l=kTl+ikT*((kTh-kTl)/mukTBn)
        kT_h=kTl+(ikT+1)*((kTh-kTl)/mukTBn)
        factor_t=(kTe-kT_l)/(kT_h-kT_l)
        get_tau_mu=(1.0-factor_c)*(1.0-factor_t)*muGTau(imu,iG,ikT) & 
             +factor_c*(1.0-factor_t)*muGTau(imu+1,iG,ikT) &
             +(1.0-factor_c)*factor_t*muGTau(imu,iG,ikT+1) &
             +factor_c*factor_t*muGTau(imu+1,iG,ikT+1)
     ELSE IF(gamma.gt.G_max)THEN
        write(*,*) "Warning! Slope=", gamma,  "is above the upper limit for kTe=",kTe, &
             "and Incl=",real(ACOS(cosi)*180.0/Pi)
        DO WHILE(gamma.gt.G_max)
           ikT=ikT-1
           G_max=FLOOR(mukTtGlim(imu,ikT+1,2)*100.0)/100.0
        ENDDO
        kTe=kTl+ikT*(kTh-kTl)/mukTBn
        write(*,*) "Setting kTe=",kTe
        GOTO 10
     ELSE IF(gamma.eq.G_max)THEN
        kT_l=kTl+ikT*((kTh-kTl)/mukTBn)
        kT_h=kTl+(ikT+1)*((kTh-kTl)/mukTBn)
        factor_t=(kTe-kT_l)/(kT_h-kT_l)
        get_tau_mu=(1.0-factor_c)*(1.0-factor_t)*muGTau(imu,iG,ikT) & 
             +factor_c*(1.0-factor_t)*muGTau(imu+1,iG,ikT) &
             +(1.0-factor_c)*factor_t*muGTau(imu,iG,ikT+1) &
             +factor_c*factor_t*muGTau(imu+1,iG,ikT+1)
     ENDIF
  ENDIF

  RETURN
end function get_tau_mu


SUBROUTINE get_slope_irrslb(gam_out,fluxscl_out,slope,fluxscl,kTBn,kTbbBn,TBn,cosiBn,kTe,kTbb,tau,incl)
  use par_global, only: pi
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_kTbb_l,g1_kTbb_h,g1_tau_l,g1_tau_h, &
       g1_incl_l,g1_incl_h
  IMPLICIT NONE

  INTEGER :: kTBn,kTbbBn,TBn,cosiBn
  REAL :: slope(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn),fluxscl(0:cosiBn,0:kTbbBn,0:kTBn,0:TBn)
  REAL :: gam_out,fluxscl_out
  REAL :: kTe,kTbb,tau,incl
  INTEGER :: in_mu,in_kT,in_kTbb,in_T

  REAL :: lg_kTe,lg_kTbb,lg_tau,cosi

  REAL, PARAMETER::lg_kTl=Log10(g1_kT_l) !10.0
  REAL, PARAMETER::lg_kTh=Log10(g1_kT_h) !500.0                                                                      
  REAL, PARAMETER::lg_kTbbl=Log10(g1_kTbb_l) !1e-3
  REAL, PARAMETER::lg_kTbbh=Log10(g1_kTbb_h) !1                                           
  REAL, PARAMETER::lg_Tl=Log10(g1_tau_l) !(0.05)
  REAL, PARAMETER::lg_Th=Log10(g1_tau_h) !(4.0)                                                                      
  REAL, PARAMETER::cosil=cos(g1_incl_h*real(pi)/180.0)
  REAL, PARAMETER::cosih=cos(g1_incl_l*real(pi)/180.0)

  REAL :: l_kT, h_kT,l_kTbb, h_kTbb,l_mu, h_mu,l_tau, h_tau
  REAL :: f_kT,f_kTbb,f_mu,f_t
  REAL :: w1,w2,w3,w4,w5,w6,w7,w8,w9,w10,w11,w12,w13,w14,w15,w16

  IF(tau.lt.0.0)THEN
     lg_tau=Log10(-tau/(0.0078277886*kTe))   !tau=y/(4*Te/511)
  ELSE
     lg_tau=Log10(tau)
  ENDIF

  lg_kTe=Log10(kTe)
  lg_kTbb=Log10(kTbb)
  cosi=cos(incl*real(pi)/180.0)

  in_kT = INT((lg_kTe-lg_kTl)*kTBn/(lg_kTh-lg_kTl))
  in_kTbb = INT((lg_kTbb-lg_kTbbl)*kTbbBn/(lg_kTbbh-lg_kTbbl))
  in_T = INT((lg_tau-lg_Tl)*TBn/(lg_Th-lg_Tl))
  in_mu = INT((cosi-cosil)*cosiBn/(cosih-cosil))
  if(in_kT.lt.0)in_kT=0
  if(in_kT.ge.kTBn)in_kT=kTBn-1
  if(in_kTbb.lt.0)in_kTbb=0
  if(in_kTbb.ge.kTbbBn)in_kTbb=kTbbBn-1
  if(in_T.lt.0)in_T=0
  if(in_T.ge.TBn)in_T=TBn-1
  if(in_mu.lt.0)in_mu=0
  if(in_mu.ge.cosiBn)in_mu=cosiBn-1

  l_kT=lg_kTl+in_kT*((lg_kTh-lg_kTl)/kTBn)
  h_kT=lg_kTl+(in_kT+1)*((lg_kTh-lg_kTl)/kTBn)

  l_kTbb=lg_kTbbl+in_kTbb*((lg_kTbbh-lg_kTbbl)/kTbbBn)
  h_kTbb=lg_kTbbl+(in_kTbb+1)*((lg_kTbbh-lg_kTbbl)/kTbbBn)

  l_tau=lg_Tl+in_T*((lg_Th-lg_Tl)/TBn)
  h_tau=lg_Tl+(in_T+1)*((lg_Th-lg_Tl)/TBn)

  l_mu=cosil+in_mu*((cosih-cosil)/cosiBn)
  h_mu=cosil+(in_mu+1)*((cosih-cosil)/cosiBn)

  f_mu=(cosi-l_mu)/(h_mu-l_mu)
  f_kTbb=(lg_kTbb-l_kTbb)/(h_kTbb-l_kTbb)
  f_kT=(lg_kTe-l_kT)/(h_kT-l_kT)
  f_t=(lg_tau-l_tau)/(h_tau-l_tau)

  w1=(1.0-f_mu)*(1.0-f_kTbb)*(1.0-f_kT)*(1.0-f_t)
  w2=f_mu*(1.0-f_kTbb)*(1.0-f_kT)*(1.0-f_t)
  w3=(1.0-f_mu)*f_kTbb*(1.0-f_kT)*(1.0-f_t)
  w4=(1.0-f_mu)*(1.0-f_kTbb)*f_kT*(1.0-f_t)
  w5=(1.0-f_mu)*(1.0-f_kTbb)*(1.0-f_kT)*f_t
  w6=f_mu*f_kTbb*(1.0-f_kT)*(1.0-f_t)
  w7=(1.0-f_mu)*f_kTbb*f_kT*(1.0-f_t)
  w8=(1.0-f_mu)*(1.0-f_kTbb)*f_kT*f_t
  w9=f_mu*(1.0-f_kTbb)*f_kT*(1.0-f_t)
  w10=(1.0-f_mu)*f_kTbb*(1.0-f_kT)*f_t
  w11=f_mu*(1.0-f_kTbb)*(1.0-f_kT)*f_t
  w12=f_mu*f_kTbb*f_kT*(1.0-f_t)
  w13=(1.0-f_mu)*f_kTbb*f_kT*f_t
  w14=f_mu*f_kTbb*(1.0-f_kT)*f_t
  w15=f_mu*(1.0-f_kTbb)*f_kT*f_t
  w16=f_mu*f_kTbb*f_kT*f_t

  gam_out=w1*slope(in_mu,in_kTbb,in_kT,in_T)+ &
       w2*slope(in_mu+1,in_kTbb,in_kT,in_T)+ &
       w3*slope(in_mu,in_kTbb+1,in_kT,in_T)+ &
       w4*slope(in_mu,in_kTbb,in_kT+1,in_T)+ &
       w5*slope(in_mu,in_kTbb,in_kT,in_T+1)+ &
       w6*slope(in_mu+1,in_kTbb+1,in_kT,in_T)+ &
       w7*slope(in_mu,in_kTbb+1,in_kT+1,in_T)+ &
       w8*slope(in_mu,in_kTbb,in_kT+1,in_T+1)+ &
       w9*slope(in_mu+1,in_kTbb,in_kT+1,in_T)+ &
       w10*slope(in_mu,in_kTbb+1,in_kT,in_T+1)+ &
       w11*slope(in_mu+1,in_kTbb,in_kT,in_T+1)+ &
       w12*slope(in_mu+1,in_kTbb+1,in_kT+1,in_T)+ &
       w13*slope(in_mu,in_kTbb+1,in_kT+1,in_T+1)+ &
       w14*slope(in_mu+1,in_kTbb+1,in_kT,in_T+1)+ &
       w15*slope(in_mu+1,in_kTbb,in_kT+1,in_T+1)+ &
       w16*slope(in_mu+1,in_kTbb+1,in_kT+1,in_T+1)

  fluxscl_out=w1*fluxscl(in_mu,in_kTbb,in_kT,in_T)+ &
       w2*fluxscl(in_mu+1,in_kTbb,in_kT,in_T)+ &
       w3*fluxscl(in_mu,in_kTbb+1,in_kT,in_T)+ &
       w4*fluxscl(in_mu,in_kTbb,in_kT+1,in_T)+ &
       w5*fluxscl(in_mu,in_kTbb,in_kT,in_T+1)+ &
       w6*fluxscl(in_mu+1,in_kTbb+1,in_kT,in_T)+ &
       w7*fluxscl(in_mu,in_kTbb+1,in_kT+1,in_T)+ &
       w8*fluxscl(in_mu,in_kTbb,in_kT+1,in_T+1)+ &
       w9*fluxscl(in_mu+1,in_kTbb,in_kT+1,in_T)+ &
       w10*fluxscl(in_mu,in_kTbb+1,in_kT,in_T+1)+ &
       w11*fluxscl(in_mu+1,in_kTbb,in_kT,in_T+1)+ &
       w12*fluxscl(in_mu+1,in_kTbb+1,in_kT+1,in_T)+ &
       w13*fluxscl(in_mu,in_kTbb+1,in_kT+1,in_T+1)+ &
       w14*fluxscl(in_mu+1,in_kTbb+1,in_kT,in_T+1)+ &
       w15*fluxscl(in_mu+1,in_kTbb,in_kT+1,in_T+1)+ &
       w16*fluxscl(in_mu+1,in_kTbb+1,in_kT+1,in_T+1)

  RETURN
END SUBROUTINE get_slope_irrslb


REAL function get_mkTe(gamma_in, kTe_in)
  use par_global, only: reflkerr_tables

  IMPLICIT NONE

  REAL :: gamma_in,kTe_in

  character(LEN=120)::fmkTe
  LOGICAL,SAVE :: cached_mkTe=.FALSE.
  INTEGER :: unit,status

  REAL, PARAMETER :: g_min=1.2
  REAL, PARAMETER :: g_max=3.0
  REAL, PARAMETER :: kTe_min=10.0
  REAL, PARAMETER :: kTe_max=320.0
  INTEGER, PARAMETER :: nG=10
  INTEGER, PARAMETER :: nkT=63
  REAL :: g_tab(nG),kTe_map(nG,nkT)
  REAL :: dkTe=(kTe_max-kTe_min)/(nkT-1)
  INTEGER :: iG,ikTe
  REAL :: gamma, kTe_l,kTe_h,g_l,g_h,f_g,f_kTe

  IF(kTe_in.lt.kTe_min)THEN
     get_mkTe=10.0
     RETURN
  ELSE IF(kTe_in.gt.kTe_max)THEN
     get_mkTe=450.0
     RETURN
  ELSE

     IF(.not.cached_mkTe) THEN
        fmkTe=trim(reflkerr_tables)//'/kTe_map.dat'

        status=0                                                                       
        call ftgiou(unit,status)                                                       
        open(unit,file=fmkTe)                                                      
        DO ikTe=1,nkT
           read(unit,*) (kTe_map(iG,ikTe),iG=1,nG) 
        ENDDO
        close(unit)
        call ftfiou(unit,status)

        cached_mkTe = .TRUE.
     ENDIF

  gamma=gamma_in
  if(gamma.lt.g_min)gamma=g_min
  if(gamma.gt.g_max)gamma=g_max

  g_tab(1:nG)=(/1.2,1.4,1.6,1.7,1.9,2.1,2.4,2.6,2.8,3.0/)

  iG=1
  DO WHILE(g_tab(iG+1).lt.gamma)
     iG=iG+1
  ENDDO

  IF(kTe_in.eq.kTe_max)THEN
     ikTe=nkT-1   
  ELSE
     ikTe=INT((kTe_in-kTe_min)/dkTe+1)
  ENDIF

  g_l=g_tab(iG)
  g_h=g_tab(iG+1)
  kTe_l=kTe_min+(ikTe-1)*dkTe
  kTe_h=kTe_min+ikTe*dkTe

  f_g=(gamma-g_l)/(g_h-g_l)
  f_kTe=(kTe_in-kTe_l)/(kTe_h-kTe_l)

  get_mkTe=(1.0-f_g)*(1.0-f_kTe)*kTe_map(iG,ikTe)+f_g*(1.0-f_kTe)*kTe_map(iG+1,ikTe)+ &
       (1.0-f_g)*f_kTe*kTe_map(iG,ikTe+1)+f_g*f_kTe*kTe_map(iG+1,ikTe+1)

  IF(get_mkTe.lt.10.0)get_mkTe=10.0
  IF(get_mkTe.gt.450.0)get_mkTe=450.0

ENDIF

  RETURN
END function get_mkTe


REAL function get_xEcut(gamma_in, Ecut_in)
  IMPLICIT NONE

  REAL :: gamma_in,Ecut_in

  REAL, PARAMETER :: g_min=1.2
  REAL, PARAMETER :: g_max=3.4
  INTEGER, PARAMETER :: nG=11
  REAL :: dg=(g_max-g_min)/nG
  INTEGER :: iG,i
  REAL :: gamma, a(nG+1,0:7),x_l,x_h,g_l,g_h,f_g

  gamma=gamma_in
  if(gamma.lt.g_min)gamma=g_min
  if(gamma.gt.g_max)gamma=g_max

  iG=INT((gamma-g_min)/dg+1)
  if(gamma.eq.g_min)iG=1
  if(gamma.eq.g_max)iG=nG
  

  a(1,0:7)=(/3.8041, -0.00793566, 1.42188e-05, -9.25726e-09, 0.0, 0.0, 0.0, 0.0/)
  a(2,0:7)=(/3.82982, -0.0119757, 3.15214e-05, -2.96396e-08, 0.0, 0.0, 0.0, 0.0/)
  a(3,0:7)=(/3.89417, -0.0156245, 4.9152e-05, -5.28423e-08, 0.0, 0.0, 0.0, 0.0/)
  a(4,0:7)=(/4.05442, -0.0205556, 7.51868e-05, -9.12016e-08, 0.0, 0.0, 0.0, 0.0/)
  a(5,0:7)=(/4.63403, -0.0407033, 0.000266889, -8.26949e-07, 1.23408e-09, -7.89967e-13, 0.0, 0.0/)
  a(6,0:7)=(/5.28432, -0.0554919, 0.000353592, -9.04991e-07, 8.27856e-10, -2.3472e-13, 0.0, 0.0/)
  a(7,0:7)=(/6.61979, -0.118317, 0.00137336, -8.16066e-06, 2.44283e-08, -2.89619e-11, 0.0, 0.0/)
  a(8,0:7)=(/10.0639, -0.394437, 0.0095643, -0.000126423, 9.55108e-07, -4.12835e-09, 9.54836e-12, -9.20351e-15/)
  a(9,0:7)=(/16.4488, -0.872962, 0.0255676, -0.000412772, 3.88474e-06, -2.12638e-08, 6.27724e-11, -7.71474e-14/)
  a(10,0:7)=(/22.9971, -1.11363, 0.028946, -0.000427833, 3.84232e-06, -2.09828e-08, 6.43841e-11, -8.47899e-14/)
  a(11,0:7)=(/38.9383, -1.59824, 0.0341554, -0.000435083, 3.69184e-06, -2.08993e-08, 7.00134e-11, -1.01391e-13/)
  a(12,0:7)=(/49.7, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/)
     
  x_l=0.0
  x_h=0.0
  g_l=g_min+(iG-1)*dg
  g_h=g_min+iG*dg
  DO i=0,7
     x_l=x_l+a(iG,i)*(Ecut_in**i)
     x_h=x_h+a(iG+1,i)*(Ecut_in**i)
  ENDDO

  f_g=(gamma-g_l)/(g_h-g_l)

  get_xEcut=((1.0-f_g)*x_l+f_g*x_h)*Ecut_in
!@@
  if(get_xEcut.le.Ecut_in)get_xEcut=Ecut_in
  IF(get_xEcut.lt.10.0)get_xEcut=10.0
  IF(get_xEcut.gt.1000.0)get_xEcut=1000.0

  RETURN
END function get_xEcut


subroutine printfitserror(status)
  integer :: status
  character (len=30) ::  errtext
  character (len=80) :: errmessage
  !     ---------------------------------
  if (status .le. 0)return
  call ftgerr(status,errtext)
  print *,'FITSIO Error Status =',status,': ',errtext
  call ftgmsg(errmessage)
  do while (errmessage .ne. ' ')
     print *,errmessage
     call ftgmsg(errmessage)
  end do
end subroutine printfitserror


subroutine check_table_paths()
  use par_global, only: reflkerr_tables
  implicit none
  character (len=24) :: env="REFLKERR_TABLES"
  character (len=240) :: var

  call GETENV(env,var)
  if (len_trim(var).gt.0) reflkerr_tables=trim(var)
  if (len_trim(reflkerr_tables).eq.0) reflkerr_tables="."

end subroutine check_table_paths


SUBROUTINE f_rxillver(ear, ne, param, ifl, photar, photer)
  use iso_c_binding
  IMPLICIT NONE
  
  interface c_interface

     subroutine lmodxillver (ear_d, ne, param_d, ifl, photar_d, photer_d, init) &
          bind (C, name = "lmodxillver")

       use iso_c_binding
       implicit none

       real (kind = c_double), intent (in), dimension (*) :: ear_d
       integer (kind=c_int), intent (in),VALUE :: ne
       integer (kind=c_int), intent (in),VALUE :: ifl
       character (kind=c_char), intent (in), dimension (*) :: init
       real (kind=c_double), intent (in),dimension (*) :: param_d
       real (kind=c_double), intent (out), dimension (*) :: photar_d
       real (kind=c_double), intent (out), dimension (*) :: photer_d

     end subroutine lmodxillver

  end interface c_interface

  INTEGER :: ne, ifl,i
  INTEGER,PARAMETER :: npm=7
  REAL :: ear(0:ne), param(npm), photar(ne), photer(ne)
  REAL(kind=8) :: photar_d(ne), photer_d(ne)
  character(LEN=40) :: init

  photer(1)=0.0

  call lmodxillver(1.0d0*ear, ne, 1.0d0*param, ifl, photar_d, photer_d, init)

  DO i=1,ne
     photar(i)=real(photar_d(i))
  ENDDO
  
  return
END SUBROUTINE f_rxillver


SUBROUTINE f_rxillvernthcomp(ear, ne, param, ifl, photar, photer)
  use iso_c_binding
  IMPLICIT NONE

  interface c_interface

     subroutine lmodxillvernthcomp (ear_d, ne, param_d, ifl, photar_d, photer_d, init) &
          bind (C, name = "lmodxillvernthcomp")

       use iso_c_binding
       implicit none

       real (kind = c_double), intent (in), dimension (*) :: ear_d
       integer (kind=c_int), intent (in),VALUE :: ne
       integer (kind=c_int), intent (in),VALUE :: ifl
       character (kind=c_char), intent (in), dimension (*) :: init
       real (kind=c_double), intent (in),dimension (*) :: param_d
       real (kind=c_double), intent (out), dimension (*) :: photar_d
       real (kind=c_double), intent (out), dimension (*) :: photer_d

     end subroutine lmodxillvernthcomp

  end interface c_interface

  INTEGER :: ne, ifl,i
  INTEGER,PARAMETER :: npm=7
  REAL :: ear(0:ne), param(npm), photar(ne), photer(ne)
  REAL(kind=8) :: photar_d(ne), photer_d(ne)
  character(LEN=40) :: init
  photer(1)=0.0

  call lmodxillvernthcomp(1.0d0*ear, ne, 1.0d0*param, ifl, photar_d, photer_d, init)

  DO i=1,ne
     photar(i)=real(photar_d(i))
  ENDDO

  return
END SUBROUTINE f_rxillvernthcomp


SUBROUTINE f_rxillverD(ear, ne, param, ifl, photar, photer)
  use iso_c_binding
  IMPLICIT NONE

  interface c_interface

     subroutine lmodxillverdens (ear_d, ne, param_d, ifl, photar_d, photer_d, init) &
          bind (C, name = "lmodxillverdens")

       use iso_c_binding
       implicit none

       real (kind = c_double), intent (in), dimension (*) :: ear_d
       integer (kind=c_int), intent (in),VALUE :: ne
       integer (kind=c_int), intent (in),VALUE :: ifl
       character (kind=c_char), intent (in), dimension (*) :: init
       real (kind=c_double), intent (in),dimension (*) :: param_d
       real (kind=c_double), intent (out), dimension (*) :: photar_d
       real (kind=c_double), intent (out), dimension (*) :: photer_d

     end subroutine lmodxillverdens

  end interface c_interface

  INTEGER :: ne, ifl,i
  INTEGER,PARAMETER :: npm=7
  REAL :: ear(0:ne), param(npm), photar(ne), photer(ne)
  REAL(kind=8) :: photar_d(ne), photer_d(ne)
  character(LEN=40) :: init
  photer(1)=0.0

  call lmodxillverdens(1.0d0*ear, ne, 1.0d0*param, ifl, photar_d, photer_d, init)

  DO i=1,ne
     photar(i)=real(photar_d(i))
  ENDDO

  return
END SUBROUTINE f_rxillverD


SUBROUTINE f_rebin_spectrum(en_out,nbn_out,en_in,nbn_in,flu_out,flu_in)
!~~~This subroutine uses the original rebin_spectrum C-function, taken~~~!
!~~~from the RELXILL 1.2.0. This function is defined in the xilutility.c~!
!~~~file, which is modified version of the original RELXILL relutility.c~!
!~~~file.~~~!

  use iso_c_binding
  IMPLICIT NONE

  interface c_interface

     subroutine rebin_spectrum(en_out, flu_out, nbn_out, en_in, flu_in, nbn_in) &
          bind (C, name = "rebin_spectrum")

       use iso_c_binding
       implicit none

       real (kind = c_double), intent (in), dimension (*) :: en_out
       real (kind = c_double), intent (out), dimension (*) :: flu_out
       integer (kind=c_int), intent (in),VALUE :: nbn_out

       real (kind = c_double), intent (in), dimension (*) :: en_in
       real (kind = c_double), intent (in), dimension (*) :: flu_in
       integer (kind=c_int), intent (in),VALUE :: nbn_in

     end subroutine rebin_spectrum

  end interface c_interface

  INTEGER :: nbn_in,nbn_out
  REAL(kind=8) :: en_in(0:nbn_in),en_out(0:nbn_out),flu_in(0:nbn_in),flu_out(0:nbn_out)

  call rebin_spectrum(en_out, flu_out, nbn_out, en_in, flu_in, nbn_in)
  
  return
END SUBROUTINE f_rebin_spectrum


SUBROUTINE f_binary_search_float(ener,k,ear,ne)
!~~~This subroutine uses the original binary_search_float C-function,~~~!
!~~~taken from the RELXILL 1.2.0. This function is defined in the xilutility.c~!
!~~~file, which is modified version of the original RELXILL relutility.c~!
!~~~file.~~~!
  use iso_c_binding
  IMPLICIT NONE

  interface c_interface

     integer(kind=c_int) function binary_search_float(ear_c, ne, ener_c) &
          bind (C, name = "binary_search_float")

       use iso_c_binding
       implicit none

       real (kind = c_float), intent (in), dimension (*) :: ear_c
       integer (kind=c_int), intent (in),VALUE :: ne
       real (kind = c_float), intent (in),VALUE :: ener_c

     end function binary_search_float

  end interface c_interface

  INTEGER :: ne,k
  REAL(kind=8) :: ener,ear(0:ne)
  k=binary_search_float(real(ear), ne, real(ener))
  k=k+1
  
  return
END SUBROUTINE f_binary_search_float
