! This file is part of the REFLKERR model. 
! You can redistribute it.
! Please refer to Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942                                                                                                                   
! This file (xreflkerr.f90) contains relativistic lamppost model routines, functions and variables

MODULE par_rel
  USE par_global
  IMPLICIT NONE
  SAVE

  real, parameter :: keV2erg = 1.6022e-9
  double precision, parameter :: kpc2cm = 3.08567758e21
  double precision, parameter :: c = 2.99792458e+10
  double precision, parameter :: Gc = 6.67259858585e-8
  double precision, parameter :: Msun = 1.9891e33
  double precision, parameter :: sigma = 5.6704e-5 
  double precision, parameter :: k_B = 1.38066e-16
  double precision, parameter :: h_P = 6.626068e-27
  
  character ftff*120
  logical :: cached_tbtff = .FALSE.
  character ftflp*120
  logical :: cached_tbtflp = .FALSE.
  character ftflp2*120
  logical :: cached_tbtflp2 = .FALSE.

  integer,parameter :: ns=12 !spin bins
  integer,parameter :: nh=49 !h bins
  integer,parameter :: nr_gr=49  ! gr: 1-49
  integer,parameter :: nr_ngr=51 ! ngr: 50-100
  integer,parameter :: nr=100 !r bins tot
  integer,parameter :: ng=1700 !g bins
  integer,parameter :: ncobs=100
  integer,parameter :: ncem=10
  integer,parameter :: nrin=100
  integer,parameter :: nint=2
  !  integer,parameter :: nntc=6
  integer,parameter :: nntc=3
  integer,parameter :: nisl=2

  integer :: mumin_d(ns,nr_gr,ncobs),mumax_d(ns,nr_gr,ncobs)
  integer :: mumin_m(ns,nr_gr,ncobs),mumax_m(ns,nr_gr,ncobs)
  integer :: mumin_u(ns,nr_gr,ncobs),mumax_u(ns,nr_gr,ncobs)
  integer :: mumin_ngr_d(nr_ngr,ncobs),mumax_ngr_d(nr_ngr,ncobs)
  integer :: mumin_ngr_m(nr_ngr,ncobs),mumax_ngr_m(nr_ngr,ncobs)
  integer :: mumin_ngr_u(nr_ngr,ncobs),mumax_ngr_u(nr_ngr,ncobs)
  integer :: ntc_d(ns,nr_gr,ncem,ncobs),ntc1_d(ns,nisl,nr_gr,ncem,ncobs,nntc)
  integer :: ntc_m(ns,nr_gr,ncem,ncobs),ntc1_m(ns,nisl,nr_gr,ncem,ncobs,nntc)
  integer :: ntc_u(ns,nr_gr,ncem,ncobs),ntc1_u(ns,nisl,nr_gr,ncem,ncobs,nntc)
  integer :: ntc_ngr_d(nr_ngr,ncem,ncobs),ntc1_ngr_d(nisl,nr_ngr,ncem,ncobs,nntc)
  integer :: ntc_ngr_m(nr_ngr,ncem,ncobs),ntc1_ngr_m(nisl,nr_ngr,ncem,ncobs,nntc)
  integer :: ntc_ngr_u(nr_ngr,ncem,ncobs),ntc1_ngr_u(nisl,nr_ngr,ncem,ncobs,nntc)
  integer :: tfnel_d(ns),tfnel_ngr_d,tfnel_tot_d
  integer :: tfnel_m(ns),tfnel_ngr_m,tfnel_tot_m
  integer :: tfnel_u(ns),tfnel_ngr_u,tfnel_tot_u
  REAL,allocatable :: tft_d(:),tft_ngr_d(:)
  REAL,allocatable :: tft_m(:),tft_ngr_m(:)
  REAL,allocatable :: tft_u(:),tft_ngr_u(:)
  integer :: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u

  character spd(ns)*5

  real, parameter :: rmin=1.237
  real, parameter :: rmax=1000.0

  REAL :: r_in(nr),r_out(nr),hi(nh),sa(ns)
  INTEGER:: ihs(ns)
  INTEGER :: sn=-1
  LOGICAL :: sn_cached=.FALSE.

  logical :: cached_sp_arr = .FALSE.
  logical :: cached_h_arr = .FALSE.
  logical :: cached_r_arr = .FALSE.
  logical :: cached_g_arr = .FALSE.

  integer :: mul,muh,gl,gh
  REAL :: gl_val,gh_val
  logical :: cached_sind_tbl = .FALSE.
  REAL :: gr(nint,nh,nr),fphr(nint,nh,nr),fphd(nint,nh)
  REAL :: gri(nint,nr),fphri(nint,nr),fphdi(nint)
  REAL :: fphr2(nint,nh,nrin,nr),fphd2(nint,nh,nrin,ncobs),fphd2_av(nint),fphd2_av_int
  REAL :: fphr2i(nint,nr),fphd2i(nint)
  REAL :: fphr_tot(nint,nr),fphr_tot_1L(nint,nr),fphd_tot(nint),fphd_tot_1L(nint),fphdin,fphdin_1L,grd
  REAL :: fphr_tot_int(nr),fphr_tot_1L_int(nr),gri_int(nr)
  REAL :: wgt(nint,nint,ncem,nr)
  INTEGER, ALLOCATABLE :: muem_app(:,:,:)
  logical :: was_wgt_invoked = .FALSE.

  INTEGER, ALLOCATABLE:: ist(:),irt(:),jt(:),ist1(:),irt1(:),jt1(:)
  REAL, ALLOCATABLE:: istr(:)
  INTEGER :: cached_lp_sn = -1
  INTEGER :: cached_lp2_sn = -1

  REAL :: rH, ra, rIncl, rRin, rRout, rTauGam, rGamma, rAfe, rkTe, rEcut, rkTbb, rLogXi, rRMtype, rAcc, rlp2ext
  REAL :: rGeom,rLogN,ruse_Ec,rRedshft,rRefl_frac
  REAL :: rLineE,rLimb
  REAL :: rMbh_bb=-1.,rD_bb=-1.,ralb_bb=-1.,rfcol_bb=-1.,rdmu_bb=-1.,rf_bb=-1.,rTherm_frac_bb=-1.,rNORM_bb=-1.
  REAL :: rTherm_frac_bb_int=-1.
  double precision :: L_int,L_intr_obs,L_intr_loc

  REAL, ALLOCATABLE :: rel_ldat_r(:,:,:,:), rel_ldat_r_final(:,:,:,:) 
  double precision, ALLOCATABLE :: conv_dat(:,:,:,:)
  double precision, ALLOCATABLE :: conv_int_dat(:,:)
  
  INTEGER :: inbn_app
  double precision, ALLOCATABLE :: inen_app(:)
  LOGICAL :: show_all_av=.FALSE.

  INTEGER :: map_a_r_to_g(nr,nint,ncem),ngmap
  REAL, ALLOCATABLE :: rel_ldat_r_maped(:,:)
  double precision, ALLOCATABLE :: conv_dat_11(:,:),conv_dat_12(:,:),conv_dat_21(:,:),conv_dat_22(:,:)

END MODULE par_rel


SUBROUTINE f_reflkerrlp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,ruse_Ec
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 16
  integer,parameter :: num_lparam = 10
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrlp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.

  IF(abs(ruse_Ec).gt.0.0)THEN
     mtp=9
  ELSE
     mtp=1
  ENDIF

  rRMtype=1.0
  RMtype=1.0

  if((abs(rGeom).gt.1.0e-22).and.(abs(rGeom).lt.4.0)) rGeom=0.0               

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rkTbb
  lparam(8)=rGeom
  lparam(9)=ruse_Ec
  lparam(10)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrlp


SUBROUTINE f_reflkerrGlp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,ruse_Ec
  use par_xilcps, only: RMtype,pkTbb
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 14
  integer,parameter :: num_lparam = 9
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrGlp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.

  IF(abs(ruse_Ec).gt.0.0)THEN
     mtp=10
  ELSE
     mtp=2
  ENDIF

  rGeom=0.0
  rRMtype=1.0
  RMtype=rRMtype
  rkTbb=1.0e-3
  pkTbb=rkTbb

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rGeom
  lparam(8)=ruse_Ec
  lparam(9)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrGlp


SUBROUTINE f_reflkerrDlp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rLogN
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 16
  integer,parameter :: num_lparam = 10
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrDlp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.
  
  mtp=5
  rRMtype=1.0
  RMtype=1.0

  if((abs(rGeom).gt.1.0e-22).and.(abs(rGeom).lt.4.0)) rGeom=0.0               

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=rkTbb
  lparam(9)=rGeom
  lparam(10)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrDlp


SUBROUTINE f_reflkerrDGlp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rLogN
  use par_xilcps, only: RMtype,pkTbb
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 14
  integer,parameter :: num_lparam = 9
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrDGlp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.
  
  mtp=6
  rRMtype=1.0
  RMtype=rRMtype
  rkTbb=1.0e-3
  pkTbb=rkTbb
  rGeom=0.0

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=rGeom
  lparam(9)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrDGlp


SUBROUTINE f_reflkerrexplp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rGamma,rRMtype,rAfe,rEcut,rLogXi,rIncl
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 13
  integer,parameter :: num_lparam = 7
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrexplp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.

  mtp=3
  rRMtype=1.0
  RMtype=1.0

  lparam(1)=rGamma
  lparam(2)=rAfe
  lparam(3)=rEcut
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrexplp 


SUBROUTINE f_reflkerrexpDlp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rGamma,rRMtype,rAfe,rEcut,rLogXi,rIncl,rLogN
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 14
  integer,parameter :: num_lparam = 8
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrexpDlp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.

  mtp=7
  rRMtype=1.0
  RMtype=1.0

  lparam(1)=rGamma
  lparam(2)=rAfe
  lparam(3)=rEcut
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrexpDlp


SUBROUTINE f_reflkerrlinelp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rLineE,rAcc,rRefl_frac
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 10
  integer,parameter :: num_lparam = 2
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrlinelp(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.FALSE.

  mtp=4
  rAcc=4.0
  rRefl_frac=-1.0
  
  lparam(1)=rLineE
  lparam(2)=0.0

  call main_lp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrlinelp

SUBROUTINE f_reflkerrlpbb(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,ruse_Ec
  use par_xilcps, only: RMtype

  use par_global,only : rchatty
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 24
  integer,parameter :: num_lparam = 10
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)
  INTEGER :: i
  INTEGER,SAVE :: k=1
  REAL :: par_1
  
  ifl=0
  photer(1)=0.0

  call init_par_reflkerrlpbb(param,num_param)

  is_lamppost=.TRUE.
  is_lamppost_bb=.TRUE.

  IF(abs(ruse_Ec).gt.0.0)THEN
     mtp=9
  ELSE
     mtp=1
  ENDIF

  rRMtype=1.0
  RMtype=1.0

  if((abs(rGeom).gt.1.0e-22).and.(abs(rGeom).lt.4.0)) rGeom=0.0               

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rkTbb
  lparam(8)=rGeom
  lparam(9)=ruse_Ec
  lparam(10)=0.0

  IF(rchatty) THEN
     write(*,*)"Iter:",k ,"################################### DUMP START ###################################"
     write(*,*)is_lamppost,is_lamppost_bb
     do i=1,num_param
        write(*,*)i,param(i)
     enddo
  ENDIF
  call main_lp_model(ear, ne, lparam, num_lparam, photar)
  IF(rchatty)THEN
     write(*,*)"Iter:",k ,"#################################### DUMP END ####################################"
  ENDIF
  
  IF((photar(1).ne.photar(1)))THEN
     par_1=lparam(3)
     write(*,*)"Iter:",k ,"############################### ERROR DUMP START ###############################"
     write(*,*)is_lamppost,is_lamppost_bb
     do i=1,num_param
        write(*,*)i,param(i)
     enddo
     lparam(3)=0.8*par_1
     call FPMSTR("RCHATTY", "0")
     rchatty=.FALSE.
     call main_lp_model(ear, ne, lparam, num_lparam, photar)
     lparam(3)=par_1
     call FPMSTR("RCHATTY", "1")
     call main_lp_model(ear, ne, lparam, num_lparam, photar)
     call FPMSTR("RCHATTY", "0")
     write(*,*)"Iter:",k ,"############################### ERROR DUMP END ###############################"
  ENDIF
  k=k+1
  
  RETURN
END SUBROUTINE f_reflkerrlpbb

SUBROUTINE main_lp_model(ear, ne, lparam, num_lparam, photar)
  USE par_global, only: Pi,mtp,inbn,inen,inenl,inenh,ginen,inphotar_p,inphotar_r,gshft,rchatty,is_lamppost_bb
  use par_rel, only: keV2erg,kpc2cm,c,Gc,sigma,Msun,h_P
  use par_rel, only: rH,ra,rIncl,rRin,rRout,rAcc,rlp2ext,rGamma,rLimb,rmin,rmax,sa,ns,sn
  use par_rel, only: rRedshft,rRefl_frac,nint,nr,gri,fphri,fphdi,fphr2i,fphd2i,fphr_tot,fphd_tot,fphdin
  use par_rel, only: fphd_tot_1L,fphdin_1L
  use par_rel, only: ncem,ncobs,wgt,muem_app,rel_ldat_r,rel_ldat_r_final
  use par_rel, only: ist,irt,jt,ist1,irt1,jt1,istr,gri,grd,mul,muh
  use par_rel, only: was_wgt_invoked,ng,conv_int_dat,inbn_app,inen_app
  use par_rel, only: rNORM_bb,rD_bb,rMbh_bb,rfcol_bb,rdmu_bb,ralb_bb,rf_bb,rTherm_frac_bb,r_in,r_out
  use par_rel, only: fphr_tot_int, gri_int,L_int,L_intr_obs,L_intr_loc,rTherm_frac_bb_int
  use par_rel, only: show_all_av,fphd2_av,fphd2_av_int,fphr_tot_1L,fphr_tot_1L_int
  use par_rel, only: map_a_r_to_g,ngmap,rel_ldat_r_maped
  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u
  !1  

  IMPLICIT NONE

  INTEGER :: ne, num_lparam, i, j, k
  REAL :: ear(0:ne), photar(ne)
  REAL :: lparam(num_lparam)

  double precision :: inphotar_p_tmp(inbn),rel_inprimary(inbn),dp_infinal(inbn)
  double precision :: dp_photar(ne),dp_photar_tmp(ne),rel_photarp(ne)
  REAL, ALLOCATABLE :: ginear(:),gphotar(:)
  REAL :: g_inphotar(inbn)

  INTEGER :: get_low_rbin,irl
  INTEGER :: get_high_rbin,irh
  INTEGER :: get_low_h_bin,ihl
  INTEGER :: get_low_inc_bin,iml
  REAL :: rms,rmsv,rhor,Rin_g, Rout_g,H_g,redshft_scl,dum
  INTEGER :: ir,is,ig,iem,get_nbg,nbg,ntot,gnbn

  LOGICAL :: do_recalculate_local_models,elim_changed,elimc_val,local_models_recalc
  REAL, ALLOCATABLE, SAVE :: rel_ldat_r_tmp(:,:,:,:)
  REAL :: get_fphd_spin_int,get_fphd_spin_int_1L,gdcal,rprim_norm

  REAL,SAVE :: eminc = 1.0e24, emaxc = -1.0e24
  REAL :: ear_l, ear_h
  REAL, SAVE :: cached_rGamma=-1.0,cached_rLimb=-1.,cached_rRedshft=-1.0
  REAL :: mu,muem,get_limbing, gline_scl, mu_limb_scl
  character FGMSTR*20, FGMSTR_val*10
  character, SAVE :: cached_FGMSTR_val*10=''
  LOGICAL::show_all,LgRin,LgRout
  REAL :: MBh,Dist,Norm,get_energy_flux,loc_flux_comp,loc_flux_comp_511
  REAL :: gamma_comp,flux_dum,Lcorr,dl,compact,compact_511

  character FGMSTR_chat_val*20
  CHARACTER(LEN=20)::FGMSTR_lgrin_val,FGMSTR_lgrout_val
  LOGICAL,SAVE :: message_diplayed=.FALSE.

  double precision :: Md,FK,Planck
  double precision :: L_irr(nr),Q1(nr),Q2(nr),Teff(nr)
  double precision, allocatable :: Nph_bb(:,:,:),dum_d
  REAL::rprim_norm_bb,rNORM_bb_int,spinl,spinh,factor
  double precision :: scale_bb,fac_1,fac_2
  LOGICAL :: bb_calc,do_recalculate_lpbb,lpbb_recalc
  double precision,SAVE :: eps_bb=-1.0
  REAL :: eta_bb_1,eta_bb_2
  REAL :: n_irr_1L,n_irr_2L
  
  INTEGER :: igmap
  INTEGER, ALLOCATABLE, SAVE :: map_a_r_to_g_tmp(:,:,:)

  IF(.not.message_diplayed)THEN
     write(*,"(A59)")"***********************************************************"
     write(*,"(A59)")"*    Relativistic reflection in the lamppost geometry.    *"
     write(*,"(A59)")"*    If you use results of this model, please refer to    *"
     write(*,"(A59)")"* Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942 *"
     write(*,"(A59)")"***********************************************************"
     message_diplayed=.TRUE.
  ENDIF

  FGMSTR_chat_val= trim(FGMSTR("RCHATTY"))

  if (len_trim(FGMSTR_chat_val).gt.0) then
     read (FGMSTR_chat_val,*) i
     if(i.eq.1)then
        rchatty=.TRUE.
     else
        rchatty=.FALSE.
     endif
  else
     rchatty=.FALSE.
  endif

  FGMSTR_val= trim(FGMSTR("CONV_ACC"))

  FGMSTR_lgrin_val= trim(FGMSTR("RINLOG"))
  if (len_trim(FGMSTR_lgrin_val).gt.0) then
     read (FGMSTR_lgrin_val,*) i
     if(i.eq.1)then
        LgRin=.TRUE.
     else
        LgRin=.FALSE.
     endif
  else
     LgRin=.FALSE.
  endif

  FGMSTR_lgrout_val= trim(FGMSTR("ROUTLOG"))
  if (len_trim(FGMSTR_lgrout_val).gt.0) then
     read (FGMSTR_lgrout_val,*) i
     if(i.eq.1)then
        LgRout=.TRUE.
     else
        LgRout=.FALSE.
     endif
  else
     LgRout=.FALSE.
  endif


  call DEFINE_SPIN_ARRAY()

  if(ra.lt.sa(1)) ra=sa(1)
  if(ra.gt.sa(ns)) ra=sa(ns)

  nbg=get_nbg(rAcc)

  call check_table_paths()

  call INIT_TFF_TABLES_PATHS()
  call INIT_LP_TABLES_PATHS()
  call INIT_LP2_TABLES_PATHS()

  call DEFINE_H_ARRAY()
  call DEFINE_R_ARRAY()
  call LOAD_SIND_TABLES()

  call set_spin_bin()

  call LOAD_LP_TF()

  IF(rlp2ext.gt.0)THEN
     call LOAD_LP2_TF()
  ENDIF

  rmsv=rms(ra)
  if(.not.LgRin)then
     IF(rRin.lt.0.0)THEN
        Rin_g=abs(rRin)*rmsv
       ELSE
        Rin_g=rRin
     ENDIF
  else
     IF(rRin.le.0.0)THEN
        Rin_g=rmsv*10.0**(abs(rRin))
     ELSE
        Rin_g=10.0**rRin
     ENDIF
  endif
  if(Rin_g.lt.rmin)Rin_g=rmin
  if(Rin_g.gt.rmax)Rin_g=rmax

  if(.not.LgRout)then
     IF(rRout.lt.0.0)THEN
        Rout_g=abs(rRout)*rmsv
     ELSE
        Rout_g=rRout
     ENDIF
  else
     IF(rRout.le.0.0)THEN
        Rout_g=rmsv*10.0**(abs(rRout))
     ELSE
        Rout_g=10.0**rRout
     ENDIF
  endif
  if(Rout_g.lt.rmin)Rout_g=rmin
  if(Rout_g.gt.rmax)Rout_g=rmax


  IF(rH.lt.0.0)THEN
     H_g=abs(rH)*rhor(ra)
  ELSE
     H_g=rH
  ENDIF

  irl=get_low_rbin(Rin_g)
  irh=get_high_rbin(Rout_g)
  ihl=get_low_h_bin(H_g)
  iml=get_low_inc_bin(rIncl)

  call INTERPOLATE_LP_TF(H_g,ihl)

  IF(abs(rTherm_frac_bb).gt.0.0)then
     bb_calc=.TRUE.
  ELSE
     bb_calc=.FALSE.
  ENDIF

  DO is=1,nint
     fphd_tot(is)=fphdi(is)
     do ir=1,nr
        fphr_tot(is,ir)=fphri(is,ir)
     enddo
  ENDDO

  call READ_XSET_SHOW_ALL_FLAG(show_all)
  show_all_av=show_all
  

  if(show_all)then
     DO is=1,nint
        fphd_tot_1L(is)=fphdi(is)
        do ir=1,nr
           fphr_tot_1L(is,ir)=fphri(is,ir)
        enddo
     ENDDO
  endif

  IF(rlp2ext.gt.0)THEN
     call INTERPOLATE_LP2_TF(H_g,Rin_g,ihl,irl,iml)
     DO is=1,nint
        fphd_tot(is)=fphd_tot(is)+rlp2ext*fphd2i(is)
        do ir=1,nr
           fphr_tot(is,ir)=fphr_tot(is,ir)+rlp2ext*fphr2i(is,ir)
        enddo
     ENDDO
  ENDIF

  if(show_all)then
     spinl=sa(sn+1)
     spinh=sa(sn+2)
     factor = (ra-spinl)/(spinh-spinl)
     fphd2_av_int=(1.0-factor)*fphd2_av(1)+factor*fphd2_av(2)
     do ir=1,nr
        fphr_tot_1L_int(ir)=(1.0-factor)*fphr_tot_1L(1,ir)+factor*fphr_tot_1L(2,ir)
     enddo
  endif

  IF(show_all.or.(is_lamppost_bb.and.bb_calc))THEN
     spinl=sa(sn+1)
     spinh=sa(sn+2)
     factor = (ra-spinl)/(spinh-spinl)
     do ir=1,nr
        fphr_tot_int(ir)=(1.0-factor)*fphr_tot(1,ir)+factor*fphr_tot(2,ir)
        gri_int(ir)=(1.0-factor)*gri(1,ir)+factor*gri(2,ir)
     enddo
  ENDIF

  fphdin=get_fphd_spin_int()
  grd=gdcal(H_g,ra)
  
  redshft_scl=(1.0/(1.0+rRedshft))
  gshft=grd*redshft_scl

  IF(rchatty)write(*,*)"Primary Redshift=",grd

  call CALCULATE_WGT_ARRAY(Rin_g,Rout_g,iml,irl,irh,rAcc,wgt)

  call GET_APP_ANGLES_BINS(nbg,wgt)

  IF(mtp.ne.4)call GET_LMOD_SERIAL_MATRIX(nbg)

  IF((mtp.eq.1).or.(mtp.eq.9)) THEN
     call init_par_hrefl(lparam,num_lparam)
  ELSE IF((mtp.eq.2).or.(mtp.eq.10)) THEN
     call init_par_hreflG(lparam,num_lparam)
  ELSE IF(mtp.eq.3) THEN
     call init_par_hreflexp(lparam,num_lparam)
  ELSE IF(mtp.eq.4) THEN
     call init_par_xline(lparam,num_lparam)
  ELSE IF(mtp.eq.5) THEN
     call init_par_hreflD(lparam,num_lparam)
  ELSE IF(mtp.eq.6) THEN
     call init_par_hreflDG(lparam,num_lparam)
  ELSE
     call init_par_hreflexpD(lparam,num_lparam)
  ENDIF

  call get_primary()

  IF(show_all)THEN
     call f_binary_search_float(1.d0,k,1.d0*inen,inbn+1)
     rprim_norm=real(inphotar_p(k)/(inen(k)-inen(k-1)))

     do i=1,inbn
        inphotar_p_tmp(i)=inphotar_p(i)/rprim_norm
     enddo
     call get_slope(1.0,20.0,real(inphotar_p_tmp),real(inen),inbn,gamma_comp,flux_dum)
     
     fphdin_1L=get_fphd_spin_int_1L()

     n_irr_1L=0.
     n_irr_2L=0.
     DO ir=irl,irh
        n_irr_1L=n_irr_1L+fphr_tot_int(ir)
        n_irr_2L=n_irr_2L+(fphr_tot_int(ir)-fphr_tot_1L_int(ir))
     ENDDO
     
     write(*,"(A12,f5.3)") "    gamma = ", gamma_comp
     write(*,"(A10,f8.3,A6)") "    Rin = ", Rin_g," [r_g]"
     write(*,"(A71,e9.3)") "    light-bending reduction of the direct radiation from the top lamp: ",2.0*fphdin_1L
     write(*,"(A58,e9.3)") "    enhancement due to contribution from the bottom lamp: ",fphdin/fphdin_1L
     write(*,"(A40,e9.3)") "    fraction of photons captured by BH: ", 1.0-n_irr_1L-n_irr_2L-fphdin_1L-fphd2_av_int
     write(*,"(A48,e9.3)") "    disk irradiating to observed photons ratio: ", (n_irr_1L+n_irr_2L)/fphdin
     write(*,*) " "

     IF(.not.is_lamppost_bb)THEN
        call READ_XSET_COMPACT_PARAM(MBh,Dist,Norm)
        IF((MBh.lt.1.0e-22).and.(Dist.lt.1.0e-22).and.(Norm.lt.1.0e-22))THEN
           write(*,"(A45)") "To get the compactness value, please provide:"
           write(*,"(A36)") "    -black hole mass in M_sun units;"
           write(*,"(A20)") "    -distance in pc;"
           write(*,"(A39)") "    -normalization factor of the model,"
           write(*,"(A35)") "by enetring the following commands:"
           write(*,"(A23)") "    'xset RMBH <value>'"
           write(*,"(A24)") "    'xset RDIST <value>'"
           write(*,"(A24)") "    'xset RNORM <value>'"
           write(*,*) " "
        ELSE IF((MBh.gt.1.0e-22).and.(Dist.gt.1.0e-22).and.(Norm.gt.1.0e-22))THEN
           loc_flux_comp= &
                get_energy_flux(inphotar_p_tmp,1.d0*inen,inbn,2.0d0,0.99d0*inenh) &
                *keV2erg
           loc_flux_comp_511= &
                get_energy_flux(inphotar_p_tmp,1.d0*inen,inbn,511.d0,0.99d0*inenh) &
                *keV2erg

           Lcorr=1.0/fphdin/gshft
           dl=(H_g-rhor(ra))/2.0
           compact=(2.00061e4)*(Dist**2.0)*loc_flux_comp*Lcorr/(dl*MBh)*Norm/2.0
           compact_511=(2.00061e4)*(Dist**2.0)*loc_flux_comp_511*Lcorr/(dl*MBh)*Norm/2.0
           write(*,"(A18,e9.3)") "    Compactness = ",compact
           write(*,"(A28,e9.3)") "    Compactness(>511 keV) = ",compact_511
           write(*,"(A6)") "where:"
           write(*,"(A10,e8.2,A8)") "    MBh = ",MBh," [M_sun]"
           write(*,"(A11,e8.2,A5)") "    Dist = ",Dist," [pc]"
           write(*,"(A11,e8.2)") "    Norm = ",Norm
        ENDIF
     ELSE
        loc_flux_comp= &
             get_energy_flux(inphotar_p_tmp,1.d0*inen,inbn,2.0d0,0.99d0*inenh) &
             *keV2erg
        loc_flux_comp_511= &
             get_energy_flux(inphotar_p_tmp,1.d0*inen,inbn,511.d0,0.99d0*inenh) &
             *keV2erg

        Lcorr=1.0/fphdin/gshft
        dl=(H_g-rhor(ra))/2.0
        Dist=1.0e3*rD_bb
        Mbh=rMbh_bb
        Norm=rNorm_bb

        compact=(2.00061e4)*(Dist**2.0)*loc_flux_comp*Lcorr/(dl*MBh)*Norm/2.0
        compact_511=(2.00061e4)*(Dist**2.0)*loc_flux_comp_511*Lcorr/(dl*MBh)*Norm/2.0
        write(*,"(A18,e9.3)") "    Compactness = ",compact
        write(*,"(A28,e9.3)") "    Compactness(>511 keV) = ",compact_511
     ENDIF

  ENDIF

  do i=1,inbn
     inphotar_p_tmp(i)=inphotar_p(i)
  enddo

  call shift_photar(real(inen),real(inphotar_p_tmp),inbn,ginen,g_inphotar,gshft)
  call f_rebin_spectrum(inen,inbn,1.d0*ginen,inbn,rel_inprimary,1.0d0*g_inphotar)

  call f_binary_search_float(1.d0,k,1.d0*inen,inbn+1)
  rprim_norm=real(rel_inprimary(k)/(inen(k)-inen(k-1)))
  IF(is_lamppost_bb.and.bb_calc)rprim_norm_bb=rprim_norm
  do i=1,inbn
     rel_inprimary(i)=rel_inprimary(i)/rprim_norm
  enddo

  call f_rebin_spectrum(1.d0*ear,ne,inen,inbn,rel_photarp,rel_inprimary)

  rprim_norm=rprim_norm*fphdin/ncobs
  ntot=ncem/nbg*nr*nint

  if(.not.allocated(rel_ldat_r)) allocate(rel_ldat_r(nint,nr,ncem,inbn))
  if(.not.allocated(rel_ldat_r_final)) allocate(rel_ldat_r_final(nint,nr,ncem,inbn))

  if(mtp.ne.4)then

     local_models_recalc=.FALSE.
     IF (do_recalculate_local_models(lparam, num_lparam,mtp))THEN
        if(allocated(rel_ldat_r_tmp))deallocate(rel_ldat_r_tmp)
        allocate(rel_ldat_r_tmp(nint,nr,ncem/nbg,inbn))
        if(allocated(map_a_r_to_g_tmp))deallocate(map_a_r_to_g_tmp)
        allocate(map_a_r_to_g_tmp(nr,nint,ncem/nbg))
        
        i=1
        DO WHILE(istr(i).lt.0.0)
           gshft=gri(ist(i),irt(i))
           lparam(6)=real(ACOS((muem_app(ist(i),irt(i),jt(i))-0.5)/ncem)*180./Pi)
           call shift_photar(real(inen),real(inphotar_p_tmp),inbn,ginen,g_inphotar,gshft)
           call f_rebin_spectrum(inen,inbn,1.d0*ginen,inbn,dp_infinal,1.0d0*g_inphotar)
           do j=1,inbn
              inphotar_p(j)=real(dp_infinal(j))
           enddo

           IF((mtp.eq.1).or.(mtp.eq.9)) THEN
              call init_par_hrefl(lparam,num_lparam)
           ELSE IF((mtp.eq.2).or.(mtp.eq.10)) THEN
              call init_par_hreflG(lparam,num_lparam)
           ELSE IF(mtp.eq.3) THEN
              call init_par_hreflexp(lparam,num_lparam)
           ELSE IF(mtp.eq.4) THEN
              call init_par_xline(lparam,num_lparam)
           ELSE IF(mtp.eq.5) THEN
              call init_par_hreflD(lparam,num_lparam)
           ELSE IF(mtp.eq.6) THEN
              call init_par_hreflDG(lparam,num_lparam)
           ELSE
              call init_par_hreflexpD(lparam,num_lparam)
           ENDIF

           call get_reflection()
           do j=1,inbn
              rel_ldat_r_tmp(ist(i),irt(i),jt(i),j)=inphotar_r(j)
           enddo
           map_a_r_to_g_tmp(irt(i),ist(i),jt(i))=i
           i=i+1
        ENDDO
        
        ngmap=i-1
        if(allocated(rel_ldat_r_maped))deallocate(rel_ldat_r_maped)
        allocate(rel_ldat_r_maped(0:ngmap,inbn))

        DO igmap=1,ngmap
           do j=1,inbn
              rel_ldat_r_maped(igmap,j)=rel_ldat_r_tmp(ist(igmap),irt(igmap),jt(igmap),j)
           enddo
        ENDDO
       
        DO WHILE(istr(i).eq.0.0)
           do j=1,inbn
              rel_ldat_r_tmp(ist1(i),irt1(i),jt1(i),j)=0.0
           enddo
           map_a_r_to_g_tmp(irt1(i),ist1(i),jt1(i))=0
           i=i+1
        ENDDO
        rel_ldat_r_maped(0,:)=0.0
        
        DO j=i,ntot
           do k=1,inbn
              rel_ldat_r_tmp(ist1(j),irt1(j),jt1(j),k)=rel_ldat_r_tmp(ist(j),irt(j),jt(j),k)
           enddo
           map_a_r_to_g_tmp(irt1(j),ist1(j),jt1(j))=map_a_r_to_g_tmp(irt(j),ist(j),jt(j))
       ENDDO
       
       DO is=1,nint
          do ir=irl,irh
             iem=1
             i=1
             do while(iem.le.ncem)
                do k=1,inbn
                   rel_ldat_r(is,ir,iem,k)=rel_ldat_r_tmp(is,ir,i,k)
                enddo
                map_a_r_to_g(ir,is,iem)=map_a_r_to_g_tmp(ir,is,i)
                if((mod(iem,nbg).eq.0.0))then
                   i=i+1
                endif
                iem=iem+1
             enddo
          enddo
       ENDDO

        deallocate(rel_ldat_r_tmp)

        local_models_recalc=.TRUE.
        
     ENDIF

     gshft=grd*redshft_scl
     IF(is_lamppost_bb.and.bb_calc)THEN
        loc_flux_comp= &
             get_energy_flux(inphotar_p_tmp,1.d0*inen,inbn,1.0001d0*inenl,0.9999d0*inenh)*keV2erg
        Lcorr=1.0/(fphdin)

        if(rNORM_bb.gt.0.0)then
           rNORM_bb_int=rNORM_bb
           L_int=Lcorr*2.0*Pi*(kpc2cm**2.0)*(rD_bb**2.0)*loc_flux_comp*rNORM_bb_int/rprim_norm_bb*gshft
        else
           rNORM_bb_int=-rNORM_bb
           L_int=Lcorr*2.0*Pi*(kpc2cm**2.0)*(rD_bb**2.0)*abs(rNORM_bb)
           rNORM_bb_int=1.0
        endif

        L_intr_obs=L_int
        L_intr_loc=L_int/(gshft**2.)


        if(abs(rRefl_frac).gt.0.)then
           rTherm_frac_bb_int=rTherm_frac_bb/abs(rRefl_frac)
        else
           rTherm_frac_bb_int=rTherm_frac_bb
        endif
     ENDIF

     lpbb_recalc=do_recalculate_lpbb()

     IF(is_lamppost_bb.and.bb_calc)THEN
        if(lpbb_recalc)then

           eta_bb_1=((Rin_g**(3./2.)-2.*Rin_g**(1./2.)+ra)/ &
                (Rin_g**(3./4.)*(Rin_g**(3./2.)-3.*Rin_g**(1./2.)+2.*ra)**(1./2.)))
           eta_bb_2=((rmsv**(3./2.)-2.*rmsv**(1./2.)+ra)/ &
                (rmsv**(3./4.)*(rmsv**(3./2.)-3.*rmsv**(1./2.)+2.*ra)**(1./2.)))
           eta_bb_2=eta_bb_1-eta_bb_2
           eta_bb_1=1.0-eta_bb_1
           eps_bb=rf_bb*eta_bb_1+eta_bb_2

           IF(eps_bb.gt.0.)THEN
              Md=L_intr_obs/eps_bb/(c**2.0)*2.0              
              fac_1=(3.*Md*(c**6.))/(8.*Pi*(Gc**2.)*((rMbh_bb*Msun)**2.))
           ELSE
              fac_1=0.
           ENDIF

           fac_2=(1.0-ralb_bb)/(Gc*rMbh_bb*Msun/(c**2.))**2.

           IF(show_all)THEN
              DO ir=1,nr
                 Q1(ir)=fac_1*FK(1d0*ra,1d0*sqrt(r_in(ir)*r_out(ir)))
                 L_irr(ir)=L_intr_loc*gri_int(ir)*gshft*fphr_tot_int(ir)
                 Q2(ir)=fac_2*L_irr(ir)/(Pi*(r_out(ir)**2.-r_in(ir)**2.))
              ENDDO
           ELSE
              DO ir=irl,irh
                 Q1(ir)=fac_1*FK(1d0*ra,1d0*sqrt(r_in(ir)*r_out(ir)))
                 L_irr(ir)=L_intr_loc*gri_int(ir)*gshft*fphr_tot_int(ir)
                 Q2(ir)=fac_2*L_irr(ir)/(Pi*(r_out(ir)**2.-r_in(ir)**2.))
              ENDDO
           ENDIF


           if(allocated(Nph_bb))deallocate(Nph_bb)
           allocate(Nph_bb(nr,ncem,inbn))

           if(rTherm_frac_bb_int.gt.0.0)then

              IF(show_all)THEN
                 DO ir=1,nr
                    Teff(ir)=(((1.-rf_bb)*Q1(ir)+Q2(ir))/sigma)**(.25)
                 ENDDO
              ELSE
                 DO ir=irl,irh
                    Teff(ir)=(((1.-rf_bb)*Q1(ir)+Q2(ir))/sigma)**(.25)
                 ENDDO
              ENDIF

           else
              Nph_bb=0.0
              GOTO 30
           endif
           
           IF(rNORM_bb_int.gt.0.0)THEN
              DO ir=irl,irh
                 do iem=mul,muh
                    muem=(iem-.5)/ncem
                    DO i=1,inbn
                       dum=real(sqrt(inen(i-1)*inen(i))*keV2erg)
                       dum_d=(inen(i)-inen(i-1))*keV2erg/h_P*(rprim_norm*ncobs)*rTherm_frac_bb_int
                       Nph_bb(ir,iem,i)=1./dum/rNORM_bb_int*(3.*(1.+rdmu_bb*muem)/(3.+2.*rdmu_bb))* &
                            1./(rfcol_bb**4.)*Planck(dum/h_P,rfcol_bb*Teff(ir))*dum_d
                    ENDDO
                 enddo
              ENDDO
           ELSE
              Nph_bb=0.0
           ENDIF
30 CONTINUE
           dum_d=Pi/((kpc2cm*rD_bb)**2.)*((Gc*rMbh_bb*Msun/(c**2.))**2.0)
           DO is=1,nint
              do ir=irl,irh
                 if(fphr_tot(is,ir).gt.0.0)then
                    dum=1.0/fphr_tot(is,ir)                    
                 else
                    dum=0.0
                 endif
                 scale_bb=((r_out(ir)**2.)-(r_in(ir)**2.))*dum_d*dum
                 DO iem=mul,muh
                    muem=(iem-0.5)/ncem
                    do i=1,inbn
                       rel_ldat_r_final(is,ir,iem,i)=rel_ldat_r(is,ir,iem,i)+ &
                            real(Nph_bb(ir,iem,i)*scale_bb*muem)
                    enddo
                 ENDDO
              enddo
           ENDDO
           
        endif
        
     ELSE
        rel_ldat_r_final=rel_ldat_r
     ENDIF
     
     IF(local_models_recalc.or.lpbb_recalc)local_models_recalc=.TRUE.
     
  else

     local_models_recalc=.FALSE.
     IF (do_recalculate_local_models(lparam, num_lparam,mtp) &
          .or.(cached_rGamma.ne.rGamma) &
          .or. (cached_rLimb.ne.rLimb))THEN
        DO is=1,nint
           do ir=1,nr
              gline_scl=gri(is,ir)**(rGamma-1.0)
              DO iem=1,ncem
                 mu = (iem-0.5)/ncem
                 mu_limb_scl=mu*get_limbing(rLimb,mu)
                 do i=1,inbn
                    rel_ldat_r(is,ir,iem,i)=real(gline_scl*mu_limb_scl*inphotar_p_tmp(i)/(inen(i)-inen(i-1)))
                 enddo
              ENDDO
           enddo
        ENDDO
        cached_rGamma=rGamma
        cached_rLimb=rLimb
        local_models_recalc=.TRUE.
     ENDIF

  endif

  IF(show_all.and.is_lamppost_bb.and.(rTherm_frac_bb.gt.0.).and. &
       ((ralb_bb.lt.1.0).or.(rf_bb.lt.1.0)))THEN
!     IF(show_all)THEN
        write(*,*)"Observed intrinsic luminosity=", L_intr_obs
        write(*,*)"Eta_1=",eta_bb_1
        write(*,*)"Eta_2=",eta_bb_2

!        if((ralb_bb.lt.1.).and.(rf_bb.eq.1.))then
!           write(*,*)"Temperature profile (2):"
!        else if((ralb_bb.eq.1.).and.(rf_bb.lt.1.))then
!           write(*,*)"Temperature profile (1):"
!        else
!           write(*,*)"Temperature profile (1)+(2):"
!        endif
!        DO ir=1,nr
!           write(*,*)ir, sqrt(r_in(ir)*r_out(ir)),Teff(ir)
!        ENDDO
!     ENDIF
  ENDIF

  lparam(6)=rIncl

  IF((mtp.eq.1).or.(mtp.eq.9)) THEN
     call init_par_hrefl(lparam,num_lparam)
  ELSE IF((mtp.eq.2).or.(mtp.eq.10)) THEN
     call init_par_hreflG(lparam,num_lparam)
  ELSE IF(mtp.eq.3) THEN
     call init_par_hreflexp(lparam,num_lparam)
  ELSE IF(mtp.eq.4) THEN
     call init_par_xline(lparam,num_lparam)
     GOTO 10
  ELSE IF(mtp.eq.5) THEN
     call init_par_hreflD(lparam,num_lparam)
  ELSE IF(mtp.eq.6) THEN
     call init_par_hreflDG(lparam,num_lparam)
  ELSE
     call init_par_hreflexpD(lparam,num_lparam)
  ENDIF
  
10 CONTINUE

  call READ_XSET_ELIM(ear_l,ear_h)
  IF((ear_l.lt.1e-34).or.(ear_h.lt.1e-34).or.(ear_h.le.ear_l))THEN
     ear_l=ear(0)
     ear_h=ear(ne)
     if(ne.eq.1) then
        eminc=ear_l
        emaxc=ear_h
     else
        IF(ear_l.lt.eminc)eminc=ear_l
        IF(ear_h.gt.emaxc)emaxc=ear_h
     endif
  ELSE
     eminc=ear_l
     emaxc=ear_h
  ENDIF

  elimc_val=elim_changed(eminc,emaxc)
  IF((elimc_val.or.was_wgt_invoked) &
     .or.local_models_recalc &
     .or.(cached_rRedshft.ne.rRedshft) &
     .or.(cached_FGMSTR_val.ne.FGMSTR_val))THEN
     call CONVOLVE_LP(iml,irl,irh,eminc,emaxc)
     cached_FGMSTR_val=FGMSTR_val
     cached_rRedshft=rRedshft
  ENDIF

  call SPIN_ANGLE_INTERP(iml)


  do i=1,ne
     dp_photar(i)=0.0
  enddo


  IF(gl_d.le.gh_d)THEN
     gnbn=gh_d-gl_d+1
     IF(allocated(ginear))deallocate(ginear)
     allocate(ginear(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))
     do i=1,inbn_app
        dum=real(redshft_scl*sqrt(inen_app(i-1)*inen_app(i))*0.007)
        DO ig=gl_d-1,gh_d
           ginear(ig-gl_d+1)=dum*ig
        ENDDO
        DO ig=gl_d,gh_d
           gphotar(ig-gl_d+1)=real(conv_int_dat(i,ig))*7.0
        ENDDO
        call f_rebin_spectrum(1.0d0*ear,ne,1.d0*ginear,gnbn,dp_photar_tmp,1.0d0*gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF
  IF(gl_m.le.gh_m)THEN
     gnbn=gh_m-gl_m+1
     IF(allocated(ginear))deallocate(ginear)
     allocate(ginear(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))
     do i=1,inbn_app
        dum=real(redshft_scl*sqrt(inen_app(i-1)*inen_app(i))*0.005)
        DO ig=gl_m-1,gh_m
           ginear(ig-gl_m+1)=dum*ig
        ENDDO
        DO ig=gl_m,gh_m
           gphotar(ig-gl_m+1)=real(conv_int_dat(i,ig))*5.0
        ENDDO
        call f_rebin_spectrum(1.0d0*ear,ne,1.d0*ginear,gnbn,dp_photar_tmp,1.0d0*gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF
  IF(gl_u.le.gh_u)THEN
     gnbn=gh_u-gl_u+1
     IF(allocated(ginear))deallocate(ginear)
     allocate(ginear(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))
     do i=1,inbn_app
        dum=real(redshft_scl*sqrt(inen_app(i-1)*inen_app(i))*0.001)
        DO ig=gl_u-1,gh_u
           ginear(ig-gl_u+1)=dum*ig
        ENDDO
        DO ig=gl_u,gh_u
           gphotar(ig-gl_u+1)=real(conv_int_dat(i,ig))
        ENDDO
        call f_rebin_spectrum(1.0d0*ear,ne,1.d0*ginear,gnbn,dp_photar_tmp,1.0d0*gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF


  IF(mtp.eq.4) GOTO 20

  
  DO i=1,ne
     dp_photar(i)=dp_photar(i)/rprim_norm
  ENDDO

  do i=1,inbn
     inphotar_p(i)=real(inphotar_p_tmp(i))
  enddo

20 CONTINUE
  call get_finalsp(ne, dp_photar, rel_photarp,rRefl_frac)


  do i=1,ne
     photar(i)=real(dp_photar(i))
  enddo

  deallocate(ginear)
  deallocate(gphotar)

  gshft=-1.0

return
END SUBROUTINE main_lp_model


SUBROUTINE set_spin_bin()
  use par_rel, only: ra,sa,ns,sn, sn_cached
  IMPLICIT NONE

  integer :: i

  i=1
  do while((ra.ge.sa(i)).and.(i.lt.ns))
     i=i+1
  enddo
  IF(sn.ne.i-2)THEN
     sn=i-2
     sn_cached=.FALSE.
  ELSE
     sn_cached=.TRUE.
  ENDIF

return
end SUBROUTINE set_spin_bin


SUBROUTINE INIT_PAR_REFLKERRLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rkTbb, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac,ruse_Ec,rGeom

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=      param(1)
  ra=      param(2)
  rIncl=   param(3)
  rRin=    param(4)
  rRout=   param(5)
  rTauGam= param(6)
  rAfe=    param(7)
  rkTe=    param(8)
  rkTbb=   param(9)
  rGeom=   param(10)
  rLogXi=  param(11)
  rRedshft= param(12)
  rRefl_frac= param(13)
  ruse_Ec= param(14)
  rlp2ext= param(15)
  rAcc=    param(16)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRLP


SUBROUTINE INIT_PAR_REFLKERRGLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac,ruse_Ec

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=   param(1)
  ra=   param(2)
  rIncl=   param(3)
  rRin=   param(4)
  rRout=   param(5)
  rTauGam=   param(6)
  rAfe=   param(7)
  rkTe=   param(8)
  rLogXi=   param(9)
  rRedshft=  param(10)
  rRefl_frac=param(11)
  ruse_Ec=param(12)
  rlp2ext=param(13)
  rAcc=   param(14)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRGLP


SUBROUTINE INIT_PAR_REFLKERRDLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rkTbb, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac,rGeom,rLogN

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=   param(1)
  ra=   param(2)
  rIncl=   param(3)
  rRin=   param(4)
  rRout=   param(5)
  rTauGam=   param(6)
  rAfe=   param(7)
  rkTe=   param(8)
  rkTbb=   param(9)
  rGeom=   param(10)
  rLogXi=   param(11)
  rRedshft=  param(12)
  rLogN=  param(13)
  rRefl_frac=param(14)
  rlp2ext=param(15)
  rAcc=   param(16)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRDLP


SUBROUTINE INIT_PAR_REFLKERRDGLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac,rLogN

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=   param(1)
  ra=   param(2)
  rIncl=   param(3)
  rRin=   param(4)
  rRout=   param(5)
  rTauGam=   param(6)
  rAfe=   param(7)
  rkTe=   param(8)
  rLogXi=   param(9)
  rRedshft=  param(10)
  rLogN=  param(11)
  rRefl_frac=param(12)
  rlp2ext=param(13)
  rAcc=   param(14)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRDGLP


SUBROUTINE INIT_PAR_REFLKERREXPLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rGamma, rAfe, rEcut, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=    param(1)
  ra=    param(2)
  rIncl= param(3)
  rRin=  param(4)
  rRout= param(5)
  rGamma=param(6)
  rAfe=  param(7)
  rEcut=param(8)
  rLogXi=   param(9)
  rRedshft=  param(10)
  rRefl_frac=param(11)
  rlp2ext=param(12)
  rAcc=   param(13)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERREXPLP


SUBROUTINE INIT_PAR_REFLKERREXPDLP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rGamma, rAfe, rEcut, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rLogN,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=    param(1)
  ra=    param(2)
  rIncl= param(3)
  rRin=  param(4)
  rRout= param(5)
  rGamma=param(6)
  rAfe=  param(7)
  rEcut=param(8)
  rLogXi=   param(9)
  rRedshft=  param(10)
  rLogN=   param(11)
  rRefl_frac=param(12)
  rlp2ext=param(13)
  rAcc=   param(14)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERREXPDLP


SUBROUTINE INIT_PAR_REFLKERRLINELP(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rGamma, rLineE, rLimb, rlp2ext
  use par_rel, only: rRedshft

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=    param(1)
  ra=    param(2)
  rIncl= param(3)
  rRin=  param(4)
  rRout= param(5)
  rGamma=param(6)
  rLineE=  param(7)
  rLimb=param(8)
  rlp2ext=param(9)
  rRedshft=  param(10)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRLINELP


SUBROUTINE INIT_PAR_REFLKERRLPBB(param, num_param)
  use par_rel, only: rH, ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rkTbb, rLogXi, rAcc, rlp2ext
  use par_rel, only: rRedshft,rRefl_frac,ruse_Ec,rGeom
  use par_rel, only: rMbh_bb,rD_bb,ralb_bb,rfcol_bb,rdmu_bb,rf_bb,rTherm_frac_bb,rNORM_bb

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rH=      param(1)
  ra=      param(2)
  rIncl=   param(3)
  rRin=    param(4)
  rRout=   param(5)
  rTauGam= param(6)
  rAfe=    param(7)
  rkTe=    param(8)
  rkTbb=   param(9)
  rGeom=   param(10)
  rLogXi=  param(11)
  rRedshft= param(12)
  rRefl_frac= param(13)
  ruse_Ec= param(14)
  rlp2ext= param(15)
  rAcc=    param(16)
  rMbh_bb= param(17)
  rD_bb=   param(18)
  ralb_bb= param(19)
  rfcol_bb=param(20)
  rdmu_bb= param(21)
  rf_bb= param(22)
  rTherm_frac_bb= param(23)
  rNORM_bb=param(24)        

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRLPBB


SUBROUTINE INIT_TFF_TABLES_PATHS()
  use par_global, only: reflkerr_tables
  use par_rel, only: ftff,spd
  use par_rel, only: cached_tbtff

  IF(.not.cached_tbtff) THEN
     spd(1:12)=(/ 'a0100', 'a0200', 'a0300', 'a0400', 'a0500','a0600','a0700','a0800','a0900','a0950','a0980','a0998' /)
     ftff=trim(reflkerr_tables)//'/tff.fits'
     cached_tbtff = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE INIT_TFF_TABLES_PATHS


SUBROUTINE INIT_LP_TABLES_PATHS()
  use par_global, only: reflkerr_tables
  use par_rel, only: ftflp
  use par_rel, only: cached_tbtflp

  IF(.not.cached_tbtflp) THEN
     ftflp=trim(reflkerr_tables)//'/lp.fits'

     cached_tbtflp = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE INIT_LP_TABLES_PATHS


SUBROUTINE INIT_LP2_TABLES_PATHS()
  use par_global, only: reflkerr_tables
  use par_rel, only: ftflp2
  use par_rel, only: cached_tbtflp2

  IF(.not.cached_tbtflp2) THEN
     ftflp2=trim(reflkerr_tables)//'/lp2.fits'

     cached_tbtflp2 = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE INIT_LP2_TABLES_PATHS


SUBROUTINE DEFINE_SPIN_ARRAY()
  use par_rel, only: ns, sa
  use par_rel, only: cached_sp_arr

  IF(.not.cached_sp_arr) THEN
     DO i=1,ns-3
        sa(i)=0.1*i
     ENDDO
     sa(10:12)=(/ 0.95, 0.98, 0.998 /)
     
     cached_sp_arr = .TRUE.

  ENDIF

  RETURN
END SUBROUTINE DEFINE_SPIN_ARRAY


SUBROUTINE DEFINE_H_ARRAY()
  use par_rel, only: ns, hi, nh
  use par_rel, only: ihs,cached_h_arr

  IF(.not.cached_h_arr) THEN

     ihs(1:ns)=(/ 8,8,8,8,7,7,7,6,5,4,3,1 /)
     hi(1:2)=(/ 1.1, 1.21 /)
     DO i=3,nh
        hi(i)=1.3*(1.1**(i-3))
     ENDDO
     cached_h_arr=.TRUE.
  ENDIF

  RETURN
END SUBROUTINE DEFINE_H_ARRAY


SUBROUTINE DEFINE_R_ARRAY()
  use par_rel, only: nr
  use par_rel, only: r_in, r_out, rmin, rmax, cached_r_arr

  REAL :: du

  IF(.not.cached_r_arr) THEN

     du=Log10(rmax/rmin)/nr
     r_in(1)=rmin
     r_out(nr)=rmax
     DO i=1,nr-1
        r_out(i)=rmin*10.0**(i*du)
        r_in(i+1)=r_out(i)
     ENDDO

     cached_r_arr=.TRUE.
  ENDIF

  RETURN
END SUBROUTINE DEFINE_R_ARRAY


SUBROUTINE LOAD_SIND_TABLES()
  use par_rel, only: ftff
  use par_rel, only: ns,nr,nr_gr,nr_ngr,ncobs,ncem,nntc,nisl
  use par_rel, only: cached_sind_tbl

  use par_rel, only: mumin_d,mumax_d,mumin_ngr_d,mumax_ngr_d
  use par_rel, only: mumin_m,mumax_m,mumin_ngr_m,mumax_ngr_m
  use par_rel, only: mumin_u,mumax_u,mumin_ngr_u,mumax_ngr_u
  use par_rel, only: ntc_d,ntc1_d,ntc_ngr_d,ntc1_ngr_d,tfnel_d,tfnel_ngr_d,tfnel_tot_d
  use par_rel, only: ntc_m,ntc1_m,ntc_ngr_m,ntc1_ngr_m,tfnel_m,tfnel_ngr_m,tfnel_tot_m
  use par_rel, only: ntc_u,ntc1_u,ntc_ngr_u,ntc1_ngr_u,tfnel_u,tfnel_ngr_u,tfnel_tot_u
  use par_rel, only: tft_d,tft_ngr_d
  use par_rel, only: tft_m,tft_ngr_m
  use par_rel, only: tft_u,tft_ngr_u

  IMPLICIT NONE

  INTEGER :: unit,status
  INTEGER :: is,ir,iem,i,j,intc
  REAL,allocatable :: tft_tmp(:)

  integer :: readwrite,blocksize
  integer :: hdunum,chdu
  integer :: hdutype,colnum
  integer :: felem=1, nulld=0
  logical :: anynull

  IF(.not.cached_sind_tbl) THEN
     write(*,*) "Loading transfer function tables..."

     if (allocated(tft_d))deallocate(tft_d)
     if (allocated(tft_m))deallocate(tft_m)
     if (allocated(tft_u))deallocate(tft_u)
     
     if (allocated(tft_ngr_d))deallocate(tft_ngr_d)
     if (allocated(tft_ngr_m))deallocate(tft_ngr_m)
     if (allocated(tft_ngr_u))deallocate(tft_ngr_u)

     status=0
     readwrite=0
     blocksize=1

     call ftgiou(unit,status)
     call ftopen(unit,ftff,readwrite,blocksize,status)
     call printfitserror(status)
     if (status.ne.0) stop

     call FTTHDU(unit, hdunum, status)
     
     if (hdunum.ne.(6*(ns+1)+3+1)) then
        print *,'Number of Extensions is not correct - check table!'
        stop
     endif

     chdu=1
     do is=1,ns  
        chdu=chdu+1
        call FTMAHD(unit,chdu,hdutype,status)
        DO ir=1,nr_gr
           colnum=1
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_d(is,ir,:), &
                anynull,status)
           colnum=2
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_d(is,ir,:), &
                anynull,status)
           do iem=1,ncem
              colnum=7*iem-4
              call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_d(is,ir,iem,:), &
                   anynull,status)
              DO intc=1,nntc
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_d(is,1,ir,iem,:,intc), &
                      anynull,status)
              ENDDO
              DO intc=4,nntc+3
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_d(is,2,ir,iem,:,intc-3), &
                      anynull,status)
              ENDDO
           enddo
        ENDDO
     enddo
     
     chdu = chdu+1
     call FTMAHD(unit,chdu,hdutype,status)
     DO ir=1,nr_ngr
        colnum=1
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_ngr_d(ir,:), &
             anynull,status)
        colnum=2
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_ngr_d(ir,:), &
             anynull,status)
        do iem=1,ncem
           colnum=7*iem-4
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_ngr_d(ir,iem,:), &
                anynull,status)
              DO intc=1,nntc
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_d(1,ir,iem,:,intc), &
                      anynull,status)
              ENDDO
              DO intc=4,nntc+3
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_d(2,ir,iem,:,intc-3), &
                      anynull,status)
              ENDDO
        enddo
     ENDDO

     do is=1,ns  
        chdu = chdu+1
        call FTMAHD(unit,chdu,hdutype,status)
        DO ir=1,nr_gr
           colnum=1
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_m(is,ir,:), &
                anynull,status)
           colnum=2
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_m(is,ir,:), &
                anynull,status)
           do iem=1,ncem
              colnum=7*iem-4
              call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_m(is,ir,iem,:), &
                   anynull,status)
              DO intc=1,nntc
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_m(is,1,ir,iem,:,intc), &
                      anynull,status)
              ENDDO
              DO intc=4,nntc+3
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_m(is,2,ir,iem,:,intc-3), &
                      anynull,status)
              ENDDO
           enddo
        ENDDO
     enddo
     chdu = chdu+1
     call FTMAHD(unit,chdu,hdutype,status)
     DO ir=1,nr_ngr
        colnum=1
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_ngr_m(ir,:), &
             anynull,status)
        colnum=2
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_ngr_m(ir,:), &
             anynull,status)
        do iem=1,ncem
           colnum=7*iem-4
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_ngr_m(ir,iem,:), &
                anynull,status)
              DO intc=1,nntc
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_m(1,ir,iem,:,intc), &
                      anynull,status)
              ENDDO
              DO intc=4,nntc+3
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_m(2,ir,iem,:,intc-3), &
                      anynull,status)
              ENDDO
        enddo
     ENDDO

     do is=1,ns  
        chdu = chdu+1
        call FTMAHD(unit,chdu,hdutype,status)
        DO ir=1,nr_gr
           colnum=1
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_u(is,ir,:), &
                anynull,status)
           colnum=2
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_u(is,ir,:), &
                anynull,status)
           do iem=1,ncem
              colnum=7*iem-4
              call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_u(is,ir,iem,:), &
                   anynull,status)
              DO intc=1,nntc
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_u(is,1,ir,iem,:,intc), &
                      anynull,status)
              ENDDO
              DO intc=4,nntc+3
                 colnum=colnum+1
                 call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_u(is,2,ir,iem,:,intc-3), &
                      anynull,status)
              ENDDO
           enddo
        ENDDO
     enddo

     chdu = chdu+1
     call FTMAHD(unit,chdu,hdutype,status)
     DO ir=1,nr_ngr
        colnum=1
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumin_ngr_u(ir,:), &
             anynull,status)
        colnum=2
        call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,mumax_ngr_u(ir,:), &
             anynull,status)
        do iem=1,ncem
           colnum=7*iem-4
           call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc_ngr_u(ir,iem,:), &
                anynull,status)
           DO intc=1,nntc
              colnum=colnum+1
              call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_u(1,ir,iem,:,intc), &
                   anynull,status)
           ENDDO
           DO intc=4,nntc+3
              colnum=colnum+1
              call ftgcvj(unit,colnum,ir,felem,ncobs,nulld,ntc1_ngr_u(2,ir,iem,:,intc-3), &
                   anynull,status)
           ENDDO
        enddo
     ENDDO

     colnum=4
     tfnel_tot_d=0
     tfnel_tot_m=0
     tfnel_tot_u=0

     chdu = 3*(2*(ns+1))+2
     call FTMAHD(unit,chdu,hdutype,status)                                                                                                              
     DO is=1,ns                                                                                                                                           
        call ftgcvj(unit,colnum,is,felem,1,nulld,tfnel_d(is), &                                                                                                 
             anynull,status) 
        tfnel_tot_d=tfnel_tot_d+tfnel_d(is)
     ENDDO
     call ftgcvj(unit,colnum,ns+1,felem,1,nulld,tfnel_ngr_d, &                                                                                                 
          anynull,status) 
     chdu=chdu+1
     call FTMAHD(unit,chdu,hdutype,status)                                                                                                              
     DO is=1,ns                                                                                                                                           
        call ftgcvj(unit,colnum,is,felem,1,nulld,tfnel_m(is), &                                                                                                 
             anynull,status) 
        tfnel_tot_m=tfnel_tot_m+tfnel_m(is)
     ENDDO
     call ftgcvj(unit,colnum,ns+1,felem,1,nulld,tfnel_ngr_m, &                          
          anynull,status) 
     chdu=chdu+1
     call FTMAHD(unit,chdu,hdutype,status)                                                                                                              
     DO is=1,ns                                                                                                                                           
        call ftgcvj(unit,colnum,is,felem,1,nulld,tfnel_u(is), &                                                                                                 
             anynull,status) 
        tfnel_tot_u=tfnel_tot_u+tfnel_u(is)
     ENDDO
     call ftgcvj(unit,colnum,ns+1,felem,1,nulld,tfnel_ngr_u, &                          
          anynull,status) 

     allocate(tft_d(tfnel_tot_d))
     allocate(tft_m(tfnel_tot_m))
     allocate(tft_u(tfnel_tot_u))

     chdu=3*(ns+1)+1
     if(allocated(tft_tmp))deallocate(tft_tmp)
     
     j=0
     DO is=1,ns
        chdu=chdu+1
        call FTMAHD(unit,chdu,hdutype,status)   
        colnum=1
        allocate(tft_tmp(tfnel_d(is)))
        call ftgcve(unit,colnum,felem,1,tfnel_d(is),nulld,tft_tmp(:), &
             anynull,status)
        do i=1,tfnel_d(is)
           tft_d(i+j)=tft_tmp(i)
        enddo
        j=j+tfnel_d(is)
        deallocate(tft_tmp)
     ENDDO

     allocate(tft_ngr_d(tfnel_ngr_d))
     chdu=chdu+1
     call FTMAHD(unit,chdu,hdutype,status)   
     colnum=1
     call ftgcve(unit,colnum,felem,1,tfnel_ngr_d,nulld,tft_ngr_d(:), &
          anynull,status)
     
     j=0
     DO is=1,ns
        chdu=chdu+1
        call FTMAHD(unit,chdu,hdutype,status)   
        colnum=1
        allocate(tft_tmp(tfnel_m(is)))
        call ftgcve(unit,colnum,felem,1,tfnel_m(is),nulld,tft_tmp(:), &
             anynull,status)
        do i=1,tfnel_m(is)
           tft_m(i+j)=tft_tmp(i)
        enddo
        j=j+tfnel_m(is)
        deallocate(tft_tmp)
     ENDDO

     allocate(tft_ngr_m(tfnel_ngr_m))
     chdu=chdu+1
     call FTMAHD(unit,chdu,hdutype,status)   
     colnum=1
     call ftgcve(unit,colnum,felem,1,tfnel_ngr_m,nulld,tft_ngr_m(:), &
          anynull,status)

     j=0
     DO is=1,ns
        chdu=chdu+1
        call FTMAHD(unit,chdu,hdutype,status)   
        colnum=1
        allocate(tft_tmp(tfnel_u(is)))
        call ftgcve(unit,colnum,felem,1,tfnel_u(is),nulld,tft_tmp(:), &
             anynull,status)
        do i=1,tfnel_u(is)
           tft_u(i+j)=tft_tmp(i)
        enddo


        j=j+tfnel_u(is)
        deallocate(tft_tmp)
     ENDDO

     allocate(tft_ngr_u(tfnel_ngr_u))
     chdu=chdu+1
     call FTMAHD(unit,chdu,hdutype,status)   
     colnum=1
     call ftgcve(unit,colnum,felem,1,tfnel_ngr_u,nulld,tft_ngr_u(:), &
          anynull,status)

     call ftfiou(unit,status)

  write(*,*) "                                   ...done."

  cached_sind_tbl = .TRUE.
  ENDIF

  RETURN
END SUBROUTINE LOAD_SIND_TABLES


SUBROUTINE LOAD_LP_TF()
  use par_rel, only: ftflp
  use par_rel, only: gr,fphr,fphd
  use par_rel, only: ns,nh,ihs,sn,nr,nint,ihs
  use par_rel, only: cached_lp_sn
  IMPLICIT NONE

  INTEGER :: unit,status,ih,is,ir,dumi

  integer :: readwrite,blocksize
  integer :: hdunum,chdu
  integer :: hdutype,colnum
  integer :: felem=1, nulld=0
  logical :: anynull

  status=0
  readwrite=0
  blocksize=1
  
  IF(cached_lp_sn.ne.sn) THEN

     call ftgiou(unit,status)

     call ftopen(unit,ftflp,readwrite,blocksize,status)
     call printfitserror(status)
     if (status.ne.0) stop

     call FTTHDU(unit, hdunum, status)
     if (hdunum.ne.(ns+1)) then
        print *,'Number of Extensions is not correct - check table!'
        stop
     endif

     DO is=1,nint
        chdu = 1+sn+is
        call FTMAHD(unit,chdu,hdutype,status)     
        DO ih=1,nh-ihs(is+sn)+1
           dumi=ihs(is+sn)+ih-1
           colnum=2
           call ftgcve(unit,colnum,ih,felem,nr,nulld,gr(is,dumi,:), &
                anynull,status)
           colnum=3
           call ftgcve(unit,colnum,ih,felem,nr,nulld,fphr(is,dumi,:), &
                anynull,status)
           colnum=4
           call ftgcve(unit,colnum,ih,felem,1,nulld,fphd(is,dumi), &
                anynull,status)
           DO ir=1,nr
              fphr(is,dumi,ir)=fphr(is,dumi,ir)*1.0e-7
           ENDDO
        ENDDO
     ENDDO

     call ftclos(unit, status)
     call ftfiou(unit, status)

     DO ih=ihs(2+sn), ihs(1+sn)-1
        do ir=1,nr
           fphr(1,ih,ir)=0.0
           gr(1,ih,ir)=gr(2,ih,ir)
        enddo
        fphd(1,ih)=0.0
     ENDDO

     cached_lp_sn=sn

  ENDIF
     
  RETURN
END SUBROUTINE LOAD_LP_TF


SUBROUTINE LOAD_LP2_TF()
  use par_rel, only: ftflp2
  use par_rel, only: fphr2,fphd2,ihs
  use par_rel, only: ns,nh,sn,nr,nrin,nint,ncobs
  use par_rel, only: cached_lp2_sn
  IMPLICIT NONE

  INTEGER :: unit,status,ih,is,ir,irin,imu,dumi

  integer :: readwrite,blocksize
  integer :: hdunum,chdu
  integer :: hdutype,colnum
  integer :: felem=1, nulld=0
  logical :: anynull

  status=0
  readwrite=0
  blocksize=1
  
  IF(cached_lp2_sn.ne.sn) THEN

     call ftgiou(unit,status)

     call ftopen(unit,ftflp2,readwrite,blocksize,status)
     call printfitserror(status)
     if (status.ne.0) stop

     call FTTHDU(unit, hdunum, status)
     if (hdunum.ne.(ns+1)) then
        print *,'Number of Extensions is not correct - check table!'
        stop
     endif

     DO is=1,nint
        chdu = 1+sn+is
        call FTMAHD(unit,chdu,hdutype,status)     

        DO ih=1,nh-ihs(is+sn)+1
           dumi=ihs(is+sn)+ih-1
           DO ir=2,nr+1
              colnum=ir
              call ftgcve(unit,colnum,ih,felem,nrin,nulld,fphr2(is,dumi,:,ir-1), &
                   anynull,status)
              do irin=1,nrin
                 fphr2(is,dumi,irin,ir-1)=fphr2(is,dumi,irin,ir-1)*1.0e-7
              enddo
           ENDDO
           DO imu=nr+2,ncobs+nr+1
              colnum=imu
              call ftgcve(unit,colnum,ih,felem,nrin,nulld,fphd2(is,dumi,:,imu-nr-1), &
                   anynull,status)
           ENDDO
        ENDDO
     ENDDO

     call ftclos(unit, status)
     call ftfiou(unit, status)

     DO ih=ihs(2+sn), ihs(1+sn)-1
        DO irin=1,nrin
           do ir=1,nr
              fphr2(1,ih,irin,ir)=0.0
           enddo
           do imu=1,ncobs
              fphd2(1,ih,irin,imu)=0.0
           enddo
        ENDDO
     ENDDO

     cached_lp2_sn=sn

  ENDIF
     
  RETURN
END SUBROUTINE LOAD_LP2_TF


INTEGER function get_low_rbin(Rin_g)
  use par_rel, only: rmin,rmax,nr
  IMPLICIT NONE

  REAL :: Rin_g

  get_low_rbin=INT(alog10(Rin_g/rmin)/alog10(rmax/rmin)*nr+1.0)
  if(Rin_g.le.rmin) get_low_rbin=1
  if(Rin_g.ge.rmax) get_low_rbin=nr
  
  return
end function get_low_rbin


INTEGER function get_high_rbin(Rout_g)
  use par_rel, only: rmin,rmax,nr
  IMPLICIT NONE

  REAL :: Rout_g

  get_high_rbin=INT(alog10(Rout_g/rmin)/alog10(rmax/rmin)*nr+1.0)
  if(Rout_g.ge.rmax) get_high_rbin=nr
  if(Rout_g.le.rmin) get_high_rbin=1
  
  return
end function get_high_rbin


REAL FUNCTION rms(a)
  IMPLICIT NONE

  REAL :: a
  REAL :: a1,a2

  a1=1.+((1.+a)**(1/3.)+(1-a)**(1/3.))*((1.-a**2)**(1/3.))
  a2=sqrt(3.*(a**2)+a1**2)
  rms=3.+a2-sqrt((3.-a1)*(3.+a1+2.*a2))

RETURN
END FUNCTION rms


REAL FUNCTION rhor(a)
  IMPLICIT NONE

  REAL :: a

  rhor=1.0 + sqrt (1.0 - a*a);

RETURN
END FUNCTION rhor


INTEGER function get_low_h_bin(H)
  use par_rel, only:sn,hi,ihs,ra,nh
  IMPLICIT NONE

  REAL, SAVE :: cached_H=-1.0, cached_sn=-1.0
  INTEGER, SAVE :: hn
  REAL :: rhor,H,Hmin

  IF((cached_sn.ne.sn).or.(cached_H.ne.H))THEN

     Hmin=rhor(ra)*1.001
     if(Hmin.lt.hi(ihs(2+sn)))Hmin=hi(ihs(2+sn))

     if(H.lt.Hmin)then
        write(*,"(A2,f5.3,A47,f5.3)") "h=",H, " is to close to BH for choosen spin. Setting h=",Hmin
        H=Hmin
     endif

     if(H.gt.hi(nh))then
        H=hi(nh)
     endif

     hn=ihs(2+sn)
     do while ((H.ge.hi(hn)).and.(hn.lt.nh))
        hn=hn+1
     enddo
     hn=hn-1
     cached_H=H
     cached_sn=sn
  ENDIF
  get_low_h_bin=hn

  return
end function get_low_h_bin


INTEGER function get_low_inc_bin(Incl)
  use par_global, only: Pi
  use par_rel, only: ncobs
  IMPLICIT NONE

  REAL :: Incl

  get_low_inc_bin=INT(0.5+COS(Incl*Pi/180.)*ncobs)

  return
end function get_low_inc_bin


SUBROUTINE INTERPOLATE_LP_TF(H,ihl)
  use par_rel, only: gr,fphr,fphd,gri,fphri,fphdi
  use par_rel, only: hi,sn,nr,nint
  IMPLICIT NONE

  REAL :: H
  INTEGER :: is,ir
  INTEGER :: ihl,ihh
  REAL :: factor
  INTEGER, SAVE :: cached_sn = -1
  REAL, SAVE :: cached_H = -1.0

  IF((cached_sn.ne.sn).or.(cached_H.ne.H)) THEN

     ihh=ihl+1
     factor = (H-hi(ihl))/(hi(ihh)-hi(ihl))
     
     DO is=1,nint
        do ir=1,nr
           IF((gr(is,ihl,ir).lt.1.0e-20).or.(gr(is,ihh,ir).lt.1.0e-20))THEN
              gri(is,ir)=1.0e-24
              fphri(is,ir)=0.0
           ELSE
              gri(is,ir)=(1.0-factor)*gr(is,ihl,ir)+factor*gr(is,ihh,ir)
              fphri(is,ir)=(1.0-factor)*fphr(is,ihl,ir)+factor*fphr(is,ihh,ir)
           ENDIF
        enddo
        fphdi(is)=(1.0-factor)*fphd(is,ihl)+factor*fphd(is,ihh)
     ENDDO

     cached_sn=sn
     cached_H=H

  ENDIF

  RETURN
END SUBROUTINE INTERPOLATE_LP_TF


SUBROUTINE INTERPOLATE_LP2_TF(H,Rin_g,ihl,irl,iml)
  use par_global, only: Pi
  use par_rel, only: fphr2,fphd2,fphr2i,fphd2i,fphd2_av,show_all_av
  use par_rel, only: hi,sn,nr,ncobs,nint,r_in,rIncl
  IMPLICIT NONE

  REAL :: H, Rin_g
  INTEGER :: ihl,irl,iml

  INTEGER :: is,ir,imu
  REAL ::  cosi,cil,cih,fact1,fact2,fact3
  INTEGER, SAVE :: cached_sn = -1
  REAL, SAVE :: cached_H = -1.0
  REAL, SAVE :: cached_rIncl = -1.0
  REAL, SAVE :: cached_Rin_g = -1.0

  IF((cached_sn.ne.sn).or.(cached_H.ne.H).or.(cached_rIncl.ne.rIncl).or.(cached_Rin_g.ne.Rin_g).or.show_all_av)THEN
     if(Rin_g.lt.r_in(1))Rin_g=r_in(1)
     if(Rin_g.gt.r_in(nr))Rin_g=r_in(nr)
     cosi=real(COS(rIncl*Pi/180.0))
     cil=(iml-0.5)/ncobs
     cih=((iml+1)-0.5)/ncobs

     fact1=(Rin_g-r_in(irl))/(r_in(irl+1)-r_in(irl))
     fact2=(H-hi(ihl))/(hi(ihl+1)-hi(ihl))
     fact3=(cosi-cil)/(cih-cil)

     DO is=1,nint
        fphd2_av(is)=0.
        DO imu=1,ncobs
           fphd2_av(is)=fphd2_av(is)+&
                ((1.0-fact1)*(1.0-fact2)*fphd2(is,ihl,irl,imu)+ &
                (1.0-fact1)*fact2*fphd2(is,ihl+1,irl,imu)+ &
                fact1*(1.0-fact2)*fphd2(is,ihl,irl+1,imu)+ &
                fact1*fact2*fphd2(is,ihl+1,irl+1,imu))
        ENDDO
        fphd2_av(is)=fphd2_av(is)/ncobs
     ENDDO

     DO is=1,nint
        do ir=1,nr
           fphr2i(is,ir)=(1.0-fact1)*(1.0-fact2)*fphr2(is,ihl,irl,ir) &
                +fact1*(1.0-fact2)*fphr2(is,ihl,irl+1,ir) &
                +(1.0-fact1)*fact2*fphr2(is,ihl+1,irl,ir) &
                +fact1*fact2*fphr2(is,ihl+1,irl+1,ir) 
        enddo
        fphd2i(is)=(1.0-fact1)*(1.0-fact2)*(1.0-fact3)*fphd2(is,ihl,irl,iml) &
             +fact1*(1.0-fact2)*(1.0-fact3)*fphd2(is,ihl,irl+1,iml) &
             +(1.0-fact1)*fact2*(1.0-fact3)*fphd2(is,ihl+1,irl,iml) &
             +(1.0-fact1)*(1.0-fact2)*fact3*fphd2(is,ihl,irl,iml+1) &
             +fact1*fact2*(1.0-fact3)*fphd2(is,ihl+1,irl+1,iml) &
             +fact1*(1.0-fact2)*fact3*fphd2(is,ihl,irl+1,iml+1) &
             +(1.0-fact1)*fact2*fact3*fphd2(is,ihl+1,irl,iml+1) &
             +fact1*fact2*fact3*fphd2(is,ihl+1,irl+1,iml+1)

     ENDDO
     cached_sn = sn
     cached_H = H
     cached_rIncl = rIncl
     cached_Rin_g = Rin_g

  ENDIF

  RETURN
END SUBROUTINE INTERPOLATE_LP2_TF


SUBROUTINE CALCULATE_WGT_ARRAY(Rin_g,Rout_g,iml,irl,irh,Acc,wgt)

  use par_rel, only: sn,nr,nr_gr,ncem,nint,r_in,r_out
  use par_rel, only: mul,muh,gl,gh
  use par_rel, only: fphr_tot,rH,rlp2ext
  use par_rel, only: was_wgt_invoked

  use par_rel, only: nntc

  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u,gl_val,gh_val
  use par_rel, only: ntc_d,ntc1_d,ntc_m,ntc1_m,ntc_u,ntc1_u
  use par_rel, only: ntc_ngr_d,ntc1_ngr_d,ntc_ngr_m,ntc1_ngr_m,ntc_ngr_u,ntc1_ngr_u

  use par_rel, only: tfnel_d,tfnel_m,tfnel_u
  use par_rel, only: tft_d,tft_ngr_d,tft_m,tft_ngr_m,tft_u,tft_ngr_u
  use par_rel, only: tfnel_tot_d,tfnel_tot_m,tfnel_tot_u
  use par_rel, only: tfnel_ngr_d,tfnel_ngr_m,tfnel_ngr_u

  use par_rel, only: mumin_d,mumax_d,mumin_m,mumax_m,mumin_u,mumax_u
  use par_rel, only: mumin_ngr_d,mumax_ngr_d,mumin_ngr_m,mumax_ngr_m,mumin_ngr_u,mumax_ngr_u

  IMPLICIT NONE

  REAL :: Rin_g,Rout_g,Acc
  REAL :: wgt(nint,nint,ncem,nr)
  INTEGER :: iml,irl,irh

  INTEGER :: i,is,ir,imu,iem,ig,nnn
  REAL :: scint_l,scint_h,wgtsum10,wgtsum100,fphr_totc,tff
  REAL :: arr(nr*ncem)
  INTEGER :: iarr(nr*ncem)
  
  REAL,PARAMETER :: conv_acc=0.1

  INTEGER, SAVE :: cached_sn = -1
  INTEGER, SAVE :: cached_iml = -1
  REAL, SAVE :: cached_Rin_g = -1.0
  REAL, SAVE :: cached_Rout_g = -1.0
  REAL, SAVE :: cached_lp2ext = -1.0
  REAL, SAVE :: cached_H = -1.0
  REAL, SAVE :: cached_Acc = -1.0

  INTEGER isl,nseek,nog,igsl
  INTEGER :: tfel_l_d(nint),tfel_l_m(nint),tfel_l_u(nint)
  INTEGER :: nav(ncem),spos,mupos,rpos
  REAL,ALLOCATABLE,SAVE :: tft_d_tmp(:),tft_m_tmp(:),tft_u_tmp(:)
  REAL,ALLOCATABLE,SAVE :: tft_ngr_d_tmp(:),tft_ngr_m_tmp(:),tft_ngr_u_tmp(:)
  LOGICAL,SAVE :: FirstCall=.TRUE.
  
  was_wgt_invoked=.FALSE.

  IF((cached_sn.ne.sn).or.(cached_iml.ne.iml).or.(cached_Rin_g.ne.Rin_g).or. &
       (cached_Rout_g.ne.Rout_g).or.(cached_lp2ext.ne.rlp2ext).or.(cached_H.ne.rH).or. &
       (cached_Acc.ne.Acc))THEN

     IF(FirstCall)THEN
        allocate(tft_d_tmp(tfnel_tot_d))
        allocate(tft_m_tmp(tfnel_tot_m))
        allocate(tft_u_tmp(tfnel_tot_u))
        allocate(tft_ngr_d_tmp(tfnel_ngr_d))
        allocate(tft_ngr_m_tmp(tfnel_ngr_m))
        allocate(tft_ngr_u_tmp(tfnel_ngr_u))     
        tft_d_tmp=tft_d
        tft_m_tmp=tft_m
        tft_u_tmp=tft_u
        tft_ngr_d_tmp=tft_ngr_d
        tft_ngr_m_tmp=tft_ngr_m
        tft_ngr_u_tmp=tft_ngr_u
        FirstCall=.FALSE.
     ELSE
        tft_d=tft_d_tmp
        tft_m=tft_m_tmp
        tft_u=tft_u_tmp
        tft_ngr_d=tft_ngr_d_tmp
        tft_ngr_m=tft_ngr_m_tmp
        tft_ngr_u=tft_ngr_u_tmp
     ENDIF

     gl=1 !delete
     gh=1700 !delete

     mul=10
     muh=1

     gl_d=75
     gh_d=1
     gl_m=155
     gh_m=106
     gl_u=1700
     gh_u=776

     scint_l=(r_out(irl)**2.0-(Rin_g)**2.0)/(r_out(irl)**2.0-r_in(irl)**2.0)
     scint_h=((Rout_g)**2.0-r_in(irh)**2.0)/(r_out(irh)**2.0-r_in(irh)**2.0)

     IF(irl.eq.irh)then
        if(Rin_g.ge.Rout_g)then
           scint_h=0.0
        else
           scint_h=((Rout_g)**2.0-(Rin_g)**2.0)/(r_out(irh)**2.0-(Rin_g)**2.0)
        endif
     ENDIF

     do is=1,nint
        DO imu=1,nint
           do iem=1,ncem
              DO ir=1,nr
                 wgt(is,imu,iem,ir)=0.0
              ENDDO
           enddo
        ENDDO
     enddo

     DO is=1,nint
        tfel_l_d(is)=0
        tfel_l_m(is)=0
        tfel_l_u(is)=0
        if((is+sn).gt.1)then
           do i=2,is+sn
              tfel_l_d(is)=tfel_l_d(is)+tfnel_d(i-1)
              tfel_l_m(is)=tfel_l_m(is)+tfnel_m(i-1)
              tfel_l_u(is)=tfel_l_u(is)+tfnel_u(i-1)
           enddo
        endif
     ENDDO

     ir=irl
     IF(ir.le.nr_gr)THEN
        do is=1,nint
           spos=is+sn
           DO imu=iml,iml+1
              mupos=imu-iml+1
              do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
                 DO isl=1,ntc_d(spos,ir,iem,imu)
                    igsl = ntc1_d(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
                    nog = ntc1_d(spos,isl,ir,iem,imu,2) 
                    do ig=igsl,igsl+nog
                       tft_d(nseek)=tft_d(nseek)*scint_l
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
                 DO isl=1,ntc_m(spos,ir,iem,imu)
                    igsl = ntc1_m(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
                    nog = ntc1_m(spos,isl,ir,iem,imu,2) 
                    do ig=igsl,igsl+nog
                       tft_m(nseek)=tft_m(nseek)*scint_l
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
                 DO isl=1,ntc_u(spos,ir,iem,imu)
                    igsl = ntc1_u(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
                    nog = ntc1_u(spos,isl,ir,iem,imu,2) 
                    do ig=igsl,igsl+nog
                       tft_u(nseek)=tft_u(nseek)*scint_l
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
           ENDDO
        enddo
     ELSE
        rpos=ir-nr_gr
        DO imu=iml,iml+1
           mupos=imu-iml+1
           do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
              DO isl=1,ntc_ngr_d(rpos,iem,imu)
                 igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_d(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_d(nseek)=tft_ngr_d(nseek)*scint_l                    
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
              DO isl=1,ntc_ngr_m(rpos,iem,imu)
                 igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_m(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_m(nseek)=tft_ngr_m(nseek)*scint_l                                        
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
              DO isl=1,ntc_ngr_u(rpos,iem,imu)
                 igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_u(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_u(nseek)=tft_ngr_u(nseek)*scint_l                                                            
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
        ENDDO
     ENDIF

     ir=irh
     IF(ir.le.nr_gr)THEN
     do is=1,nint
        spos=is+sn
        DO imu=iml,iml+1
           mupos=imu-iml+1
           do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
              DO isl=1,ntc_d(spos,ir,iem,imu)
                 igsl = ntc1_d(spos,isl,ir,iem,imu,1)
                 nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
                 nog = ntc1_d(spos,isl,ir,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_d(nseek)=tft_d(nseek)*scint_h
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
              DO isl=1,ntc_m(spos,ir,iem,imu)
                 igsl = ntc1_m(spos,isl,ir,iem,imu,1)
                 nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
                 nog = ntc1_m(spos,isl,ir,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_m(nseek)=tft_m(nseek)*scint_h                    
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
                 DO isl=1,ntc_u(spos,ir,iem,imu)
                    igsl = ntc1_u(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
                    nog = ntc1_u(spos,isl,ir,iem,imu,2) 
                    do ig=igsl,igsl+nog
                       tft_u(nseek)=tft_u(nseek)*scint_h                                           
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
           ENDDO
        enddo
     ELSE
        rpos=ir-nr_gr
        DO imu=iml,iml+1
           mupos=imu-iml+1
           do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
              DO isl=1,ntc_ngr_d(rpos,iem,imu)
                 igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_d(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_d(nseek)=tft_ngr_d(nseek)*scint_h
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
              DO isl=1,ntc_ngr_m(rpos,iem,imu)
                 igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_m(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_m(nseek)=tft_ngr_m(nseek)*scint_h
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
              DO isl=1,ntc_ngr_u(rpos,iem,imu)
                 igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_u(isl,rpos,iem,imu,2) 
                 do ig=igsl,igsl+nog
                    tft_ngr_u(nseek)=tft_ngr_u(nseek)*scint_h
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
        ENDDO
     ENDIF

     DO ir=irl,min(irh,nr_gr)
        do is=1,nint
           spos=is+sn
           DO imu=iml,iml+1
              mupos=imu-iml+1
              nav=0
              do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
                 if(mul.gt.mumin_d(spos,ir,imu))mul=mumin_d(spos,ir,imu)
                 if(muh.lt.mumax_d(spos,ir,imu))muh=mumax_d(spos,ir,imu)
                 DO isl=1,ntc_d(spos,ir,iem,imu)
                    igsl = ntc1_d(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
                    nog = ntc1_d(spos,isl,ir,iem,imu,2) 
                    if(gl_d.gt.igsl)gl_d=igsl
                    if(gh_d.lt.(igsl+nog))gh_d=igsl+nog
                    do ig=igsl,igsl+nog
                       tff=tft_d(nseek)
                       wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)+tff
                       if(tff.gt.0.0)nav(iem)=nav(iem)+1
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
                 if(mul.gt.mumin_m(spos,ir,imu))mul=mumin_m(spos,ir,imu)
                 if(muh.lt.mumax_m(spos,ir,imu))muh=mumax_m(spos,ir,imu)
                 DO isl=1,ntc_m(spos,ir,iem,imu)
                    igsl = ntc1_m(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
                    nog = ntc1_m(spos,isl,ir,iem,imu,2) 
                    if(gl_m.gt.igsl)gl_m=igsl
                    if(gh_m.lt.(igsl+nog))gh_m=igsl+nog
                    do ig=igsl,igsl+nog
                       tff=tft_m(nseek)
                       wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)+tff
                       if(tff.gt.0.0)nav(iem)=nav(iem)+1
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
                 if(mul.gt.mumin_u(spos,ir,imu))mul=mumin_u(spos,ir,imu)
                 if(muh.lt.mumax_u(spos,ir,imu))muh=mumax_u(spos,ir,imu)
                 DO isl=1,ntc_u(spos,ir,iem,imu)
                    igsl = ntc1_u(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
                    nog = ntc1_u(spos,isl,ir,iem,imu,2) 
                    if(gl_u.gt.igsl)gl_u=igsl
                    if(gh_u.lt.(igsl+nog))gh_u=igsl+nog
                    do ig=igsl,igsl+nog
                       tff=tft_u(nseek)
                       wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)+tff
                       if(tff.gt.0.0)nav(iem)=nav(iem)+1
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mul,muh
                 if(nav(iem).gt.0)then
                    wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)/nav(iem)
                 endif
              enddo
           ENDDO
        enddo
     ENDDO

     DO ir=max((nr_gr+1),irl),irh
        rpos=ir-nr_gr
        DO imu=iml,iml+1
           mupos=imu-iml+1
           nav=0
           do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
              if(mul.gt.mumin_ngr_d(rpos,imu))mul=mumin_ngr_d(rpos,imu)
              if(muh.lt.mumax_ngr_d(rpos,imu))muh=mumax_ngr_d(rpos,imu)
              DO isl=1,ntc_ngr_d(rpos,iem,imu)
                 igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_d(isl,rpos,iem,imu,2) 
                 if(gl_d.gt.igsl)gl_d=igsl
                 if(gh_d.lt.(igsl+nog))gh_d=igsl+nog
                 do ig=igsl,igsl+nog
                    tff=tft_ngr_d(nseek)
                    wgt(1,mupos,iem,ir)=wgt(1,mupos,iem,ir)+tff
                    nseek = nseek + 1
                    if(tff.gt.0.0)nav(iem)=nav(iem)+1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
              if(mul.gt.mumin_ngr_m(rpos,imu))mul=mumin_ngr_m(rpos,imu)
              if(muh.lt.mumax_ngr_m(rpos,imu))muh=mumax_ngr_m(rpos,imu)
              DO isl=1,ntc_ngr_m(rpos,iem,imu)
                 igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_m(isl,rpos,iem,imu,2) 
                 if(gl_m.gt.igsl)gl_m=igsl
                 if(gh_m.lt.(igsl+nog))gh_m=igsl+nog
                 do ig=igsl,igsl+nog
                    tff=tft_ngr_m(nseek)
                    wgt(1,mupos,iem,ir)=wgt(1,mupos,iem,ir)+tff
                    nseek = nseek + 1
                    if(tff.gt.0.0)nav(iem)=nav(iem)+1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
              if(mul.gt.mumin_ngr_u(rpos,imu))mul=mumin_ngr_u(rpos,imu)
              if(muh.lt.mumax_ngr_u(rpos,imu))muh=mumax_ngr_u(rpos,imu)
              DO isl=1,ntc_ngr_u(rpos,iem,imu)
                 igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_u(isl,rpos,iem,imu,2) 
                 if(gl_u.gt.igsl)gl_u=igsl
                 if(gh_u.lt.(igsl+nog))gh_u=igsl+nog
                 do ig=igsl,igsl+nog
                    tff=tft_ngr_u(nseek)
                    wgt(1,mupos,iem,ir)=wgt(1,mupos,iem,ir)+tff
                    nseek = nseek + 1
                    if(tff.gt.0.0)nav(iem)=nav(iem)+1
                 enddo
              ENDDO
           enddo
           do iem=mul,muh
              if(nav(iem).gt.0)then
                 wgt(1,mupos,iem,ir)=wgt(1,mupos,iem,ir)/nav(iem)
                 wgt(2,mupos,iem,ir)=wgt(1,mupos,iem,ir)
              endif
           enddo
        ENDDO
     ENDDO

     DO ir=irl,min(irh,nr_gr)
        do is=1,nint
           spos=is+sn
           fphr_totc=fphr_tot(is,ir)
           DO imu=iml,iml+1
              mupos=imu-iml+1
              do iem=mul,muh
                 wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)*fphr_totc
              enddo
           ENDDO
        enddo
     ENDDO
     DO ir=max((nr_gr+1),irl),irh
        do is=1,nint
           spos=is+sn
           fphr_totc=fphr_tot(is,ir)
           DO imu=iml,iml+1
              mupos=imu-iml+1
              do iem=mul,muh
                 wgt(is,mupos,iem,ir)=wgt(is,mupos,iem,ir)*fphr_totc
              enddo
           ENDDO
        enddo
     ENDDO

     IF(Acc.lt.3.5)THEN
        DO is=1,nint
           do imu=1,nint
              wgtsum100=0.0
              nnn=0
              DO ir=1,nr
                 do iem=1,ncem
                    arr((ir-1)*ncem+iem)=wgt(is,imu,iem,ir)
                    wgtsum100=wgtsum100+arr((ir-1)*ncem+iem)
                    if(wgt(is,imu,iem,ir).gt.0.0)nnn=nnn+1
                 enddo
              ENDDO
              wgtsum100=wgtsum100/nnn
              call piksrt(nr*ncem,arr,iarr)
              wgtsum10=0.0
              DO i=1,nr*ncem
                 wgtsum10=arr(i)
                 if((mod(iarr(i),ncem).eq.0)) then
                    ir=iarr(i)/ncem
                    iem=ncem
                    wgt(is,imu,iem,ir)=0.0
                 else
                    ir=(iarr(i)/ncem+1)
                    iem=mod(iarr(i),ncem)
                    wgt(is,imu,iem,ir)=0.0
                 endif
                 if((wgtsum10/wgtsum100).ge.conv_acc)then
                    exit
                 endif

              ENDDO
           enddo
        ENDDO
     ENDIF

     if(gl_d.le.gh_d)then
        gl_val=(gl_d-0.5)*0.007
        gl=gl_d
     else if(gl_m.le.gh_m)then
        gl_val=(gl_m-0.5)*0.005
        gl=gl_m
     else if(gl_u.le.gh_u)then
        gl_val=(gl_u-0.5)*0.001
        gl=gl_u
     endif

     if(gh_u.ge.gl_u)then
        gh_val=(gh_u-0.5)*0.001
        gh=gh_u
     else if(gh_m.ge.gl_m)then
        gh_val=(gh_m-0.5)*0.005
        gh=gh_m
     else if(gh_d.ge.gl_d)then
        gh_val=(gl_d-0.5)*0.007
        gh=gh_d
     endif
  

     cached_sn = sn
     cached_iml = iml
     cached_Rin_g = Rin_g
     cached_Rout_g = Rout_g
     cached_lp2ext=rlp2ext
     cached_H=rH
     cached_Acc=Acc

     was_wgt_invoked=.TRUE.
  ENDIF

  RETURN
END SUBROUTINE CALCULATE_WGT_ARRAY


SUBROUTINE GET_APP_ANGLES_BINS(nbg,wgt)
  use par_rel, only: was_wgt_invoked
  use par_rel, only: nint,ncem,nr,muem_app
  IMPLICIT NONE

  INTEGER :: nbg
  REAL :: wgt(nint,nint,ncem,nr)

  INTEGER :: is,ir,i,j
  REAL :: mu,wgttot

  IF(was_wgt_invoked)THEN
     
     if(allocated(muem_app))then
        deallocate(muem_app)
     endif
     allocate(muem_app(nint,nr,ncem/nbg))

     IF(nbg.gt.1)THEN
        DO is=1,nint
           DO ir=1,nr
              i=1
              j=1
              mu=0.0
              wgttot=0.0
              do while(i.le.ncem)
                 if((wgt(is,1,i,ir).gt.1.0e-22).or.(wgt(is,2,i,ir).gt.1.0e-22))then
                    wgttot=wgttot+wgt(is,1,i,ir)+wgt(is,2,i,ir)
                    mu=mu+((i-0.5)/ncem)*(wgt(is,1,i,ir)+wgt(is,2,i,ir))
                 endif
                 if((mod(i,nbg).eq.0.0))then
                    if(wgttot.lt.1.0e-22)then
                       muem_app(is,ir,j)=-1
                    else
                       muem_app(is,ir,j)=INT(0.5+mu/wgttot*ncem)
                    endif
                    mu=0.0
                    wgttot=0.0
                    j=j+1
                 endif
                 i=i+1
              enddo
           ENDDO
        ENDDO

     ELSE

        DO is=1,nint
           DO ir=1,nr
              DO j=1,ncem
                 if((wgt(is,1,j,ir).gt.1.0e-22).or.(wgt(is,2,j,ir).gt.1.0e-22))then
                    muem_app(is,ir,j)=j
                 else
                    muem_app(is,ir,j)=-1
                 endif
              ENDDO
           ENDDO
        ENDDO

     ENDIF
     
  ENDIF

  RETURN
END SUBROUTINE GET_APP_ANGLES_BINS


SUBROUTINE GET_LMOD_SERIAL_MATRIX(nbg)
  use par_rel, only: was_wgt_invoked,gri
  use par_rel, only: nint,ncem,nr,muem_app
  use par_rel, only: ist,irt,jt,ist1,irt1,jt1,istr
  IMPLICIT NONE

  INTEGER nbg

  INTEGER :: is,ir,i,j,k,l,m,n,o,ntot
  REAL, PARAMETER :: g_acc = 0.1 !0.12
  REAL :: dum
  REAL,ALLOCATABLE :: dimas(:),ser_muem(:)
  INTEGER, ALLOCATABLE :: iarr(:), inds(:),inds2(:)

  ntot=nint*ncem/nbg*nr

  IF (was_wgt_invoked) THEN

     if(allocated(ser_muem))deallocate(ser_muem)
     allocate(ser_muem(ntot))

     if(allocated(iarr))deallocate(iarr)
     allocate(iarr(ntot))

     if(allocated(ist1))deallocate(ist1)
     allocate(ist1(ntot))
     if(allocated(irt1))deallocate(irt1)
     allocate(irt1(ntot))
     if(allocated(jt1))deallocate(jt1)
     allocate(jt1(ntot))

     if(allocated(ist))deallocate(ist)
     allocate(ist(ntot))
     if(allocated(irt))deallocate(irt)
     allocate(irt(ntot))
     if(allocated(jt))deallocate(jt)
     allocate(jt(ntot))

     k=0
     do is=1,nint
        do j=1,ncem/nbg
           do ir=1,nr
              k=k+1
              ser_muem(k)=real(muem_app(is,ir,j))

              ist1(k)=is
              irt1(k)=ir
              jt1(k)=j
           enddo
        enddo
     enddo

     call piksrt(ntot,ser_muem,iarr)

     dum=-10.0
     k=0
     do i=1,ntot
        if(dum.ne.ser_muem(i))then
           if(k.ne.0) then
              if(allocated(dimas))deallocate(dimas)
              allocate(dimas(k))
              if(allocated(inds))deallocate(inds)
              allocate(inds(k))
              if(allocated(inds2))deallocate(inds2)
              allocate(inds2(k))

              do j=1,k
                 inds(j)=(i-1-k+j)
                 dimas(j)=gri(ist1(iarr(inds(j))), irt1(iarr(inds(j))))
              enddo
              call  piksrt(k,dimas,inds2)
              l=0
              do j=1,k
                 if((l.eq.0).or.abs(dimas(l)-dimas(j))/dimas(l).gt.g_acc)then
!                 if((l.eq.0).or.abs(dimas(l)-dimas(j)).gt.g_acc)then
                    l=j
                    ist(iarr(inds(inds2(j))))=ist1(iarr(inds(inds2(j))))
                    irt(iarr(inds(inds2(j))))=irt1(iarr(inds(inds2(j))))
                    jt(iarr(inds(inds2(j))))=jt1(iarr(inds(inds2(j))))
                 else
                    ist(iarr(inds(inds2(j))))=ist1(iarr(inds(inds2(l))))
                    irt(iarr(inds(inds2(j))))=irt1(iarr(inds(inds2(l))))
                    jt(iarr(inds(inds2(j))))=jt1(iarr(inds(inds2(l))))
                 endif
              enddo
           endif
           k=0
           dum=ser_muem(i)
        endif
        k=k+1
     enddo
     if(allocated(dimas))deallocate(dimas)
     allocate(dimas(k))
     if(allocated(inds))deallocate(inds)
     allocate(inds(k))
     if(allocated(inds2))deallocate(inds2)
     allocate(inds2(k))

     do j=1,k
        inds(j)=(i-1-k+j)
        dimas(j)=gri(ist1(iarr(inds(j))), irt1(iarr(inds(j))))
     enddo
     call piksrt(k,dimas,inds2)
     l=0
     do j=1,k
        if((l.eq.0).or.abs(dimas(l)-dimas(j))/dimas(l).gt.g_acc)then
!        if((l.eq.0).or.abs(dimas(l)-dimas(j)).gt.g_acc)then
           l=j
           ist(iarr(inds(inds2(j))))=ist1(iarr(inds(inds2(j))))
           irt(iarr(inds(inds2(j))))=irt1(iarr(inds(inds2(j))))
           jt(iarr(inds(inds2(j))))=jt1(iarr(inds(inds2(j))))
        else
           ist(iarr(inds(inds2(j))))=ist1(iarr(inds(inds2(l))))
           irt(iarr(inds(inds2(j))))=irt1(iarr(inds(inds2(l))))
           jt(iarr(inds(inds2(j))))=jt1(iarr(inds(inds2(l))))
        endif

     enddo

     if(allocated(istr))deallocate(istr)
     allocate(istr(ntot))
     if(allocated(dimas))deallocate(dimas)
     allocate(dimas(ntot))

     m=0
     n=0
     o=0

     do k=1,ntot
        if(muem_app(ist1(k),irt1(k),jt1(k)).lt.0)then
           istr(k)=0.0
           m=m+1
        else if((ist1(k).eq.ist(k)).and.(irt1(k).eq.irt(k)).and.(jt1(k).eq.jt(k)))then
           istr(k)=real(-ist(k))
           n=n+1
        else 
           istr(k)=real(ist(k))
           o=o+1
        endif
     enddo

     call  piksrt(ntot,istr,iarr)

     do k=1,ntot
        dimas(k)=real(ist1(iarr(k)))
     enddo
     do k=1,ntot
        ist1(k)=int(dimas(k))
     enddo

     do k=1,ntot
        dimas(k)=real(irt1(iarr(k)))
     enddo
     do k=1,ntot
        irt1(k)=int(dimas(k))
     enddo

     do k=1,ntot
        dimas(k)=real(jt1(iarr(k)))
     enddo
     do k=1,ntot
        jt1(k)=int(dimas(k))
     enddo

     do k=1,ntot
        dimas(k)=real(ist(iarr(k)))
     enddo
     do k=1,ntot
        ist(k)=int(dimas(k))
     enddo

     do k=1,ntot
        dimas(k)=real(irt(iarr(k)))
     enddo
     do k=1,ntot
        irt(k)=int(dimas(k))
     enddo

     do k=1,ntot
        dimas(k)=real(jt(iarr(k)))
     enddo
     do k=1,ntot
        jt(k)=int(dimas(k))
     enddo

     deallocate(ser_muem)
     deallocate(iarr)
     deallocate(dimas)
     deallocate(inds)
     deallocate(inds2)
     
  ENDIF

  RETURN
END SUBROUTINE GET_LMOD_SERIAL_MATRIX


INTEGER function get_nbg(Acc)
  IMPLICIT NONE

  REAL :: Acc

  if(Acc.lt.0.5) then
     get_nbg=10
  elseif(Acc.lt.1.5) then
     get_nbg=5
  elseif(Acc.lt.2.5) then
     get_nbg=2
  else
     get_nbg=1
  endif

  return
end function get_nbg


LOGICAL function elim_changed(earl, earh)
  IMPLICIT NONE

  REAL :: earl, earh
  REAL, SAVE :: cached_earl = -1.0
  REAL, SAVE :: cached_earh = -1.0

  if((cached_earl.ne.earl).or.(cached_earh.ne.earh)) then
     elim_changed=.TRUE.
     cached_earl=earl
     cached_earh=earh
  else
     elim_changed=.FALSE.
  endif

  return
end function elim_changed

SUBROUTINE piksrt(n,arr,iarr)
  INTEGER n
  REAL arr(n)
  INTEGER iarr(n)
  INTEGER i,j,ind
  REAL a

  do i=1,n
     iarr(i)=i
  enddo

  do j=2,n
     a=arr(j)
     ind=iarr(j)
     do i=j-1,1,-1
        if(arr(i).le.a)goto 10
        arr(i+1)=arr(i)
        iarr(i+1)=iarr(i)
     enddo
     i=0                                                                                                                  
10   arr(i+1)=a                                                                                                           
     iarr(i+1)=ind                                                                                                        
  enddo

  return                                                                                                                  
END SUBROUTINE piksrt


LOGICAL function do_recalculate_local_models(lparam, npar, mtp)
  use par_global, only: rnobb
  use par_xilcps, only: use_iref,use_xil
  use par_rel, only: was_wgt_invoked
  IMPLICIT NONE

  INTEGER :: npar,mtp,i 
  REAL :: lparam(npar)
  LOGICAL, SAVE :: firstcall=.TRUE.
  INTEGER, SAVE :: mtp_cached=-1
  REAL,SAVE, ALLOCATABLE :: cached_lparam(:)

  character(len=20) :: FGSOLR,FGMSTR
  character(len=10) :: FGSOLR_val,FGUI_val
  character(len=10), SAVE :: cached_FGSOLR_val=''

  LOGICAL, SAVE :: cached_rnobb,cached_use_iref,cached_use_xil
                                                                                                                               
  FGSOLR_val = trim(FGSOLR())

  FGUI_val= trim(FGMSTR("RUSEIREF"))
  if (len_trim(FGUI_val).gt.0) then
     read (FGUI_val,*) i
     if(i.eq.1)then
        use_iref=.TRUE.
     else
        use_iref=.FALSE.
     endif
  else
     use_iref=.TRUE.
  endif

  FGUI_val= trim(FGMSTR("RUSEXIL"))
  if (len_trim(FGUI_val).gt.0) then
     read (FGUI_val,*) i
     if(i.eq.1)then
        use_xil=.TRUE.
     else
        use_xil=.FALSE.
     endif
  else
     use_xil=.TRUE.
  endif
     do_recalculate_local_models=.FALSE.
     IF(firstcall.or.(mtp_cached.ne.mtp)) THEN
     if(allocated(cached_lparam))deallocate(cached_lparam)
     allocate(cached_lparam(npar))
     do i=1,npar
        cached_lparam(i)=lparam(i)
     enddo
     mtp_cached=mtp
     firstcall=.FALSE.
     do_recalculate_local_models=.TRUE.

     cached_FGSOLR_val=FGSOLR_val
     cached_rnobb=rnobb
     cached_use_iref=use_iref
     cached_use_xil=use_xil
  ELSE
     if(was_wgt_invoked)then
        do_recalculate_local_models=.TRUE.
     else
        do_recalculate_local_models=.FALSE.
        do i=1,npar
           if(cached_lparam(i).ne.lparam(i))then
              do_recalculate_local_models=.TRUE.
              cached_lparam(i)=lparam(i)
           endif
        enddo
        if(cached_FGSOLR_val.ne.FGSOLR_val) do_recalculate_local_models=.TRUE.
        if(cached_rnobb.neqv.rnobb) do_recalculate_local_models=.TRUE.
        if(cached_use_iref.neqv.use_iref) do_recalculate_local_models=.TRUE.
        if(cached_use_xil.neqv.use_xil) do_recalculate_local_models=.TRUE.
        cached_FGSOLR_val=FGSOLR_val
        cached_rnobb=rnobb
        cached_use_iref=use_iref
        cached_use_xil=use_xil
     endif

  ENDIF

  return
end function do_recalculate_local_models


REAL function get_limbing(rLmbp, mu)
  IMPLICIT NONE

  REAL :: rLmbp,mu
  real,parameter :: delta = 2.06 

  if(rLmbp.lt.0.5)then
     !         isotropic                                                                     
     get_limbing = 1.0
  elseif(rLmbp.lt.1.5)then
     !         electron scattering                                                           
     get_limbing = (1.0 + delta*mu)/delta
  else
     !         dabrowski lasenby                                                             
     get_limbing = alog(1.0 + 1.0/mu)
  endif

  return
end function get_limbing


SUBROUTINE CONVOLVE_LP(iml,irl,irh,emin,emax)

  use par_global, only: inenl,inenh,inen
  use par_rel, only: rel_ldat_r_final,conv_dat
  use par_rel, only: inbn,ng,nr,nr_gr,ncem,nint
  use par_rel, only: rRedshft
  use par_rel, only: gl_val,gh_val,mul,muh
  use par_rel, only: fphr_tot
  use par_rel, only: inbn_app, inen_app

  use par_rel, only: sn,nntc

  use par_rel, only: ngmap,rel_ldat_r_maped,is_lamppost_bb
  use par_rel, only: conv_dat_11,conv_dat_12,conv_dat_21,conv_dat_22,map_a_r_to_g

  use par_rel, only: ntc_d,ntc1_d,ntc_m,ntc1_m,ntc_u,ntc1_u
  use par_rel, only: ntc_ngr_d,ntc1_ngr_d,ntc_ngr_m,ntc1_ngr_m,ntc_ngr_u,ntc1_ngr_u

  use par_rel, only: tfnel_d,tfnel_m,tfnel_u
  use par_rel, only: tft_d,tft_ngr_d,tft_m,tft_ngr_m,tft_u,tft_ngr_u

  use par_rel, only: mumin_d,mumax_d,mumin_m,mumax_m,mumin_u,mumax_u
  use par_rel, only: mumin_ngr_d,mumax_ngr_d,mumin_ngr_m,mumax_ngr_m,mumin_ngr_u,mumax_ngr_u

  IMPLICIT NONE

  REAL :: emin,emax
  INTEGER :: iml,irl,irh,nbn_conv

  INTEGER :: is,ir,ig,iem,imu,i,binl,binh
  REAL :: enl,enh,nbin_scale,de,tff,fphr_totc
  REAL :: shft_fac
  double precision, ALLOCATABLE :: rel_ldat_r_app(:,:,:,:)
  double precision, ALLOCATABLE :: rel_ldat_r_tmp(:),rel_ldat_r_app_tmp(:)
  character FGMSTR*20, FGMSTR_val*10

  INTEGER isl,nseek,nog,igsl
  REAL tff1,tff2

  double precision, ALLOCATABLE :: rel_ldat_r_maped_app(:,:),rldr_maped(:,:)
  INTEGER :: igmap,spos,rpos,imupos,tfel_l_d(nint),tfel_l_m(nint),tfel_l_u(nint)
  INTEGER :: inbn_app_break
  REAL :: y,e,de_rec
  INTEGER :: icur
  FGMSTR_val = trim(FGMSTR("CONV_ACC"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) nbin_scale
     if(nbin_scale.lt.1.0e-2)then
        nbin_scale=1.0e-2
     else if(nbin_scale.gt.1.0)then
        nbin_scale=1.0
     endif
  else
     nbin_scale=0.5 !1.0
  endif

  shft_fac=1.0/(1.0+rRedshft)

  enl=max(emin/gh_val/shft_fac,real(inenl))
  enh=min(emax/gl_val/shft_fac,real(inenh))

  if(enl.eq.real(inenl))then
     binl=1
  else
     call f_binary_search_float(1.d0*enl,binl,1.d0*inen,inbn+1)
  endif
  if(enh.eq.real(inenh))then
     binh=inbn
  else
     call f_binary_search_float(1.d0*enh,binh,1.d0*inen,inbn+1)
  endif

  nbn_conv=binh-binl+1

  inbn_app=INT(nbn_conv*nbin_scale)
  if(inbn_app.lt.10)inbn_app=10
  if(inbn_app.gt.inbn)inbn_app=inbn
  if(inbn_app.lt.inbn)then

     IF(allocated(inen_app))deallocate(inen_app)
     allocate(inen_app(0:inbn_app))

     de=Log10(enh/enl)/inbn_app

     i=0
     inen_app(i)=enl*10d0**(i*de)
     do while(inen_app(i).lt.20.0)
        i=i+1
        inen_app(i)=enl*10d0**(i*de)
     enddo
     inbn_app_break=i
     y=1.428571429*(2.7-2.0*nbin_scale)
     i=1
     e=real(inen_app(inbn_app_break))
     icur=i+inbn_app_break
     de_rec=(de/(enh-enl)*((y-1.0)*e+enh-enl-y+1.0))
     inen_app(icur)=e*10d0**de_rec

     do while((inen_app(icur).lt.enh).and.(icur.lt.inbn_app))
        e=real(inen_app(icur))
        i=i+1        
        icur=i+inbn_app_break
        de_rec=(de/(enh-enl)*((y-1.0)*e+enh-enl-y+1.0))
        inen_app(icur)=e*10d0**de_rec
     enddo
     inbn_app=i+inbn_app_break

     IF(allocated(rel_ldat_r_tmp))deallocate(rel_ldat_r_tmp)
     allocate(rel_ldat_r_tmp(inbn))
     IF(allocated(rel_ldat_r_app_tmp))deallocate(rel_ldat_r_app_tmp)
     allocate(rel_ldat_r_app_tmp(inbn_app))

     IF(is_lamppost_bb)THEN

        IF(allocated(rel_ldat_r_app))deallocate(rel_ldat_r_app)
        allocate(rel_ldat_r_app(nint,nr,ncem,inbn_app))

        DO is=1,nint
           do ir=1,nr
              DO iem=1,ncem
                 do i=1,inbn
                    rel_ldat_r_tmp(i)=rel_ldat_r_final(is,ir,iem,i)
                 enddo
                 call f_rebin_spectrum(inen_app,inbn_app,inen,inbn,rel_ldat_r_app_tmp,rel_ldat_r_tmp)
                 do i=1,inbn_app
                    rel_ldat_r_app(is,ir,iem,i)=rel_ldat_r_app_tmp(i)
                 enddo
              ENDDO
           enddo
        ENDDO

     ELSE

        IF(allocated(rel_ldat_r_maped_app))deallocate(rel_ldat_r_maped_app)
        allocate(rel_ldat_r_maped_app(0:ngmap,inbn_app))

        rel_ldat_r_maped_app(0,:)=0.0
        DO igmap=1,ngmap
           DO iem=1,ncem
              do i=1,inbn
                 rel_ldat_r_tmp(i)=rel_ldat_r_maped(igmap,i)
              enddo
              call f_rebin_spectrum(inen_app,inbn_app,inen,inbn,rel_ldat_r_app_tmp,rel_ldat_r_tmp)
              do i=1,inbn_app
                 rel_ldat_r_maped_app(igmap,i)=rel_ldat_r_app_tmp(i)
              enddo
           ENDDO
        ENDDO


     ENDIF

     deallocate(rel_ldat_r_tmp)
     deallocate(rel_ldat_r_app_tmp)

  else

     inbn_app=inbn
     IF(allocated(inen_app))deallocate(inen_app)
     allocate(inen_app(0:inbn))
     DO i=0,inbn
        inen_app(i)=inen(i)
     ENDDO
     
     IF(is_lamppost_bb)THEN
        
        IF(allocated(rel_ldat_r_app))deallocate(rel_ldat_r_app)
        allocate(rel_ldat_r_app(nint,nr,ncem,inbn))

        DO is=1,nint
           do ir=1,nr
              DO iem=1,ncem
                 do i=1,inbn
                    rel_ldat_r_app(is,ir,iem,i)=rel_ldat_r_final(is,ir,iem,i)
                 enddo
              ENDDO
           enddo
        ENDDO
        
     ELSE
        
        IF(allocated(rel_ldat_r_maped_app))deallocate(rel_ldat_r_maped_app)
        allocate(rel_ldat_r_maped_app(0:ngmap,inbn_app))
     
        rel_ldat_r_maped_app=rel_ldat_r_maped
        
     ENDIF
     
  endif

  IF(is_lamppost_bb)THEN
  
     IF(allocated(conv_dat))deallocate(conv_dat)
     allocate(conv_dat(inbn_app,ng,nint,nint))

     DO is=1,nint
        DO imu=1,nint
           do i=1,inbn_app
              do ig=1,ng
                 conv_dat(i,ig,is,imu)=0.0
              enddo
           enddo
        ENDDO
     ENDDO

  ELSE
     
     IF(allocated(conv_dat_11))deallocate(conv_dat_11)
     allocate(conv_dat_11(inbn_app,ng))
     IF(allocated(conv_dat_12))deallocate(conv_dat_12)
     allocate(conv_dat_12(inbn_app,ng))
     IF(allocated(conv_dat_21))deallocate(conv_dat_21)
     allocate(conv_dat_21(inbn_app,ng))
     IF(allocated(conv_dat_22))deallocate(conv_dat_22)
     allocate(conv_dat_22(inbn_app,ng))
     
     conv_dat_11=0.0
     conv_dat_12=0.0
     conv_dat_21=0.0
     conv_dat_22=0.0

  ENDIF


  DO is=1,nint
     tfel_l_d(is)=0
     tfel_l_m(is)=0
     tfel_l_u(is)=0
     if((is+sn).gt.1)then
        do i=2,is+sn
           tfel_l_d(is)=tfel_l_d(is)+tfnel_d(i-1)
           tfel_l_m(is)=tfel_l_m(is)+tfnel_m(i-1)
           tfel_l_u(is)=tfel_l_u(is)+tfnel_u(i-1)
        enddo
     endif
  ENDDO

  IF(is_lamppost_bb)THEN

     DO ir=irl,min(irh,nr_gr)
        do is=1,nint
           spos=is+sn
           fphr_totc=fphr_tot(is,ir)
           DO imu=iml,iml+1
              imupos=imu-iml+1
              do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
                 DO isl=1,ntc_d(spos,ir,iem,imu)
                    igsl = ntc1_d(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
                    nog = ntc1_d(spos,isl,ir,iem,imu,2)
                    do ig=igsl,igsl+nog
                       tff=tft_d(nseek)*fphr_totc
                       DO i=1,inbn_app                
                          conv_dat(i,ig,is,imupos)=conv_dat(i,ig,is,imupos) &
                               +rel_ldat_r_app(is,ir,iem,i)*tff
                       ENDDO
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo

              do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
                 DO isl=1,ntc_m(spos,ir,iem,imu)
                    igsl = ntc1_m(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
                    nog = ntc1_m(spos,isl,ir,iem,imu,2)
                    do ig=igsl,igsl+nog
                       tff=tft_m(nseek)*fphr_totc
                       DO i=1,inbn_app                
                          conv_dat(i,ig,is,imupos)=conv_dat(i,ig,is,imupos) &
                               +rel_ldat_r_app(is,ir,iem,i)*tff
                       ENDDO
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
              do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
                 DO isl=1,ntc_u(spos,ir,iem,imu)
                    igsl = ntc1_u(spos,isl,ir,iem,imu,1)
                    nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
                    nog = ntc1_u(spos,isl,ir,iem,imu,2)
                    do ig=igsl,igsl+nog
                       tff=tft_u(nseek)*fphr_totc
                       DO i=1,inbn_app                
                          conv_dat(i,ig,is,imupos)=conv_dat(i,ig,is,imupos) &
                               +rel_ldat_r_app(is,ir,iem,i)*tff
                       ENDDO
                       nseek = nseek + 1
                    enddo
                 ENDDO
              enddo
           ENDDO
        enddo
     ENDDO

     DO ir=max((nr_gr+1),irl),irh                                                                     
        rpos=ir-nr_gr
        DO imu=iml,iml+1
           imupos=imu-iml+1
           do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
              DO isl=1,ntc_ngr_d(rpos,iem,imu)
                 igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff1=tft_ngr_d(nseek)*fphr_tot(1,ir)
                    tff2=tft_ngr_d(nseek)*fphr_tot(2,ir)
                    DO i=1,inbn_app                
                       conv_dat(i,ig,1,imupos)=conv_dat(i,ig,1,imupos) &
                            +rel_ldat_r_app(1,ir,iem,i)*tff1
                       conv_dat(i,ig,2,imupos)=conv_dat(i,ig,2,imupos) &
                            +rel_ldat_r_app(2,ir,iem,i)*tff2
                    ENDDO
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
              DO isl=1,ntc_ngr_m(rpos,iem,imu)
                 igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff1=tft_ngr_m(nseek)*fphr_tot(1,ir)
                    tff2=tft_ngr_m(nseek)*fphr_tot(2,ir)
                    DO i=1,inbn_app                
                       conv_dat(i,ig,1,imupos)=conv_dat(i,ig,1,imupos) &
                            +rel_ldat_r_app(1,ir,iem,i)*tff1
                       conv_dat(i,ig,2,imupos)=conv_dat(i,ig,2,imupos) &
                            +rel_ldat_r_app(2,ir,iem,i)*tff2
                    ENDDO
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
              DO isl=1,ntc_ngr_u(rpos,iem,imu)
                 igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
                 nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
                 nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff1=tft_ngr_u(nseek)*fphr_tot(1,ir)
                    tff2=tft_ngr_u(nseek)*fphr_tot(2,ir)
                    DO i=1,inbn_app                
                       conv_dat(i,ig,1,imupos)=conv_dat(i,ig,1,imupos) &
                            +rel_ldat_r_app(1,ir,iem,i)*tff1
                       conv_dat(i,ig,2,imupos)=conv_dat(i,ig,2,imupos) &
                            +rel_ldat_r_app(2,ir,iem,i)*tff2
                    ENDDO
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
        ENDDO
     ENDDO

     deallocate(rel_ldat_r_app)
     
  ELSE
    

     IF(allocated(rldr_maped))deallocate(rldr_maped)
     allocate(rldr_maped(ncem,inbn_app))

     DO ir=irl,min(irh,nr_gr)
        is=1
        spos=is+sn
        DO iem=mul,muh
           igmap=map_a_r_to_g(ir,is,iem)
           do i=1,inbn_app
              rldr_maped(iem,i)=rel_ldat_r_maped_app(igmap,i)
           enddo
        ENDDO
        fphr_totc=fphr_tot(is,ir)

        imu=iml
        do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
           DO isl=1,ntc_d(spos,ir,iem,imu)
              igsl = ntc1_d(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
              nog = ntc1_d(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_d(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
           DO isl=1,ntc_m(spos,ir,iem,imu)
              igsl = ntc1_m(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
              nog = ntc1_m(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_m(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
           DO isl=1,ntc_u(spos,ir,iem,imu)
              igsl = ntc1_u(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
              nog = ntc1_u(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_u(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        
        imu=iml+1
        do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
           DO isl=1,ntc_d(spos,ir,iem,imu)
              igsl = ntc1_d(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
              nog = ntc1_d(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_d(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
           DO isl=1,ntc_m(spos,ir,iem,imu)
              igsl = ntc1_m(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
              nog = ntc1_m(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_m(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
           DO isl=1,ntc_u(spos,ir,iem,imu)
              igsl = ntc1_u(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
              nog = ntc1_u(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_u(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo

        is=2
        spos=is+sn
        DO iem=mul,muh
           igmap=map_a_r_to_g(ir,is,iem)
           do i=1,inbn_app
              rldr_maped(iem,i)=rel_ldat_r_maped_app(igmap,i)
           enddo
        ENDDO
        fphr_totc=fphr_tot(is,ir)
        
        imu=iml
        do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
           DO isl=1,ntc_d(spos,ir,iem,imu)
              igsl = ntc1_d(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
              nog = ntc1_d(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_d(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
           DO isl=1,ntc_m(spos,ir,iem,imu)
              igsl = ntc1_m(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
              nog = ntc1_m(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_m(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
           DO isl=1,ntc_u(spos,ir,iem,imu)
              igsl = ntc1_u(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
              nog = ntc1_u(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_u(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo

        imu=iml+1
        do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
           DO isl=1,ntc_d(spos,ir,iem,imu)
              igsl = ntc1_d(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
              nog = ntc1_d(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_d(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
           DO isl=1,ntc_m(spos,ir,iem,imu)
              igsl = ntc1_m(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
              nog = ntc1_m(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_m(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
           DO isl=1,ntc_u(spos,ir,iem,imu)
              igsl = ntc1_u(spos,isl,ir,iem,imu,1)
              nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
              nog = ntc1_u(spos,isl,ir,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_u(nseek)*fphr_totc
                 DO i=1,inbn_app                
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
     ENDDO

     DO ir=max((nr_gr+1),irl),irh
        rpos=ir-nr_gr

        is=1
        DO iem=mul,muh
           igmap=map_a_r_to_g(ir,is,iem)
           do i=1,inbn_app
              rldr_maped(iem,i)=rel_ldat_r_maped_app(igmap,i)
           enddo
        ENDDO
        fphr_totc=fphr_tot(is,ir)

        imu=iml
        do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
           DO isl=1,ntc_ngr_d(rpos,iem,imu)
              igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_d(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
           DO isl=1,ntc_ngr_m(rpos,iem,imu)
              igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_m(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
           DO isl=1,ntc_ngr_u(rpos,iem,imu)
              igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_u(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_11(i,ig)=conv_dat_11(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo

        imu=iml+1
        do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
           DO isl=1,ntc_ngr_d(rpos,iem,imu)
              igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_d(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
           DO isl=1,ntc_ngr_m(rpos,iem,imu)
              igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_m(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
           DO isl=1,ntc_ngr_u(rpos,iem,imu)
              igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_u(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_12(i,ig)=conv_dat_12(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo

        is=2
        DO iem=mul,muh
           igmap=map_a_r_to_g(ir,is,iem)
           do i=1,inbn_app
              rldr_maped(iem,i)=rel_ldat_r_maped_app(igmap,i)
           enddo
        ENDDO
        fphr_totc=fphr_tot(is,ir)

        imu=iml
        do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
           DO isl=1,ntc_ngr_d(rpos,iem,imu)
              igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_d(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
           DO isl=1,ntc_ngr_m(rpos,iem,imu)
              igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_m(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
           DO isl=1,ntc_ngr_u(rpos,iem,imu)
              igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_u(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_21(i,ig)=conv_dat_21(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo

        imu=iml+1
        do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
           DO isl=1,ntc_ngr_d(rpos,iem,imu)
              igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_d(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
           DO isl=1,ntc_ngr_m(rpos,iem,imu)
              igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_m(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
           DO isl=1,ntc_ngr_u(rpos,iem,imu)
              igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_u(nseek)*fphr_totc
                 DO i=1,inbn_app
                    conv_dat_22(i,ig)=conv_dat_22(i,ig)+rldr_maped(iem,i)*tff
                 ENDDO
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        
     ENDDO

     deallocate(rldr_maped)

  ENDIF
  
  RETURN
END SUBROUTINE CONVOLVE_LP


SUBROUTINE SPIN_ANGLE_INTERP(iml)
  use par_global, only: Pi
  use par_rel, only: sn,sa,rIncl,ra,is_lamppost_bb
  use par_rel, only: conv_dat, conv_int_dat, nint, ng, inbn_app,ncobs
  use par_rel, only: conv_dat_11,conv_dat_12,conv_dat_21,conv_dat_22
  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u
  IMPLICIT NONE

  INTEGER :: i, iml, ig
  REAL:: spinl, spinh, incl, inch, cosi, factor1, factor2
  INTEGER,SAVE :: inbn_app_cached = -1

  IF(inbn_app_cached.ne.inbn_app) THEN 
     if(.not.allocated(conv_int_dat))then
        allocate(conv_int_dat(inbn_app,ng))
     else
        deallocate(conv_int_dat)
        allocate(conv_int_dat(inbn_app,ng))
     endif
     inbn_app_cached=inbn_app
  ENDIF

  spinl=sa(sn+1)
  spinh=sa(sn+2)

  incl=(iml-0.5)/ncobs
  inch=(iml+1-0.5)/ncobs
  cosi=real(COS(rIncl*Pi/180.0))

  factor1 = (ra-spinl)/(spinh-spinl)
  factor2 = (cosi-incl)/(inch-incl)

  IF(is_lamppost_bb)THEN
     DO i=1,inbn_app
        do ig=gl_d,gh_d
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat(i,ig,1,1) &
                +factor1*(1.0-factor2)*conv_dat(i,ig,2,1) &
                +(1.0-factor1)*factor2*conv_dat(i,ig,1,2) &
                +factor1*factor2*conv_dat(i,ig,2,2)
        enddo
        do ig=gl_m,gh_m
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat(i,ig,1,1) &
                +factor1*(1.0-factor2)*conv_dat(i,ig,2,1) &
                +(1.0-factor1)*factor2*conv_dat(i,ig,1,2) &
                +factor1*factor2*conv_dat(i,ig,2,2)
        enddo
        do ig=gl_u,gh_u
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat(i,ig,1,1) &
                +factor1*(1.0-factor2)*conv_dat(i,ig,2,1) &
                +(1.0-factor1)*factor2*conv_dat(i,ig,1,2) &
                +factor1*factor2*conv_dat(i,ig,2,2)
        enddo
     ENDDO
  ELSE
     DO i=1,inbn_app
        do ig=gl_d,gh_d
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat_11(i,ig) &
                +factor1*(1.0-factor2)*conv_dat_21(i,ig) &
                +(1.0-factor1)*factor2*conv_dat_12(i,ig) &
                +factor1*factor2*conv_dat_22(i,ig)
        enddo
        do ig=gl_m,gh_m
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat_11(i,ig) &
                +factor1*(1.0-factor2)*conv_dat_21(i,ig) &
                +(1.0-factor1)*factor2*conv_dat_12(i,ig) &
                +factor1*factor2*conv_dat_22(i,ig)
        enddo
        do ig=gl_u,gh_u
           conv_int_dat(i,ig)=(1.0-factor1)*(1.0-factor2)*conv_dat_11(i,ig) &
                +factor1*(1.0-factor2)*conv_dat_21(i,ig) &
                +(1.0-factor1)*factor2*conv_dat_12(i,ig) &
                +factor1*factor2*conv_dat_22(i,ig)
        enddo
     ENDDO
  ENDIF

  RETURN
END SUBROUTINE SPIN_ANGLE_INTERP


REAL function get_fphd_spin_int()
  use par_rel, only: sn,sa,ra,fphd_tot,nint
  IMPLICIT NONE

  REAL:: spinl, spinh, factor

  spinl=sa(sn+1)
  spinh=sa(sn+2)
  factor = (ra-spinl)/(spinh-spinl)

  get_fphd_spin_int=(1.0-factor)*fphd_tot(1)+factor*fphd_tot(2)

  return
end function get_fphd_spin_int


REAL function get_fphd_spin_int_1L()
  use par_rel, only: sn,sa,ra,fphd_tot_1L,nint
  IMPLICIT NONE

  REAL:: spinl, spinh, factor

  spinl=sa(sn+1)
  spinh=sa(sn+2)
  factor = (ra-spinl)/(spinh-spinl)

  get_fphd_spin_int_1L=(1.0-factor)*fphd_tot_1L(1)+factor*fphd_tot_1L(2)

  return
end function get_fphd_spin_int_1L




LOGICAL function rmodel_param_changed(param,num_param)
  IMPLICIT NONE

  INTEGER :: num_param,i
  REAL :: param(num_param)
  REAL, SAVE, ALLOCATABLE :: param_cached(:)
  INTEGER, SAVE :: num_param_cached=-1

  IF(num_param_cached.ne.num_param)THEN
     if(allocated(param_cached))deallocate(param_cached)
     allocate(param_cached(num_param))
     do i=1,num_param
        param_cached(i)=param(i)
     enddo
     num_param_cached=num_param
     rmodel_param_changed=.FALSE.
  ELSE
     rmodel_param_changed=.FALSE.
     do i=1,num_param
        if(param_cached(i).ne.param(i)) rmodel_param_changed=.TRUE.
        param_cached(i)=param(i)
     enddo
  ENDIF

  return
end function rmodel_param_changed


SUBROUTINE  READ_XSET_ELIM(ear_l,ear_h)
  IMPLICIT NONE
  REAL :: ear_l, ear_h

  character :: FGMSTR*20, FGMSTR_val*20

  FGMSTR_val=FGMSTR("RMIN")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) ear_l
  else
     ear_l=0.0
  endif

  FGMSTR_val=FGMSTR("RMAX")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) ear_h
  else
     ear_h=0.0
  endif

  RETURN
END SUBROUTINE READ_XSET_ELIM


SUBROUTINE READ_XSET_SHOW_ALL_FLAG(show_all)
  IMPLICIT NONE

  LOGICAL :: show_all
  REAL :: show_all_val

  character :: FGMSTR*20, FGMSTR_val*20

  FGMSTR_val=FGMSTR("SHOW_ALL")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) show_all_val
     if(show_all_val.ne.0)then
        show_all=.TRUE.
     else
        show_all=.FALSE.
     endif
  else
     show_all=.FALSE.
  endif

  RETURN
END SUBROUTINE READ_XSET_SHOW_ALL_FLAG


SUBROUTINE READ_XSET_COMPACT_PARAM(MBh,Dist,Norm)
  IMPLICIT NONE

  REAL :: MBh,Dist,Norm !MBh in solar mases, Dist in pc

  character :: FGMSTR*20, FGMSTR_val*20

  FGMSTR_val=FGMSTR("RMBH")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) MBh
  else
     MBh=0.0
  endif

  FGMSTR_val=FGMSTR("RDIST")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) Dist
  else
     Dist=0.0
  endif

  FGMSTR_val=FGMSTR("RNORM")
  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) Norm
  else
     Norm=0.0
  endif

  RETURN
END SUBROUTINE READ_XSET_COMPACT_PARAM


REAL FUNCTION gdcal(h,a)
  IMPLICIT NONE

  REAL ::h,a,r,ds,dc,delta,sigma,aa,omega,gnu
  r       = h
!  theta   = 0.
!  ds      = SIN(theta)*SIN(theta)
!  dc      = COS(theta)*COS(theta)
  ds=0.0
  dc=1.0
  delta   = r**2 - 2.*r + a**2
  sigma   = r**2 + dc*a**2
  aa      = (r**2 + a**2)**2 - ds*delta*a**2
  omega   = 2.*a*r/aa
  gnu     = delta*sigma/aa
  gdcal       = sqrt(gnu)

  return
END FUNCTION gdcal

doubleprecision Function FK(a, r)
  Implicit None

  REAL(kind=8):: a, r

  REAL(kind=8):: r_last, A1, A2, ac
  REAL(kind=8):: x, x0, x1, x2, x3
  REAL(kind=8):: F
  REAL(kind=8):: Pi3, Sq3, Sq6
  Parameter (Pi3 = 1.04719755, Sq3 = 1.73205081, Sq6 = 2.44948974)

  x = dsqrt(r)

  if(a .eq. 0.) then
     if(r .lt. 6.) goto 999
     F = x - Sq6 - 0.5*sq3 * dLog(5.82842712 * (x - Sq3) / (x + Sq3))
     FK = F / (x**5 * (x * x - 3))
  else
     A1 = 1 + ((1 - a * a)**(1./3.)) &
          * ((1 + a)**(1./3.) + (1 - a)**(1./3.))
     A2 = dsqrt(3 * a * a + A1 * A1)
     r_last = 3 + A2 - &
          dsign(1.d0,a) * dsqrt((3 - A1) * (3 + A1 + 2 * A2))
     if(r .lt. r_last) goto 999
     ac = dacos(a) / 3
     x0 = dsqrt(r_last)
     x1 = 2 * dcos(ac - Pi3)
     x2 = 2 * dcos(ac + Pi3)
     x3 = -2 * dcos(ac)

!     write(*,*) a, r, A1, A2, r_last, x0, x1, x2, x3                                                                              \
                                                                                                                                    
     F = x - x0 - 1.5 * a * dLog(x / x0)
     F = F - 3 * (x1 - a)**2 / (x1 * (x1 - x2) * (x1 - x3)) &
          * dLog((x - x1)/(x0 - x1))
     F = F - 3 * (x2 - a)**2 / (x2 * (x2 - x1) * (x2 - x3)) &
          * dLog((x - x2)/(x0 - x2))
     F = F - 3 * (x3 - a)**2 / (x3 * (x3 - x1) * (x3 - x2)) &
          * dLog((x - x3)/(x0 - x3))
     FK =  F / (x**4 * (2 * a - 3 * x + x**3))
  endif

  return

999 FK = 0.
  return

end Function FK


doubleprecision Function Planck(nu,T)
  use par_rel, only: h_P, k_B, c
  Implicit None

  REAL(kind=8):: nu,T


  Planck=(2.0*h_P*(nu**3.0))/(c**2.0)*(1.0/(exp((h_P*nu)/(k_B*T))-1.0))
  
  return

end Function Planck


LOGICAL function do_recalculate_lpbb()

  use par_global, only: is_lamppost_bb
  use par_rel, only: L_intr_loc,rD_bb,rMbh_bb,rNORM_bb,ralb_bb,rf_bb,rdmu_bb,rTherm_frac_bb_int,rfcol_bb,rRedshft
  use par_rel, only: ra,rIncl,rH,rRin,rRout,rlp2ext,rnobb,rRefl_frac,rTherm_frac_bb
  IMPLICIT NONE

  LOGICAL, SAVE :: firstcall=.TRUE.
  double precision, SAVE :: cached_L_intr_loc,cached_rMbh_bb,cached_rD_bb
  REAL, SAVE :: cached_rNORM_bb
  REAL, SAVE :: cached_ralb_bb
  REAL, SAVE :: cached_rf_bb
  REAL, SAVE :: cached_rTherm_frac_bb
  REAL, SAVE :: cached_rTherm_frac_bb_int
  REAL, SAVE :: cached_rfcol_bb
  REAL, SAVE :: cached_rdmu_bb
  REAL, SAVE :: cached_rRedshft
  REAL, SAVE :: cached_ra
  REAL, SAVE :: cached_rIncl
  REAL, SAVE :: cached_rH
  REAL, SAVE :: cached_rRin
  REAL, SAVE :: cached_rRout
  REAL, SAVE :: cached_rlp2ext
  LOGICAL, SAVE :: cached_is_lamppost_bb
  LOGICAL, SAVE :: cached_rnobb
  LOGICAL, SAVE :: cached_rf_gt0
  LOGICAL rf_gt0

  if(abs(rRefl_frac).gt.0.)then
     rf_gt0=.TRUE.
  else
     rf_gt0=.FALSE.
  endif

  do_recalculate_lpbb=.FALSE.
  IF(firstcall) THEN
     firstcall=.FALSE.

     cached_L_intr_loc=L_intr_loc
     cached_rD_bb=rD_bb
     cached_rMbh_bb=rMbh_bb
     cached_rNORM_bb=rNORM_bb
     cached_ralb_bb=ralb_bb
     cached_rf_bb=rf_bb
     cached_rTherm_frac_bb_int=rTherm_frac_bb_int
     cached_rTherm_frac_bb=rTherm_frac_bb
     cached_rfcol_bb=rfcol_bb
     cached_rdmu_bb=rdmu_bb
     cached_rRedshft=rRedshft
     cached_ra=ra
     cached_rIncl=rIncl
     cached_rH=rH
     cached_rRin=rRin
     cached_rRout=rRout
     cached_rlp2ext=rlp2ext
     cached_is_lamppost_bb=is_lamppost_bb
     cached_rnobb=rnobb
     cached_rf_gt0=rf_gt0
     
     do_recalculate_lpbb=.TRUE.
  ELSE
     if(cached_L_intr_loc.ne.L_intr_loc) do_recalculate_lpbb=.TRUE.
     if(cached_rD_bb.ne.rD_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rMbh_bb.ne.rMbh_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rNORM_bb.ne.rNORM_bb) do_recalculate_lpbb=.TRUE.
     if(cached_ralb_bb.ne.ralb_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rf_bb.ne.rf_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rTherm_frac_bb_int.ne.rTherm_frac_bb_int) do_recalculate_lpbb=.TRUE.
     if(cached_rTherm_frac_bb.ne.rTherm_frac_bb) do_recalculate_lpbb=.TRUE.     
     if(cached_rfcol_bb.ne.rfcol_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rdmu_bb.ne.rdmu_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rRedshft.ne.rRedshft) do_recalculate_lpbb=.TRUE.
     if(cached_ra.ne.ra) do_recalculate_lpbb=.TRUE.
     if(cached_rIncl.ne.rIncl) do_recalculate_lpbb=.TRUE.
     if(cached_rH.ne.rH) do_recalculate_lpbb=.TRUE.
     if(cached_rRin.ne.rRin) do_recalculate_lpbb=.TRUE.
     if(cached_rRout.ne.rRout) do_recalculate_lpbb=.TRUE.
     if(cached_rlp2ext.ne.rlp2ext) do_recalculate_lpbb=.TRUE.
     if(cached_is_lamppost_bb.neqv.is_lamppost_bb) do_recalculate_lpbb=.TRUE.
     if(cached_rnobb.neqv.rnobb) do_recalculate_lpbb=.TRUE.
     if(cached_rf_gt0.neqv.rf_gt0) do_recalculate_lpbb=.TRUE.

     cached_L_intr_loc=L_intr_loc
     cached_rD_bb=rD_bb
     cached_rNORM_bb=rNORM_bb
     cached_ralb_bb=ralb_bb
     cached_rf_bb=rf_bb
     cached_rTherm_frac_bb_int=rTherm_frac_bb_int
     cached_rTherm_frac_bb=rTherm_frac_bb
     cached_rfcol_bb=rfcol_bb
     cached_rdmu_bb=rdmu_bb
     cached_rRedshft=rRedshft
     cached_ra=ra
     cached_rIncl=rIncl
     cached_rH=rH
     cached_rRin=rRin
     cached_rRout=rRout
     cached_rlp2ext=rlp2ext
     cached_is_lamppost_bb=is_lamppost_bb
     cached_rnobb=rnobb
     cached_rf_gt0=rf_gt0
  ENDIF

  return
end function do_recalculate_lpbb

