! This file is part of the REFLKERR model.                                                                                                                                                
! You can redistribute it.                                                                                                          
! Please refer to Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942

! This file (xreflkerr_nolp.f90) contains relativistic coronal model routines, functions and variables 

MODULE par_rel_nolp
  USE par_global
  IMPLICIT NONE
  SAVE

  REAL :: rIndex1, rIndex2, rRbr
  double precision, ALLOCATABLE :: tflux_nlp(:,:,:,:),tflux_nlp_int(:,:)
  INTEGER :: inbn_app_p,inbn_app_r
  double precision, ALLOCATABLE :: inen_app_p(:), inen_app_r(:)
  double precision, ALLOCATABLE :: data_app_p(:,:), data_app_r(:,:)
  double precision, ALLOCATABLE :: conv_int_dat_p(:,:), conv_int_dat_r(:,:)
  REAL :: enrm_l,enrm_h
  LOGICAL :: tflux_reint

END MODULE par_rel_nolp


SUBROUTINE f_reflkerrline(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rLineE,rAcc,rRefl_frac
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 10
  integer,parameter :: num_lparam = 2
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrline(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.

  mtp=4
  rAcc=4.0
  rRefl_frac=-1.0

  lparam(1)=rLineE
  lparam(2)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrline


SUBROUTINE f_reflkerrexp(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rGamma,rRMtype,rAfe,rEcut,rLogXi,rIncl,rAcc
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 13
  integer,parameter :: num_lparam = 7
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)


  ifl=0
  photer(1)=0.0

  call init_par_reflkerrexp(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  mtp=3
  rRMtype=1.0
  RMtype=1.0
  rAcc=4.0

  lparam(1)=rGamma
  lparam(2)=rAfe
  lparam(3)=rEcut
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrexp


SUBROUTINE f_reflkerr(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rAcc,ruse_Ec
  use par_xilcps, only: RMtype
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_kTbb_l,g1_kTbb_h,g1_tau_l,g1_tau_h, &
       g1_incl_l,g1_incl_h,g1_y_l,g1_y_h
    IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 16
  integer,parameter :: num_lparam = 10
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerr(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  IF(abs(ruse_Ec).gt.0.0)THEN
     mtp=9
  ELSE
     mtp=1
  ENDIF

  rRMtype=1.0
  RMtype=1.0
  rAcc=4.0

  if(rGeom.eq.-1.0) rGeom=1.0
  if((abs(rGeom).gt.1.0).and.(abs(rGeom).lt.4.0)) rGeom=0.0

  IF(rGeom.eq.1.0)THEN
     if(rTauGam.lt.0.0)then
        g1_y_l=-g1_tau_h*(0.0078277886*rkTe)
        g1_y_h=-g1_tau_l*(0.0078277886*rkTe)
        IF(rTauGam.lt.g1_y_l)rTauGam=g1_y_l
        IF(rTauGam.gt.g1_y_h)rTauGam=g1_y_h
     else
        IF(rTauGam.lt.g1_tau_l)rTauGam=g1_tau_l
        IF(rTauGam.gt.g1_tau_h)rTauGam=g1_tau_h
     endif

     if(rkTe.lt.g1_kT_l)rkTe=g1_kT_l
     if(rkTe.gt.g1_kT_h)rkTe=g1_kT_h
     IF(rkTbb.gt.0.)THEN
        if(rkTbb.lt.g1_kTbb_l)rkTbb=g1_kTbb_l
        if(rkTbb.gt.g1_kTbb_h)rkTbb=g1_kTbb_h
     ELSE
        if((-rkTbb).lt.g1_kTbb_l)rkTbb=-g1_kTbb_l
        if((-rkTbb).gt.g1_kTbb_h)rkTbb=-g1_kTbb_h
     ENDIF
     if(rIncl.lt.g1_incl_l)rIncl=g1_incl_l
     if(rIncl.gt.g1_incl_h)rIncl=g1_incl_h
  ENDIF

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rkTbb
  lparam(8)=rGeom
  lparam(9)=ruse_Ec
  lparam(10)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerr


SUBROUTINE f_reflkerrG(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rAcc,ruse_Ec
  use par_xilcps, only: RMtype,pkTbb
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_incl_l,g1_incl_h

  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 15
  integer,parameter :: num_lparam = 9
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrG(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  IF(abs(ruse_Ec).gt.0.0)THEN
     mtp=10
  ELSE
     mtp=2
  ENDIF

  if(rGeom.ne.0.0)rGeom=1.0
  rkTbb=1.0e-3
  pkTbb=rkTbb
  rRMtype=1.0
  RMtype=rRMtype
  rAcc=4.0

  IF(rGeom.eq.1.0)THEN
     if(rkTe.lt.g1_kT_l)rkTe=g1_kT_l
     if(rkTe.gt.g1_kT_h)rkTe=g1_kT_h
     if(rIncl.lt.g1_incl_l)rIncl=g1_incl_l
     if(rIncl.gt.g1_incl_h)rIncl=g1_incl_h
  ENDIF
 
  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rGeom
  lparam(8)=ruse_Ec
  lparam(9)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrG

SUBROUTINE f_reflkerrexpD(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rGamma,rRMtype,rAfe,rEcut,rLogXi,rIncl,rAcc,rLogN
  use par_xilcps, only: RMtype
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 14
  integer,parameter :: num_lparam = 8
  REAL :: ear(0:ne),param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrexpD(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  mtp=7
  rRMtype=1.0
  RMtype=1.0
  rAcc=4.0

  lparam(1)=rGamma
  lparam(2)=rAfe
  lparam(3)=rEcut
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrexpD


SUBROUTINE f_reflkerrD(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rAcc,rLogN
  use par_xilcps, only: RMtype
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_kTbb_l,g1_kTbb_h,g1_tau_l,g1_tau_h, &
       g1_incl_l,g1_incl_h,g1_y_l,g1_y_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 16
  integer,parameter :: num_lparam = 10
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrD(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  if(rGeom.eq.-1.0) rGeom=1.0
  if((abs(rGeom).gt.1.0).and.(abs(rGeom).lt.4.0)) rGeom=0.0

  mtp=5
  rRMtype=1.0
  RMtype=1.0
  rAcc=4.0

  IF(rGeom.eq.1.0)THEN
     if(rTauGam.lt.0.0)then
        g1_y_l=-g1_tau_h*(0.0078277886*rkTe)
        g1_y_h=-g1_tau_l*(0.0078277886*rkTe)
        IF(rTauGam.lt.g1_y_l)rTauGam=g1_y_l
        IF(rTauGam.gt.g1_y_h)rTauGam=g1_y_h
     else
        IF(rTauGam.lt.g1_tau_l)rTauGam=g1_tau_l
        IF(rTauGam.gt.g1_tau_h)rTauGam=g1_tau_h
     endif

     if(rkTe.lt.g1_kT_l)rkTe=g1_kT_l
     if(rkTe.gt.g1_kT_h)rkTe=g1_kT_h
     IF(rkTbb.gt.0.)THEN
        if(rkTbb.lt.g1_kTbb_l)rkTbb=g1_kTbb_l
        if(rkTbb.gt.g1_kTbb_h)rkTbb=g1_kTbb_h
     ELSE
        if((-rkTbb).lt.g1_kTbb_l)rkTbb=-g1_kTbb_l
        if((-rkTbb).gt.g1_kTbb_h)rkTbb=-g1_kTbb_h
     ENDIF
     if(rIncl.lt.g1_incl_l)rIncl=g1_incl_l
     if(rIncl.gt.g1_incl_h)rIncl=g1_incl_h
  ENDIF

  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=rkTbb
  lparam(9)=rGeom
  lparam(10)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrD


SUBROUTINE f_reflkerrDG(ear, ne, param, ifl, photar, photer)
  USE par_global, only: mtp,is_lamppost,is_lamppost_bb
  use par_rel, only: rTauGam,rRMtype,rAfe,rkTe,rLogXi,rIncl,rkTbb,rGeom,rAcc,rLogN
  use par_xilcps, only: RMtype,pkTbb
  use par_xilcps, only: g1_kT_l,g1_kT_h,g1_incl_l,g1_incl_h
  IMPLICIT NONE

  INTEGER :: ne, ifl
  integer,parameter :: num_param = 15
  integer,parameter :: num_lparam = 9
  REAL :: ear(0:ne), param(num_param),photar(ne),photer(ne)
  REAL :: lparam(num_lparam)

  ifl=0
  photer(1)=0.0

  call init_par_reflkerrDG(param,num_param)

  is_lamppost=.FALSE.
  is_lamppost_bb=.FALSE.  

  if(rGeom.ne.0.0)rGeom=1.0

  mtp=6
  rkTbb=1.0e-3
  pkTbb=rkTbb
  rRMtype=1.0
  RMtype=rRMtype
  rAcc=4.0

  IF(rGeom.eq.1.0)THEN
     if(rkTe.lt.g1_kT_l)rkTe=g1_kT_l
     if(rkTe.gt.g1_kT_h)rkTe=g1_kT_h
     if(rIncl.lt.g1_incl_l)rIncl=g1_incl_l
     if(rIncl.gt.g1_incl_h)rIncl=g1_incl_h
  ENDIF
 
  lparam(1)=rTauGam
  lparam(2)=rAfe
  lparam(3)=rkTe
  lparam(4)=rLogXi
  lparam(5)=0.0
  lparam(6)=rIncl
  lparam(7)=rLogN
  lparam(8)=rGeom
  lparam(9)=0.0

  call main_nolp_model(ear, ne, lparam, num_lparam, photar)

  RETURN
END SUBROUTINE f_reflkerrDG


SUBROUTINE main_nolp_model(ear, ne, lparam, num_lparam, photar)
  USE par_global, only: Pi,mtp,inbn,inen,inphotar_p,inphotar_r,rchatty
  use par_rel, only: ra,rIncl,rRin,rRout,rAcc,rLimb,rmin,rmax,sa,ns,ng,ncem
  use par_rel, only: rRedshft,rRefl_frac,nint,nr,fphr_tot,wgt,muem_app
  use par_rel, only: rGeom,was_wgt_invoked
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel_nolp, only: inbn_app_p,inbn_app_r,inen_app_p,inen_app_r,data_app_p,data_app_r
  use par_rel_nolp, only: enrm_l,enrm_h,conv_int_dat_p,conv_int_dat_r,tflux_reint
  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u
  IMPLICIT NONE

  INTEGER :: ne
  REAL :: ear(0:ne), photar(ne)
  integer num_lparam
  REAL :: lparam(num_lparam)

  double precision :: inphotar_r_tmp(inbn),dp_photar(ne),dp_photar_tmp(ne)
  REAL,SAVE :: rel_ldat_fsp(ncem,inbn),rel_ldat_p(ncem,inbn)
  INTEGER :: ir,is,ig,iem,get_nbg,nbg,i,j
  INTEGER :: get_low_rbin,irl
  INTEGER :: get_high_rbin,irh
  INTEGER :: get_low_inc_bin,iml
  REAL :: rms,rmsv,Rin_g,Rout_g
  LOGICAL :: do_recalculate_local_models,local_models_recalc,elim_changed,elimc_val

  REAL :: mu,get_limbing,mu_limb_scl,redshft_scl
  INTEGER :: muems(ncem)

  REAL, SAVE :: cached_rIndex1=-1.0
  REAL, SAVE :: cached_rIndex2=-1.0
  REAL, SAVE :: cached_rRbr=-1.0
  REAL, SAVE :: cached_rLimb=-1.0
  REAL, SAVE :: cached_rRefl_frac=-1e-20
  REAL, SAVE :: cached_rRedshft=-1.0

  REAL,SAVE :: eminc = 1.0e24, emaxc = -1.0e24
  REAL :: ear_l, ear_h

  character FGMSTR*20, FGMSTR_val*10
  character, SAVE :: cached_FGMSTR_val*10=''

  REAL :: dum
  DOUBLE PRECISION,SAVE :: rprim_norm
 
  INTEGER :: nbnrm
  REAL,ALLOCATABLE :: ear_nrm(:)
  double precision,ALLOCATABLE ::dp_ph_nrm(:),dp_ph_nrm_tmp(:)
  LOGICAL :: was_tflux_calculated, was_prim_app, was_refl_app
  DOUBLE PRECISION,ALLOCATABLE :: gphotar(:)
  REAL,ALLOCATABLE :: ginear_nlp(:)
  INTEGER :: gnbn
  LOGICAL::LgRin,LgRout

  character FGMSTR_chat_val*20
  CHARACTER(LEN=20)::FGMSTR_lgrin_val,FGMSTR_lgrout_val

  LOGICAL,SAVE :: message_diplayed=.FALSE.
  
  IF(.not.message_diplayed)THEN                                                                                                                         
     write(*,"(A59)")"***********************************************************"                                                        
     write(*,"(A59)")"*   Relativistic reflection in the disc-corona geometry.  *"                                                        
     write(*,"(A59)")"*    If you use results of this model, please refer to    *"                                                        
     write(*,"(A59)")"* Niedzwiecki, Szanecki & Zdziarski 2019 MNRAS, 485, 2942 *"                                                        
     write(*,"(A59)")"***********************************************************"                                                        
     message_diplayed=.TRUE.                                                                                       
  ENDIF

  FGMSTR_chat_val= trim(FGMSTR("RCHATTY"))

  if (len_trim(FGMSTR_chat_val).gt.0) then
     read (FGMSTR_chat_val,*) i
     if(i.eq.1)then
        rchatty=.TRUE.
     else
        rchatty=.FALSE.
     endif
  else
     rchatty=.FALSE.
  endif

  FGMSTR_val = trim(FGMSTR("CONV_ACC"))

  FGMSTR_lgrin_val= trim(FGMSTR("RINLOG"))
  if (len_trim(FGMSTR_lgrin_val).gt.0) then
     read (FGMSTR_lgrin_val,*) i
     if(i.eq.1)then
        LgRin=.TRUE.
     else
        LgRin=.FALSE.
     endif
  else
     LgRin=.FALSE.
  endif

  FGMSTR_lgrout_val= trim(FGMSTR("ROUTLOG"))
  if (len_trim(FGMSTR_lgrout_val).gt.0) then
     read (FGMSTR_lgrout_val,*) i
     if(i.eq.1)then
        LgRout=.TRUE.
     else
        LgRout=.FALSE.
     endif
  else
     LgRout=.FALSE.
  endif


  call DEFINE_SPIN_ARRAY()

  if(ra.lt.sa(1)) ra=sa(1)
  if(ra.gt.sa(ns)) ra=sa(ns)

  nbg=get_nbg(rAcc)

  call check_table_paths()

  call INIT_TFF_TABLES_PATHS()

  call DEFINE_R_ARRAY()
  call LOAD_SIND_TABLES()

  call set_spin_bin()

  rmsv=rms(ra)
  if(.not.LgRin)then
     IF(rRin.lt.0.0)THEN
        Rin_g=abs(rRin)*rmsv
       ELSE
        Rin_g=rRin
     ENDIF
  else
     IF(rRin.le.0.0)THEN
        Rin_g=rmsv*10.0**(abs(rRin))
     ELSE
        Rin_g=10.0**rRin
     ENDIF
  endif
  if(Rin_g.lt.rmin)Rin_g=rmin
  if(Rin_g.gt.rmax)Rin_g=rmax

  if(.not.LgRout)then
     IF(rRout.lt.0.0)THEN
        Rout_g=abs(rRout)*rmsv
     ELSE
        Rout_g=rRout
     ENDIF
  else
     IF(rRout.le.0.0)THEN
        Rout_g=rmsv*10.0**(abs(rRout))
     ELSE
        Rout_g=10.0**rRout
     ENDIF
  endif
  if(Rout_g.lt.rmin)Rout_g=rmin
  if(Rout_g.gt.rmax)Rout_g=rmax

  irl=get_low_rbin(Rin_g)
  irh=get_high_rbin(Rout_g)
  iml=get_low_inc_bin(rIncl)

  DO is=1,nint
     do ir=1,nr
        fphr_tot(is,ir)=1.0
     enddo
  ENDDO

  call CALCULATE_WGT_ARRAY(Rin_g,Rout_g,iml,irl,irh,rAcc,wgt)

  call GET_APP_ANGLES_BINS(nbg,wgt)

  DO iem=1,ncem
     muems(iem)=-1
  ENDDO

  DO is=1,nint
     do ir=1,nr
        DO iem=1,ncem
           if(muem_app(is,ir,iem).gt.0) muems(iem)=muem_app(is,ir,iem)
        ENDDO
     enddo
  ENDDO
  
  IF((mtp.eq.1).or.(mtp.eq.9))call init_par_hrefl(lparam,num_lparam)
  IF((mtp.eq.2).or.(mtp.eq.10))THEN
     lparam(6)=30.0
     call init_par_hreflG(lparam,num_lparam)
  ENDIF
  IF(mtp.eq.3)call init_par_hreflexp(lparam,num_lparam)
  IF(mtp.eq.4)call init_par_xline(lparam,num_lparam)
  IF(mtp.eq.5)call init_par_hreflD(lparam,num_lparam)
  IF(mtp.eq.6)THEN
     lparam(6)=30.0
     call init_par_hreflDG(lparam,num_lparam)
  ENDIF
  IF(mtp.eq.7)call init_par_hreflexpD(lparam,num_lparam)

  if(mtp.ne.4)then

     local_models_recalc=.FALSE.
     IF (do_recalculate_local_models(lparam, num_lparam,mtp).or. & 
          (cached_rRefl_frac.ne.rRefl_frac))THEN

        call get_primary()

        DO iem=1,ncem
           if(muems(iem).gt.0)then
              lparam(6)=real(ACOS((muems(iem)-0.5)/ncem)*180./Pi)

              IF((mtp.eq.1).or.(mtp.eq.9))THEN
                 call init_par_hrefl(lparam,num_lparam)
                 if(rGeom.eq.1)then
                    call get_primary()
                 endif
              ENDIF
              IF((mtp.eq.2).or.(mtp.eq.10))THEN
                 call init_par_hreflG(lparam,num_lparam)
                 if(rGeom.eq.1)then
                    call get_primary()
                 endif
              ENDIF
              IF(mtp.eq.3)call init_par_hreflexp(lparam,num_lparam)
              IF(mtp.eq.5)THEN
                 call init_par_hreflD(lparam,num_lparam)
                 if(rGeom.eq.1)then
                    call get_primary()
                 endif
              ENDIF
              IF(mtp.eq.6)THEN
                 call init_par_hreflDG(lparam,num_lparam)
                 if(rGeom.eq.1)then
                    call get_primary()
                 endif
              ENDIF
              IF(mtp.eq.7)call init_par_hreflexpD(lparam,num_lparam)

              if(abs(rRefl_frac).gt.1.0e-22)then  
                 call get_reflection()

                 do i=1,inbn
                    inphotar_r_tmp(i)=1.0d0*inphotar_r(i)
                 enddo
              else
                 do i=1,inbn
                    inphotar_r_tmp(i)=0.0d0
                 enddo
              endif

              call get_finalsp(inbn, inphotar_r_tmp, 1.0d0*inphotar_p, rRefl_frac)

              do i=1,inbn
                 rel_ldat_fsp(iem,i)=real(inphotar_r_tmp(i))
                 rel_ldat_p(iem,i)=inphotar_p(i)
              enddo
           else
              do i=1,inbn
                 rel_ldat_fsp(iem,i)=0.0
                 rel_ldat_p(iem,i)=0.0
              enddo
           endif
        ENDDO
        local_models_recalc=.TRUE.
        cached_rRefl_frac=rRefl_frac

     ENDIF

  else 

     local_models_recalc=.FALSE.
     IF (do_recalculate_local_models(lparam, num_lparam,mtp) &
          .or.(cached_rLimb.ne.rLimb))THEN
        call get_primary()
        DO iem=1,ncem
           if(muems(iem).gt.0)then
              mu = (iem-0.5)/ncem
              mu_limb_scl=mu*get_limbing(rLimb,mu)
              do i=1,inbn
                 rel_ldat_fsp(iem,i)=real(mu_limb_scl*inphotar_p(i)/(inen(i)-inen(i-1)))
              enddo
           else
              do i=1,inbn
                 rel_ldat_fsp(iem,i)=0.0
              enddo
           endif
        ENDDO
        cached_rLimb=rLimb
        local_models_recalc=.TRUE.
     ENDIF

  endif


  was_tflux_calculated=.FALSE.
  IF(was_wgt_invoked.or. &
       (cached_rIndex1.ne.rIndex1).or. &
       (cached_rIndex2.ne.rIndex2).or. &
       (cached_rRbr.ne.rRbr))THEN
     call CALCULATE_TFLUX_NOLP(iml,irl,irh)
     cached_rIndex1=rIndex1
     cached_rIndex2=rIndex2
     cached_rRbr=rRbr
     was_tflux_calculated=.TRUE.
  ENDIF
  tflux_reint=.FALSE.
  call SPIN_ANGLE_TFLUX_INTERP(iml,was_tflux_calculated)

  redshft_scl=1.0/(1.0+rRedshft)
  lparam(6)=rIncl

  IF((mtp.eq.1).or.(mtp.eq.9))call init_par_hrefl(lparam,num_lparam)
  IF((mtp.eq.2).or.(mtp.eq.10))call init_par_hreflG(lparam,num_lparam)
  IF(mtp.eq.3)call init_par_hreflexp(lparam,num_lparam)
  IF(mtp.eq.4)call init_par_xline(lparam,num_lparam)
  IF(mtp.eq.5)call init_par_hreflD(lparam,num_lparam)
  IF(mtp.eq.6)call init_par_hreflDG(lparam,num_lparam)
  IF(mtp.eq.7)call init_par_hreflexpD(lparam,num_lparam)

  call READ_XSET_ELIM(ear_l,ear_h)
  IF((ear_l.lt.1e-34).or.(ear_h.lt.1e-34).or.(ear_h.le.ear_l))THEN
     ear_l=ear(0)
     ear_h=ear(ne)
     if(ne.eq.1) then
        eminc=ear_l
        emaxc=ear_h
     else
        IF(ear_l.lt.eminc)eminc=ear_l
        IF(ear_h.gt.emaxc)emaxc=ear_h
     endif
  ELSE
     eminc=ear_l
     emaxc=ear_h
  ENDIF

  elimc_val=elim_changed(eminc,emaxc)
  rprim_norm=1.0e11
  
  if(mtp.ne.4)then

     call CREATE_NORM_ENGRID()

     was_prim_app=.FALSE.
     IF(local_models_recalc.or. &
          (cached_rRedshft.ne.rRedshft))THEN

        IF(allocated(data_app_p))deallocate(data_app_p)
        allocate(data_app_p(ncem,inbn_app_p))

        call GET_APP_LOCAL_DATA(inbn,inen,rel_ldat_p,inbn_app_p,inen_app_p,data_app_p)
        was_prim_app=.TRUE.
     ENDIF

     IF(was_tflux_calculated.or.tflux_reint.or.was_prim_app)THEN
        if(allocated(conv_int_dat_p))deallocate(conv_int_dat_p)
        allocate(conv_int_dat_p(inbn_app_p,ng))

        call CONVOLVE_NLP(data_app_p,inbn_app_p,conv_int_dat_p)
     ENDIF

     nbnrm=1

     if(allocated(ear_nrm))deallocate(ear_nrm)
     if(allocated(dp_ph_nrm))deallocate(dp_ph_nrm)
     if(allocated(dp_ph_nrm_tmp))deallocate(dp_ph_nrm_tmp)
     allocate(ear_nrm(0:nbnrm))
     allocate(dp_ph_nrm(nbnrm))
     allocate(dp_ph_nrm_tmp(nbnrm))

     ear_nrm(0)=enrm_l
     ear_nrm(1)=enrm_h

     DO i=1,nbnrm
        dp_ph_nrm(i)=0.0
     ENDDO
     IF(gl_d.le.gh_d)THEN
        gnbn=gh_d-gl_d+1
        IF(allocated(ginear_nlp))deallocate(ginear_nlp)
        allocate(ginear_nlp(0:gnbn))
        IF(allocated(gphotar))deallocate(gphotar)
        allocate(gphotar(gnbn))

        do i=1,inbn_app_p
           dum=real(redshft_scl*sqrt(inen_app_p(i-1)*inen_app_p(i))*0.007)
           DO ig=gl_d-1,gh_d
              ginear_nlp(ig-gl_d+1)=dum*ig
           ENDDO
           DO ig=gl_d,gh_d
              gphotar(ig-gl_d+1)=conv_int_dat_p(i,ig)*7.0
           ENDDO
           call f_rebin_spectrum(1.d0*ear_nrm,nbnrm,1.d0*ginear_nlp,gnbn,dp_ph_nrm_tmp,gphotar)
           dp_ph_nrm(1)=dp_ph_nrm(1)+dp_ph_nrm_tmp(1)
        enddo
     ENDIF
     IF(gl_m.le.gh_m)THEN
        gnbn=gh_m-gl_m+1
        IF(allocated(ginear_nlp))deallocate(ginear_nlp)
        allocate(ginear_nlp(0:gnbn))
        IF(allocated(gphotar))deallocate(gphotar)
        allocate(gphotar(gnbn))

        do i=1,inbn_app_p
           dum=real(redshft_scl*sqrt(inen_app_p(i-1)*inen_app_p(i))*0.005)
           DO ig=gl_m-1,gh_m
              ginear_nlp(ig-gl_m+1)=dum*ig
           ENDDO
           DO ig=gl_m,gh_m
              gphotar(ig-gl_m+1)=conv_int_dat_p(i,ig)*5.0
           ENDDO
           call f_rebin_spectrum(1.d0*ear_nrm,nbnrm,1.d0*ginear_nlp,gnbn,dp_ph_nrm_tmp,gphotar)
           dp_ph_nrm(1)=dp_ph_nrm(1)+dp_ph_nrm_tmp(1)
        enddo
     ENDIF
     IF(gl_u.le.gh_u)THEN
        gnbn=gh_u-gl_u+1
        IF(allocated(ginear_nlp))deallocate(ginear_nlp)
        allocate(ginear_nlp(0:gnbn))
        IF(allocated(gphotar))deallocate(gphotar)
        allocate(gphotar(gnbn))

        do i=1,inbn_app_p
           dum=real(redshft_scl*sqrt(inen_app_p(i-1)*inen_app_p(i))*0.001)
           DO ig=gl_u-1,gh_u
              ginear_nlp(ig-gl_u+1)=dum*ig
           ENDDO
           DO ig=gl_u,gh_u
              gphotar(ig-gl_u+1)=conv_int_dat_p(i,ig)
           ENDDO
           call f_rebin_spectrum(1.d0*ear_nrm,nbnrm,1.d0*ginear_nlp,gnbn,dp_ph_nrm_tmp,gphotar)
           dp_ph_nrm(1)=dp_ph_nrm(1)+dp_ph_nrm_tmp(1)
        enddo
     ENDIF

     rprim_norm=dp_ph_nrm(1)/(enrm_h-enrm_l)
     deallocate(ear_nrm)
     deallocate(dp_ph_nrm)
     deallocate(dp_ph_nrm_tmp)

  endif

  was_refl_app=.FALSE.
  IF((elimc_val.or.was_wgt_invoked).or. &
       local_models_recalc.or. &
       (cached_rRedshft.ne.rRedshft).or. &
       (cached_FGMSTR_val.ne.FGMSTR_val))THEN

     call CREATE_FSP_ENGRID(eminc,emaxc)

     IF(allocated(data_app_r))deallocate(data_app_r)
     allocate(data_app_r(ncem,inbn_app_r))
     call GET_APP_LOCAL_DATA(inbn,inen,rel_ldat_fsp,inbn_app_r,inen_app_r,data_app_r)
     was_refl_app=.TRUE.

     cached_rRedshft=rRedshft
     cached_FGMSTR_val=FGMSTR_val
  ENDIF

  IF(was_tflux_calculated.or.tflux_reint.or.was_refl_app)THEN
     if(allocated(conv_int_dat_r))deallocate(conv_int_dat_r)
     allocate(conv_int_dat_r(inbn_app_r,ng))

     call CONVOLVE_NLP(data_app_r,inbn_app_r,conv_int_dat_r)
  ENDIF

  DO i=1,ne
     dp_photar(i)=0.0
  ENDDO

  IF(gl_d.le.gh_d)THEN
     gnbn=gh_d-gl_d+1
     IF(allocated(ginear_nlp))deallocate(ginear_nlp)
     allocate(ginear_nlp(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))

     do i=1,inbn_app_r
        dum=real(redshft_scl*sqrt(inen_app_r(i-1)*inen_app_r(i))*0.007)
        DO ig=gl_d-1,gh_d
           ginear_nlp(ig-gl_d+1)=dum*ig
        ENDDO
        DO ig=gl_d,gh_d
           gphotar(ig-gl_d+1)=conv_int_dat_r(i,ig)*7.0
        ENDDO
        call f_rebin_spectrum(1.d0*ear,ne,1.d0*ginear_nlp,gnbn,dp_photar_tmp,gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF
  IF(gl_m.le.gh_m)THEN
     gnbn=gh_m-gl_m+1
     IF(allocated(ginear_nlp))deallocate(ginear_nlp)
     allocate(ginear_nlp(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))

     do i=1,inbn_app_r
        dum=real(redshft_scl*sqrt(inen_app_r(i-1)*inen_app_r(i))*0.005)
        DO ig=gl_m-1,gh_m
           ginear_nlp(ig-gl_m+1)=dum*ig
        ENDDO
        DO ig=gl_m,gh_m
           gphotar(ig-gl_m+1)=conv_int_dat_r(i,ig)*5.0
        ENDDO
        call f_rebin_spectrum(1.d0*ear,ne,1.d0*ginear_nlp,gnbn,dp_photar_tmp,gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF
  IF(gl_u.le.gh_u)THEN
     gnbn=gh_u-gl_u+1
     IF(allocated(ginear_nlp))deallocate(ginear_nlp)
     allocate(ginear_nlp(0:gnbn))
     IF(allocated(gphotar))deallocate(gphotar)
     allocate(gphotar(gnbn))

     do i=1,inbn_app_r
        dum=real(redshft_scl*sqrt(inen_app_r(i-1)*inen_app_r(i))*0.001)
        DO ig=gl_u-1,gh_u
           ginear_nlp(ig-gl_u+1)=dum*ig
        ENDDO
        DO ig=gl_u,gh_u
           gphotar(ig-gl_u+1)=conv_int_dat_r(i,ig)
        ENDDO
        call f_rebin_spectrum(1.d0*ear,ne,1.d0*ginear_nlp,gnbn,dp_photar_tmp,gphotar)
        DO j=1,ne
           dp_photar(j)=dp_photar(j)+dp_photar_tmp(j)
        ENDDO
     enddo
  ENDIF

  IF(rprim_norm.gt.0.0)THEN
     DO i=1,ne
        photar(i)=real(dp_photar(i)/rprim_norm)
     ENDDO
  ELSE
     DO i=1,ne
        photar(i)=1.0
     ENDDO
  ENDIF

  deallocate(ginear_nlp)
  deallocate(gphotar)

  return
END SUBROUTINE main_nolp_model


SUBROUTINE CALCULATE_TFLUX_NOLP(iml,irl,irh)
  use par_rel, only: ng,nr_gr,ncem,nint
  use par_rel, only: r_in,r_out
  use par_rel_nolp, only: tflux_nlp,rIndex1,rIndex2,rRbr

  use par_rel, only: sn,nntc

  use par_rel, only: ntc_d,ntc1_d,ntc_m,ntc1_m,ntc_u,ntc1_u
  use par_rel, only: ntc_ngr_d,ntc1_ngr_d,ntc_ngr_m,ntc1_ngr_m,ntc_ngr_u,ntc1_ngr_u

  use par_rel, only: tfnel_d,tfnel_m,tfnel_u
  use par_rel, only: tft_d,tft_ngr_d,tft_m,tft_ngr_m,tft_u,tft_ngr_u

  use par_rel, only: mumin_d,mumax_d,mumin_m,mumax_m,mumin_u,mumax_u
  use par_rel, only: mumin_ngr_d,mumax_ngr_d,mumin_ngr_m,mumax_ngr_m,mumin_ngr_u,mumax_ngr_u

  IMPLICIT NONE

  INTEGER :: iml,irl,irh

  INTEGER :: i,is,ir,ig,iem,imu
  REAL :: q,r
  DOUBLE PRECISION :: scale,tff

  INTEGER isl,nseek,nog,igsl
  INTEGER :: spos,rpos,imupos,tfel_l_d(nint),tfel_l_m(nint),tfel_l_u(nint)

  IF(allocated(tflux_nlp))deallocate(tflux_nlp)
  allocate(tflux_nlp(ng,ncem,nint,nint))

  DO is=1,nint
     do imu=1,nint
        DO ig=1,ng
           do iem=1,ncem
              tflux_nlp(ig,iem,is,imu)=0.0
           enddo
        ENDDO
     enddo
  ENDDO

  DO is=1,nint
     tfel_l_d(is)=0
     tfel_l_m(is)=0
     tfel_l_u(is)=0
     if((is+sn).gt.1)then
        do i=2,is+sn
           tfel_l_d(is)=tfel_l_d(is)+tfnel_d(i-1)
           tfel_l_m(is)=tfel_l_m(is)+tfnel_m(i-1)
           tfel_l_u(is)=tfel_l_u(is)+tfnel_u(i-1)
        enddo
     endif
  ENDDO

  DO ir=irl,min(irh,nr_gr)
     r = sqrt(r_out(ir)*r_in(ir))
     if(r.gt.rRbr)then                                                                                                
        q = rIndex2
     else                                                                                                             
        q = rIndex1
     endif
     scale = (r_out(ir)-r_in(ir))*r*REAL(r/rRbr,kind=8)**(-q)
     do is=1,nint
        spos=is+sn
        DO imu=iml,iml+1
           imupos=imu-iml+1
           do iem=mumin_d(spos,ir,imu),mumax_d(spos,ir,imu)
              DO isl=1,ntc_d(spos,ir,iem,imu)
                 igsl = ntc1_d(spos,isl,ir,iem,imu,1)
                 nseek = tfel_l_d(is)+ntc1_d(spos,isl,ir,iem,imu,3)
                 nog = ntc1_d(spos,isl,ir,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff=tft_d(nseek)*scale
                    tflux_nlp(ig,iem,is,imupos)=tflux_nlp(ig,iem,is,imupos)+tff
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_m(spos,ir,imu),mumax_m(spos,ir,imu)
              DO isl=1,ntc_m(spos,ir,iem,imu)
                 igsl = ntc1_m(spos,isl,ir,iem,imu,1)
                 nseek = tfel_l_m(is)+ntc1_m(spos,isl,ir,iem,imu,3)
                 nog = ntc1_m(spos,isl,ir,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff=tft_m(nseek)*scale
                    tflux_nlp(ig,iem,is,imupos)=tflux_nlp(ig,iem,is,imupos)+tff
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
           do iem=mumin_u(spos,ir,imu),mumax_u(spos,ir,imu)
              DO isl=1,ntc_u(spos,ir,iem,imu)
                 igsl = ntc1_u(spos,isl,ir,iem,imu,1)
                 nseek = tfel_l_u(is)+ntc1_u(spos,isl,ir,iem,imu,3)
                 nog = ntc1_u(spos,isl,ir,iem,imu,2)
                 do ig=igsl,igsl+nog
                    tff=tft_u(nseek)*scale
                    tflux_nlp(ig,iem,is,imupos)=tflux_nlp(ig,iem,is,imupos)+tff
                    nseek = nseek + 1
                 enddo
              ENDDO
           enddo
        ENDDO
     enddo
  ENDDO

  DO ir=max((nr_gr+1),irl),irh
     rpos=ir-nr_gr
     r = sqrt(r_out(ir)*r_in(ir))
     if(r.gt.rRbr)then
        q = rIndex2
     else
        q = rIndex1
     endif
     scale = (r_out(ir)-r_in(ir))*r*REAL(r/rRbr,kind=8)**(-q)
     DO imu=iml,iml+1
        imupos=imu-iml+1
        do iem=mumin_ngr_d(rpos,imu),mumax_ngr_d(rpos,imu)
           DO isl=1,ntc_ngr_d(rpos,iem,imu)
              igsl = ntc1_ngr_d(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_d(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_d(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_d(nseek)*scale
                 tflux_nlp(ig,iem,1,imupos)=tflux_nlp(ig,iem,1,imupos)+tff
                 tflux_nlp(ig,iem,2,imupos)=tflux_nlp(ig,iem,2,imupos)+tff
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_m(rpos,imu),mumax_ngr_m(rpos,imu)
           DO isl=1,ntc_ngr_m(rpos,iem,imu)
              igsl = ntc1_ngr_m(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_m(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_m(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_m(nseek)*scale
                 tflux_nlp(ig,iem,1,imupos)=tflux_nlp(ig,iem,1,imupos)+tff
                 tflux_nlp(ig,iem,2,imupos)=tflux_nlp(ig,iem,2,imupos)+tff
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
        do iem=mumin_ngr_u(rpos,imu),mumax_ngr_u(rpos,imu)
           DO isl=1,ntc_ngr_u(rpos,iem,imu)
              igsl = ntc1_ngr_u(isl,rpos,iem,imu,1)
              nseek = ntc1_ngr_u(isl,rpos,iem,imu,3)
              nog = ntc1_ngr_u(isl,rpos,iem,imu,2)
              do ig=igsl,igsl+nog
                 tff=tft_ngr_u(nseek)*scale
                 tflux_nlp(ig,iem,1,imupos)=tflux_nlp(ig,iem,1,imupos)+tff
                 tflux_nlp(ig,iem,2,imupos)=tflux_nlp(ig,iem,2,imupos)+tff
                 nseek = nseek + 1
              enddo
           ENDDO
        enddo
     ENDDO
  ENDDO

  RETURN
END SUBROUTINE CALCULATE_TFLUX_NOLP


SUBROUTINE SPIN_ANGLE_TFLUX_INTERP(iml,was_tflux_calculated)
  use par_global, only: Pi
  use par_rel, only: sn,sa,rIncl,ra
  use par_rel, only: nint,ng,ncobs,ncem,mul,muh
  use par_rel_nolp, only: tflux_nlp,tflux_nlp_int,tflux_reint
  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u
  IMPLICIT NONE

  INTEGER :: iml, iem,ig
  LOGICAL :: was_tflux_calculated
  REAL:: spinl, spinh, incl, inch, cosi, factor1, factor2
  REAL,SAVE :: cached_ra=-1.0
  REAL,SAVE :: cached_rIncl=-1.0


  IF((cached_ra.ne.ra).or.(cached_rIncl.ne.rIncl).or.was_tflux_calculated)THEN

     IF(.not.allocated(tflux_nlp_int))allocate(tflux_nlp_int(ng,ncem))

     spinl=sa(sn+1)
     spinh=sa(sn+2)

     incl=(iml-0.5)/ncobs
     inch=(iml+1-0.5)/ncobs
     cosi=real(COS(rIncl*Pi/180.0))

     factor1 = (ra-spinl)/(spinh-spinl)
     factor2 = (cosi-incl)/(inch-incl)

     DO ig=1,ng
        do iem=1,ncem
           tflux_nlp_int(ig,iem)=0.0
        enddo
     ENDDO

     DO ig=gl_d,gh_d
        do iem=mul,muh
           tflux_nlp_int(ig,iem)=(1.0-factor1)*(1.0-factor2)*tflux_nlp(ig,iem,1,1) &
                +factor1*(1.0-factor2)*tflux_nlp(ig,iem,2,1) &
                +(1.0-factor1)*factor2*tflux_nlp(ig,iem,1,2) &
                +factor1*factor2*tflux_nlp(ig,iem,2,2)
        enddo
     ENDDO
     DO ig=gl_m,gh_m
        do iem=mul,muh
           tflux_nlp_int(ig,iem)=(1.0-factor1)*(1.0-factor2)*tflux_nlp(ig,iem,1,1) &
                +factor1*(1.0-factor2)*tflux_nlp(ig,iem,2,1) &
                +(1.0-factor1)*factor2*tflux_nlp(ig,iem,1,2) &
                +factor1*factor2*tflux_nlp(ig,iem,2,2)
        enddo
     ENDDO
     DO ig=gl_u,gh_u
        do iem=mul,muh
           tflux_nlp_int(ig,iem)=(1.0-factor1)*(1.0-factor2)*tflux_nlp(ig,iem,1,1) &
                +factor1*(1.0-factor2)*tflux_nlp(ig,iem,2,1) &
                +(1.0-factor1)*factor2*tflux_nlp(ig,iem,1,2) &
                +factor1*factor2*tflux_nlp(ig,iem,2,2)
        enddo
     ENDDO
     
     cached_ra=ra
     cached_rIncl=rIncl
     tflux_reint=.TRUE.
  ENDIF

  RETURN
END SUBROUTINE SPIN_ANGLE_TFLUX_INTERP


SUBROUTINE CREATE_NORM_ENGRID()
  use par_global, only: inen,inbn,inenl,inenh
  use par_rel, only: gl,gh,rRedshft
  use par_rel_nolp, only: inbn_app_p,inen_app_p,enrm_l,enrm_h
  use par_rel, only: gl_val,gh_val
  IMPLICIT NONE

  REAL :: de,elow,ehigh
  INTEGER :: i
  REAL :: shft_fac

  INTEGER, SAVE :: gl_cached=-1
  INTEGER, SAVE :: gh_cached=-1
  REAL, SAVE :: rRedshft_cached=-1.0
  INTEGER::inbn_app_break
  REAL::y,e,nbin_scale,de_rec
  INTEGER::icur

  character FGMSTR*20, FGMSTR_val*10

  FGMSTR_val = trim(FGMSTR("CONV_ACC"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) nbin_scale
     if(nbin_scale.lt.1.0e-2)then
        nbin_scale=1.0e-2
     else if(nbin_scale.gt.1.0)then
        nbin_scale=1.0
     endif
  else
     nbin_scale=0.5 !1.0
  endif

  IF((gl_cached.ne.gl).or.(gh_cached.ne.gh).or.(rRedshft_cached.ne.rRedshft))THEN

     shft_fac=1.0/(1.0+rRedshft)

     call f_binary_search_float(1.d0,i,inen,inbn)
     enrm_l=real(inen(i-1))
     enrm_h=real(inen(i))

     elow=max(enrm_l/gh_val/shft_fac,real(inenl))
     ehigh=min(enrm_h/gl_val/shft_fac,real(inenh))
 
     inbn_app_p=gh-gl

     IF(allocated(inen_app_p))deallocate(inen_app_p)
     allocate(inen_app_p(0:inbn_app_p))

     de=Log10(ehigh/elow)/inbn_app_p

     if(ehigh.gt.20.0)then
     i=0
     inen_app_p(i)=elow*10d0**(i*de)
     do while(inen_app_p(i).lt.20.0)
        i=i+1
        inen_app_p(i)=elow*10d0**(i*de)
     enddo
     inbn_app_break=i

     y=1.428571429*(2.7-2.0*nbin_scale)
     i=1
     e=real(inen_app_p(inbn_app_break))
     icur=i+inbn_app_break
     de_rec=(de/(ehigh-elow)*((y-1.0)*e+ehigh-elow-y+1.0))
     inen_app_p(icur)=e*10d0**de_rec
     do while((inen_app_p(icur).lt.ehigh).and.(icur.lt.inbn_app_p))
        e=real(inen_app_p(icur))
        i=i+1
        icur=i+inbn_app_break
        de_rec=(de/(ehigh-elow)*((y-1.0)*e+ehigh-elow-y+1.0))        
        inen_app_p(icur)=e*10d0**de_rec
     enddo
     inbn_app_p=i+inbn_app_break
     
  else
     
     DO i=0,inbn_app_p
        inen_app_p(i)=elow*10d0**(i*de)
     ENDDO
     
  endif
     
     gl_cached=gl
     gh_cached=gh
     rRedshft_cached=rRedshft
  ENDIF

  RETURN
END SUBROUTINE CREATE_NORM_ENGRID


SUBROUTINE CREATE_FSP_ENGRID(emin,emax)
  use par_global, only: inenl,inenh,inen,inbn
  use par_rel, only: rRedshft
  use par_rel_nolp, only: inbn_app_r,inen_app_r
  use par_rel, only: gl_val,gh_val,rRedshft
  IMPLICIT NONE

  REAL :: emin,emax,enl,enh,de
  INTEGER :: i,binl,binh,nbn_conv
  REAL :: nbin_scale 
  character FGMSTR*20, FGMSTR_val*10
  REAL :: shft_fac
  INTEGER::inbn_app_break
  REAL::y,e,de_rec
  INTEGER::icur

  FGMSTR_val = trim(FGMSTR("CONV_ACC"))

  if (len_trim(FGMSTR_val).gt.0) then
     read (FGMSTR_val,*) nbin_scale
     if(nbin_scale.lt.1.0e-2)then
        nbin_scale=1.0e-2
     else if(nbin_scale.gt.1.0)then
        nbin_scale=1.0
     endif
  else
     nbin_scale=0.5 !1.0
  endif

  shft_fac=1.0/(1.0+rRedshft)

  enl=max(emin/gh_val/shft_fac,real(inenl))
  enh=min(emax/gl_val/shft_fac,real(inenh))

  if(enl.eq.real(inenl))then                                                                                                            
     binl=1                                                                                                                             
  else                                                                                                                                  
     call f_binary_search_float(1.d0*enl,binl,1.d0*inen,inbn+1)                                                                                      
  endif
  if(enh.eq.real(inenh))then                                                                                                            
     binh=inbn
  else
     call f_binary_search_float(1.d0*enh,binh,1.d0*inen,inbn+1)
  endif

  nbn_conv=binh-binl+1

  inbn_app_r=INT(nbn_conv*nbin_scale)
  if(inbn_app_r.lt.10)inbn_app_r=10
  if(inbn_app_r.gt.inbn)inbn_app_r=inbn

  IF(allocated(inen_app_r))deallocate(inen_app_r)

  IF(inbn_app_r.ge.inbn)THEN
     inbn_app_r=inbn
     allocate(inen_app_r(0:inbn_app_r))
     DO i=0,inbn_app_r
        inen_app_r(i)=inen(i)
     ENDDO
  ELSE
     allocate(inen_app_r(0:inbn_app_r))
     de=Log10(enh/enl)/inbn_app_r

     i=0
     inen_app_r(i)=enl*10d0**(i*de)
     do while(inen_app_r(i).lt.20.0)
        i=i+1
        inen_app_r(i)=enl*10d0**(i*de)
     enddo
     inbn_app_break=i

     y=1.428571429*(2.7-2.0*nbin_scale)
     i=1
     e=real(inen_app_r(inbn_app_break))
     icur=i+inbn_app_break
     de_rec=(de/(enh-enl)*((y-1.0)*e+enh-enl-y+1.0))
     inen_app_r(icur)=e*10d0**de_rec

     do while((inen_app_r(icur).lt.enh).and.(icur.lt.inbn_app_r))
        e=real(inen_app_r(icur))
        i=i+1
        icur=i+inbn_app_break
        de_rec=(de/(enh-enl)*((y-1.0)*e+enh-enl-y+1.0))        
        inen_app_r(icur)=e*10d0**de_rec
     enddo
     inbn_app_r=i+inbn_app_break

  ENDIF

  RETURN
END SUBROUTINE CREATE_FSP_ENGRID


SUBROUTINE GET_APP_LOCAL_DATA(nbin_in,en_in,data_in,nbin_out,en_out,data_out)
  use par_rel, only: ncem
  IMPLICIT NONE

  INTEGER :: nbin_in,nbin_out
  REAL :: data_in(ncem,nbin_in)
  double precision :: en_in(0:nbin_in),en_out(0:nbin_out),data_out(ncem,nbin_out)
  INTEGER ::iem,i
  double precision, ALLOCATABLE :: rel_ldat_tmp(:),rel_ldat_app_tmp(:)

  IF(nbin_out.ge.nbin_in)THEN

     DO iem=1,ncem
        do i=1,nbin_in
           data_out(iem,i)=data_in(iem,i)
        enddo
     ENDDO

  ELSE

     allocate(rel_ldat_tmp(nbin_in))
     allocate(rel_ldat_app_tmp(nbin_out))

     DO iem=1,ncem
        do i=1,nbin_in
           rel_ldat_tmp(i)=data_in(iem,i)
        enddo
        call f_rebin_spectrum(en_out,nbin_out,en_in,nbin_in,rel_ldat_app_tmp,rel_ldat_tmp)
        do i=1,nbin_out
           data_out(iem,i)=rel_ldat_app_tmp(i)
        enddo
     ENDDO

     deallocate(rel_ldat_tmp)
     deallocate(rel_ldat_app_tmp)

  ENDIF

  RETURN
END SUBROUTINE GET_APP_LOCAL_DATA


SUBROUTINE CONVOLVE_NLP(rel_ldat_in,inbn_in,conv_int_dat_out)
  use par_rel, only: ng,ncem,mul,muh
  use par_rel_nolp, only: tflux_nlp_int
  use par_rel, only: gl_d,gh_d,gl_m,gh_m,gl_u,gh_u
  IMPLICIT NONE

  INTEGER :: inbn_in
  double precision :: rel_ldat_in(ncem,inbn_in),conv_int_dat_out(inbn_in,ng)

  INTEGER :: ig,iem,i

  DO ig=1,ng
     do i=1,inbn_in
        conv_int_dat_out(i,ig)=0.0
     enddo
  ENDDO

  DO ig=gl_d,gh_d
     do i=1,inbn_in
        DO iem=mul,muh
           conv_int_dat_out(i,ig)=conv_int_dat_out(i,ig) &                            
                +rel_ldat_in(iem,i)*tflux_nlp_int(ig,iem)
        ENDDO
     enddo
  ENDDO
  DO ig=gl_m,gh_m
     do i=1,inbn_in
        DO iem=mul,muh
           conv_int_dat_out(i,ig)=conv_int_dat_out(i,ig) &                            
                +rel_ldat_in(iem,i)*tflux_nlp_int(ig,iem)
        ENDDO
     enddo
  ENDDO
  DO ig=gl_u,gh_u
     do i=1,inbn_in
        DO iem=mul,muh
           conv_int_dat_out(i,ig)=conv_int_dat_out(i,ig) &                            
                +rel_ldat_in(iem,i)*tflux_nlp_int(ig,iem)
        ENDDO
     enddo
  ENDDO


  RETURN
END SUBROUTINE CONVOLVE_NLP


SUBROUTINE INIT_PAR_REFLKERRLINE(param, num_param)                            
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rLineE, rLimb, rRedshft
                                                                                                                  
  IMPLICIT NONE                                                                                                   
  integer :: num_param                                                                                            
  real :: param(num_param)                                                                                        

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rLineE=    param(8)                                                                                               
  rLimb=     param(9)                                                                                                  
  rRedshft=      param(10)

  if(rIncl.gt.89.2)rIncl=89.2                                                                                     
  if(rIncl.lt.9.1)rIncl=9.1                                                                                       
                                                                                                                  
  RETURN                                                                                                          
END SUBROUTINE INIT_PAR_REFLKERRLINE


SUBROUTINE INIT_PAR_REFLKERREXP(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rGamma, rAfe, rEcut, rLogXi
  use par_rel, only: rRedshft,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rGamma=    param(8)
  rAfe=      param(9)
  rEcut=     param(10)
  rLogXi=       param(11)
  rRedshft=      param(12)
  rRefl_frac=param(13)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERREXP

SUBROUTINE INIT_PAR_REFLKERREXPD(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rGamma, rAfe, rEcut, rLogXi, rLogN
  use par_rel, only: rRedshft,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rGamma=    param(8)
  rAfe=      param(9)
  rEcut=     param(10)
  rLogXi=       param(11)
  rRedshft=      param(12)
  rLogN=     param(13)
  rRefl_frac=param(14)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERREXPD


SUBROUTINE INIT_PAR_REFLKERR(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rkTbb, rLogXi, rGeom
  use par_rel, only: rRedshft,ruse_Ec,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)                                                                                 
  ra=        param(4)                                                                                 
  rIncl=     param(5)                                                                                 
  rRin=      param(6)                                                                                 
  rRout=     param(7)
  rTauGam=   param(8)
  rAfe=      param(9)
  rkTe=      param(10)
  rkTbb=     param(11)
  rGeom=     param(12)
  ruse_Ec=   param(13)
  rLogXi=    param(14)
  rRedshft=  param(15)
  rRefl_frac=param(16)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERR


SUBROUTINE INIT_PAR_REFLKERRG(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rLogXi, rGeom
  use par_rel, only: rRedshft,ruse_Ec,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rTauGam=   param(8)
  rAfe=      param(9)
  rkTe=      param(10)
  rGeom=     param(11)
  ruse_Ec=   param(12)
  rLogXi=    param(13)
  rRedshft=  param(14)
  rRefl_frac=param(15)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRG


SUBROUTINE INIT_PAR_REFLKERRD(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rkTbb, rGeom, rLogXi, rLogN
  use par_rel, only: rRedshft,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rTauGam=   param(8)
  rAfe=      param(9)
  rkTe=      param(10)
  rkTbb=     param(11)
  rGeom=     param(12)
  rLogXi=    param(13)
  rRedshft=  param(14)
  rLogN=     param(15)
  rRefl_frac=param(16)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRD


SUBROUTINE INIT_PAR_REFLKERRDG(param, num_param)
  use par_rel_nolp, only: rIndex1, rIndex2, rRbr
  use par_rel, only: ra, rIncl, rRin, rRout, rTauGam, rAfe, rkTe, rGeom, rLogXi,rLogN
  use par_rel, only: rRedshft,rRefl_frac

  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)

  rIndex1=   param(1)
  rIndex2=   param(2)
  rRbr=      param(3)
  ra=        param(4)
  rIncl=     param(5)
  rRin=      param(6)
  rRout=     param(7)
  rTauGam=   param(8)
  rAfe=      param(9)
  rkTe=      param(10)
  rGeom=     param(11)
  rLogXi=    param(12)
  rRedshft=  param(13)
  rLogN=     param(14)
  rRefl_frac=param(15)

  if(rIncl.gt.89.2)rIncl=89.2
  if(rIncl.lt.9.1)rIncl=9.1

  RETURN
END SUBROUTINE INIT_PAR_REFLKERRDG
